// gmodstat.cpp
//
// Version 0.1, September 2001
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// CAVEAT: gmodstat is growing organically, so the code struggles to
// keep itself clean and orderly. Like an old city, many function
// and variable names are built on previous code that has since
// disappeared or evolved. gmodstat was not designed and written in
// one sitting. I'm gradually trying to clean it up... apologies
// for those sections that do not (yet) make any sense.
//
// gmodstat provides post-analysis of QuakeIII Arena logfiles generated
// by servers running gmod1.0. You can analyse client NAT usage, playing
// time trends and ping distributions. gmod1.0 modifies the server to
// produce more information in its logfiles (to add ipaddr/port per client
// Connect, and log ping times and ping aggregate histograms - none of this
// is done by the default Quake3 server).
//

#include "stdafx.h"
#define QLOGFIRSTPASS
#include "gmodstat.h"

int main(int argc, char* argv[])
{
	struct global_names gn;
	struct global_state gs;

	// Initialize major parameters
	
	init_gmodstat_state(&gs, &gn);

	// Parse command line

	while (argc>1)
	 if (argv[1][0]=='-')
      {
        switch (argv[1][1])
		{
			
		 case 'c':
			{	// Set alternate name for the config file
				strcpy(gn.configfile, argv[2]);
			}
			break;

		 case 'o':
			{	// Specify a different output filename suffix
				// (default is logout.txt)
				strcpy(gn.global_outfile, argv[2]);
			}
			break;

        case 'v':
               {       // Print the version number then quit.
                       printf("%s, version %s\n", PACKAGE, VERSION);
                       exit(-1);
               }
               break;

		default:
		case 'h':
			printf("%s -c <file> -o <file> -v\n", PACKAGE);
			printf("\n");
			printf("\t-c <file>  Config file is <file>\n");
			printf("\t-o <file>  Output file suffix is <file>\n");
			printf("\t-v         Print name and version number\n");
			printf("\n");
			printf("\tQuakeIII server logfile analysis.\n");
			printf("\t(c) 2001, Grenville Armitage, gj_armitage@yahoo.com\n\n");
			exit(-1);
			break;
		}

          argv +=2;
          argc -=2;
	 }

	// Open the global output file

	gs.g_outf = fopen(gn.global_outfile, "w");
	if (gs.g_outf == NULL)
	{
		printf("Opening output file %s failed, aborting.\n",
			gn.global_outfile);
		exit(1);
	}

	// Read in configuration parameters to set/modify most
	// of the operating behavior(s)
	
	parse_config_file(&gn, &gs);

	// Minor hack for parsing FreeBSD 'ping' output. If this
	// was set in the config file, no other processing is
	// performed.
	
	if (FLAG_SET(gs.ping_stats, PROCESS_NIXPING))
	{
		process_unix_ping(&gn, &gs);
		exit(0);
	}

	// Preload the DNS ipaddr:name mappings from external files
	// (only occurs if appropriate filenames were supplied in the
	// config file). Note that client detection requires a substantial
	// set of resolved (or resolvable) domain names.

	init_dns_lists(&gn, &gs);

	// Preload the RADB ipaddr->ASnumber mappings from external
	// files (if appropriate filenames were supplied in the
	// config file)

	init_asnum_list(&gn, &gs);

	// Before embarking on the (usually time consuming) walk through
	// the logfile(s), perform a final sanity check on the
	// configuration parameters

	if (sanity_check(&gs) == 1)
		exit(-1);

	// Read/process the logfile(s) and aggregate the game information

	{
		long sec1 = time(NULL);
		long sec2;
		walk_log_files(&gn, &gs);
		sec2 = time(NULL);
		printf("Walking logfile(s) took %d seconds\n", sec2 - sec1);
		fprintf(gs.g_outf, "Walking logfile(s) took %d seconds\n", sec2 - sec1);
		fflush(NULL);
	}
	
	// Dump the entire internal table of ipaddr:domainname mappings
	// (primarily used to generate a new local mapping file after
	// reading - and resolving - a logfile containing new clients)

	if (FLAG_SET(gs.dom_stats, DUMP_DOMAINMAP))
		list_all_domain_names(&gn, &gs);

	// If required, resolved AS numbers for all ip addresses
	// seen.
	//
	// [gja NOTE: 081701: Need to update resolve_asnums() so it uses
	// seen_client list rather than the old, and now removed,
	// seen_playername list]

//	if (FLAG_SET(gs.dom_stats, RESOLVE_ASNUM))
//		resolve_asnums(&gn, &gs);

	// Dump the entire internal table of ipaddr->ASnum mappings

	if (FLAG_SET(gs.dom_stats, DUMP_ASNUMMAP))
		list_all_asnums(&gs);

	// Dump a sorted list of AS numbers, sorted in descending order of
	// how frequently the AS number appears in the seen_playername list
	//
	// [gja NOTE: 081701: Need to update dump_asnum_tree() so it uses
	// seen_client list rather than the old, and now removed,
	// seen_playername list]
	
//	if (FLAG_SET(gs.dom_stats, DUMP_ASTREE))
//		dump_asnum_tree(&gs);

	// If required, analyse the number of players, clients,
	// and how many look like NAT users

	if (FLAG_SET(gs.dom_stats, SUFFIX_STATS))
		create_pname_suffix_stats(&gs);

	// If command token stats are required, print them out
	// (list of how many times "ClientBegin", etc, were
	// seen)

	if (FLAG_SET(gs.stats, C_TOKEN_STATS))
		print_token_stats(&gs, gs.cmdlist, gs.stats);

	// If the list of derived CLIENTS is required, print them
	// out. (Dumped in descending order of number of games
	// played.)

	if (FLAG_SET(gs.stats, CLIENT_NAMES))
		print_client_list(&gs, &gn);

	// If required, create histogram of median ping values
	
	if (FLAG_SET(gs.graph_stats, GRAPH_PINGS))
		create_ping_histo(&gs, &gn);

	// If required, create histogram(s) of playedtime vs day
	// of week, or vs hour of day
	
	if (FLAG_SET(gs.graph_stats, PTIME_HISTO))
		create_ptime_histo(&gs, &gn);
		
	// Dump the cumulative time (in hours) played by non-bot
	// players versus calendar time.

	if (FLAG_SET(gs.stats, CUMULATE_TIME))
		cumulative_gametime(&gs);

	// Print some per-game stats (start/stop, player kills, etc)

	if (FLAG_SET(gs.stats, GAME_STATS))
		print_game_stats(&gs);

	// Generate ping information for a single specified
	// client, using "ClientPing:" logfile entries (internal
	// mode I used prior to gmod1.0 being released, superseded
	// by "CPhisto2:" logged compressed histograms from gmod1.0)
	
	if (FLAG_SET(gs.ping_stats, SINGLE_C_PING))
		single_client_pings(&gs);

	// Generate ping information for a single specified
	// client, using "CPhisto2:" logfile entries from gmod1.0
	
	if (FLAG_SET(gs.ping_stats, SINGLE_C_PHISTO))
		single_client_phisto(&gs);

	if (FLAG_SET(gs.ping_stats, TOPN_PHISTOS))
		topN_client_phistos(&gs);

	//
	if (FLAG_SET(gs.ping_stats, TOPN_CLIENT))
		topN_client_pings(&gs);

	//
	if (FLAG_SET(gs.ping_stats, MEDIAN_PINGS))
		median_client_pings(&gs);

	//
	if (FLAG_SET(gs.ping_stats, PINGVSFRAGS))
		player_ping_vs_frags(&gs, &gn);

	//
	if (FLAG_SET(gs.ping_stats, PINGVSGRATIO))
		player_ping_vs_gameratio(&gs, &gn);

	//
	if (FLAG_SET(gs.ping_stats, PINGVSGAMENUM))
		client_ping_vs_gamenum(&gs);

	//
	if (FLAG_SET(gs.ping_stats, PINGVSGAMETIME))
		client_ping_vs_gametime(&gs);

	// CHECK_FRAG_COUNTS(&gs);

	return(0);
}


//
// init_gmodstat_state()
//
// Initialize the gs and gn structures
//
void init_gmodstat_state(struct global_state *gs,
					struct global_names *gn)
{
	memset((char *)gn, 0, sizeof(struct global_names));
	memset((char *)gs, 0, sizeof(struct global_state));

	strcpy(gn->configfile, CONFDEFAULT);		// Default configuration file
	strcpy(gn->logfile, LOGDEFAULT);			// Default logfile file
	strcpy(gn->global_outfile, OUTFDEFAULT);	// Default output file suffix
	strcpy(gn->whoisaddr, WHOIS_ADDR);		// IP address of RADB whois server (in ASCII)
	gs->maptime = -1;
	gs->saycid = -1;		// Suppress collecting and printing 'say' lines if this parameter is -1
	gs->rangeinc = 240;
	gs->faked_suffix = 4;	// Default, selects ".w-x-y-z.unresolved" as the fake suffix
							// for seen_client entries.
	gs->max_ping = 998;	// Default ping limit is 998 ms (999 is infinity in Q3)
	gs->ping_step = 1;	// Default ping histogram bucket sizes
	gs->jitterwidth = (float)0.90;	// Default, jitter is lower/upper ping bounds encompassing
									// 90% of all ping samples
	gs->gmod_ver = (float)0.0;

	return;
}

//
// walk_log_files()
//
// Opens one or more log files and repeatedly calls
// the read_log_file() routine.
//
// Returns the number of seconds total elapsed game time
// covered by the logfiles
//

long walk_log_files(struct global_names *gn,
					struct global_state *gs)
{
	FILE	*infile;		/* Stream descriptor for input file. */
	long gametime;		// Aggregate gametime in seconds

	// The input file is either LOGDEFAULT, a single file set by the user
	// on the command line, or a pointer to a file containing a set of
	// file names to be processed in sequence. If logfile2 is non-empty
	// then we ignore logfile.

	gametime = 0;

	if (strlen(gn->logfile2) > 0)
	{
		// If a file of logfiles was provided, open it and begin reading
		// filenames, aggregating game stats from each logfile

		FILE	*filelist;		/* Stream descriptor for file of input files. */
		int listeof = 0;

		fprintf(gs->g_outf, "Retrieving logfiles from %s\n", gn->logfile2);

		if ((filelist = fopen(gn->logfile2, "rb")) == NULL)
			{
			 fprintf(gs->g_outf, "Opening %s failed: %s\n", gn->logfile2, strerror(errno));
			 exit(1);
			}

		while (listeof != 1)
		{
    		memset(gn->logfile, 0, 128);
    		if (fgets(gn->logfile, 128, filelist) == NULL)
    			 listeof = 1;
    		else
    		{
    		 // First, ensure there's a null character at the end of
    		 // the filename string (to account for conversions between
    		 // windows and unix from time to time)
    		
    		  char *tmpchr = strchr(gn->logfile, '\n');
    		  if (tmpchr != NULL)
    			  *tmpchr = 0;

    		  tmpchr = strchr(gn->logfile, '\r');
    		  if (tmpchr != NULL)
    				 *tmpchr = 0;

    		  // Now, if the line was a real name rather than a comment
    		  // we open the file (using gn->logfile2_base to create the
    		  // full path name) and process the specified logfile
    		
    		  if ((gn->logfile[0] != '#') && (strlen(gn->logfile) > 0))
    		  {
    		  	  char tmp_logfile[128];
    		  	
    		  	  sprintf(tmp_logfile, "%s%s\0", gn->logfile2_base, gn->logfile);
    		  	
    			  fprintf(gs->g_outf, "Reading from %s\n", tmp_logfile);
    			  printf("Reading from %s\n", tmp_logfile);

    			  // Open in binary mode to cope with special codes in certain playernames
    			  if ((infile = fopen(tmp_logfile, "rb")) == NULL)
    				 fprintf(gs->g_outf, "Opening %s failed: %s\n", tmp_logfile, strerror(errno));
    			  else
    				  gametime += read_log_file(infile, gn->logfile, gs);
    		  }
    		  else
    			  fprintf(gs->g_outf, "Skipping %s\n", gn->logfile);

    		 }
		}
	}
	else
	{
	 fprintf(gs->g_outf, "Reading from %s\n", gn->logfile);
	 printf("Reading from %s\n", gn->logfile);

	 // Open in binary mode to cope with special codes in certain playernames
	 if ((infile = fopen(gn->logfile, "rb")) == NULL)
		{
		 fprintf(gs->g_outf, "Opening %s failed: %s\n", gn->logfile, strerror(errno));
		 exit(1);
		}

	 gametime += read_log_file(infile, gn->logfile, gs);
	}

	fprintf(gs->g_outf, "\nTotal parsed gametime covers ");
	fprintf(gs->g_outf, "%d days, %d hours, %d minutes, %d seconds (total %d seconds)\n\n",
			 gametime / 86400, (gametime / 3600) % 24,
			 (gametime / 60) % 60, gametime % 60, gametime);

	printf("\nTotal parsed gametime covers ");
	printf("%d days, %d hours, %d minutes, %d seconds (total %d seconds)\n\n",
			 gametime / 86400, (gametime / 3600) % 24,
			 (gametime / 60) % 60, gametime % 60, gametime);

	return(gametime);
}




//
// read_log_file()
//
// Read through the opened log file.
//
// Returns the number of seconds total elapsed game time
// covered by the logfile
//


int read_log_file(FILE *inf, char *logfilename,
				  struct global_state *gs)
{
	int		pkteof, flag;
	int		tstamp = 0;
	char *filetail;			// Points to last component of filename (in case
							// the name includes "blah\blah\filename.log" form)

	gs->time_base = 0;	// Initialize time base to an invalid value (Jan.1st, 1970)

	// Now, if BASE_TIME is set we try to extract a valid time_base
	// from the format of the logfile's ASCII filename. (If not set,
	// we might still pick up a valid time_base from a BaseTime: line
	// in the logfile itself.)

#ifdef WIN32
	filetail = strrchr(logfilename, '\\');
#endif

#ifndef WIN32
	filetail = strrchr(logfilename, '/');
#endif

	if (filetail == NULL)
		filetail = logfilename;
	else
		filetail++;

	// If BASE_TIME is set, we assume that the start time of
	// this logfile is embedded in the logfile name. Extract
	// it and set the base timestamp (in seconds since 1/1/70)
	if (FLAG_SET(gs->stats, BASE_TIME))
		if (extract_base_time(filetail, gs) == -1)
		{
			fprintf(gs->g_outf, "read_log_file: Error, couldn't extract base time from %s\n",
				filetail);
		}

	pkteof = 1;
	while (pkteof != 0)
	{
	 if (feof(inf))
	 {	
		 fprintf(gs->g_outf, "Reached end of log file after ");
		 fprintf(gs->g_outf, "%d days, %d hours, %d minutes, %d seconds (total %d seconds)\n\n",
			 tstamp / 86400, (tstamp / 3600) % 24, (tstamp / 60) % 60, tstamp % 60, tstamp);
		 pkteof = 0;
	 }
	 else
	 {
		// Parse single line
		flag = parse_line2(inf, &tstamp, gs);
		if (flag == -1)
		{
		 fprintf(gs->g_outf, "Parsing process terminated.\n");
		 pkteof = 0;
		}
		if (flag == -2)
		 fprintf(gs->g_outf, "Parsing process non-fatal error, skipping line.\n");
	 }
	}
	
	return(tstamp);
}


// Takes a filename of form xxxx-ddmmyy-hhhh-pppp
// and extracts the ddmmyy and hhhhh into a single
// 'long' number of seconds since midnight jan.1st, 1970
// then stores this into gs->timebase
//
// (this is a pre-gmod1.0 hack, nothing to worry about)

int extract_base_time(char *name, struct global_state *gs)
{
	int day, month, year, hours;
	char temp[32], datestr[32];
	struct tm tmnow;

	if (name == NULL)
		return(-1);

	if (strlen(name) == 0)
		return(-1);

	sscanf(name, "%[^-]-%[^-]-%d-", temp, datestr, &hours);

	sprintf(temp, "%c%c\0", datestr[0], datestr[1]);
	sscanf(temp, "%d", &day);
	sprintf(temp, "%c%c\0", datestr[2], datestr[3]);
	sscanf(temp, "%d", &month);
	sprintf(temp, "%c%c\0", datestr[4], datestr[5]);
	sscanf(temp, "%d", &year);

	if (year < 70)
		year += 100;

	memset((char *)&tmnow, 0, sizeof(struct tm));
	tmnow.tm_isdst = -1; // Force system to decide if daylight savings was on
	tmnow.tm_hour = (hours / 100) + gs->time_offset;
	tmnow.tm_min = hours % 100;
	tmnow.tm_year = year;
	tmnow.tm_mon = month - 1;	// NB. In tm_mon january is month "0" not "1"
	tmnow.tm_mday = day;

	gs->time_base = mktime(&tmnow);

	printf("Setting basetime to %d - %s", gs->time_base, asctime(&tmnow));
	fprintf(gs->g_outf, "Setting basetime to %d - %s", gs->time_base, asctime(&tmnow));

	if (gs->time_offset != 0)
	{
		printf("(Note: offset of %d hours added to base time)\n",
			gs->time_offset);
		fprintf(gs->g_outf, "(Note: offset of %d hours added to base time)\n",
			gs->time_offset);
	}

	return(0);
}

// setbasetime()
//
// Called when a "BaseTime:" token is encountered in the logfile.
// (gmod1.0 uses this to embed the server's starttime in the
// logfile itself.)

void setbasetime(char *ptr, int tstamp, struct global_state *gs)
{
	char tmptime[32];

	sprintf(tmptime, "xxx0-%s\0", ptr);
	extract_base_time(tmptime, gs);
	return;
}



// setgmodversion()
//
// Called when a "ModVersion:" token is encountered in the logfile.
// This entry is inserted by the gmod server modification that
// creates CPhisto2 lines, etc. Doesn't appear in logfiles pre-0901.

void setgmodversion(char *ptr, int tstamp, struct global_state *gs)
{
	float tmpver;
	
	gs->gmod_ver = (float)0.0;
	
	if (sscanf(ptr, "gja%f", &tmpver) < 1)
	{
		printf("Warning, ModVersion %s not recognized.\n", ptr);
		fprintf(gs->g_outf, "Warning, ModVersion %s not recognized.\n", ptr);
	}
	else
		gs->gmod_ver = tmpver;
	
	return;
}




//
// findclient_id()
//
// Search a client_id list for a matching entry to
// the supplied client ID number, return pointer to
// matched entry (or NULL if none found)
//
struct client_id *findclient_id(struct client_id *cidptr, int cidnum)
{
	while (cidptr != NULL)
	{
		if (cidptr->c_id == cidnum)
			return(cidptr);

		cidptr = cidptr->next;
	}

	return(NULL);
}


//  update_kill_list()
//
void update_kill_list(struct player_kill **klist,
					  struct seen_player *who,
					  int wp,
					  struct global_state *gs)
{
	struct player_kill *ktmp;

	// Add or update entry for a kill/death involving "who" using weapon "wp"

	// Deal with the case where this is the first kill/death entry

	if (*klist == NULL)
	{
		*klist = (struct player_kill *)malloc(sizeof(struct player_kill));
		if (*klist == NULL)
		{
			fprintf(gs->g_outf,"Error in update_kill_list - memory allocation failed.\n");
			exit(-1);
		}
		memset((char *)(*klist), 0, sizeof(struct player_kill));

		(*klist)->count = 1;
		(*klist)->how = wp;
		(*klist)->who = who;

		return;;
	}

	// Walk through the existing list to see if this who/weapon combination
	// has happened before

	ktmp = *klist;
	while (ktmp != NULL)
	{
		if ((ktmp->how == wp) && (ktmp->who == who))
			break;
		ktmp = ktmp->next;
	}

	if (ktmp == NULL)
	{
		// Need to create a new entry

		ktmp = (struct player_kill *)malloc(sizeof(struct player_kill));
		if (ktmp == NULL)
		{
			fprintf(gs->g_outf,"Error in update_kill_list - memory allocation failed.\n");
			exit(-1);
		}
		memset((char *)(ktmp), 0, sizeof(struct player_kill));

		ktmp->count = 1;
		ktmp->how = wp;
		ktmp->who = who;

		ktmp->next = *klist;
		*klist = ktmp;

		return;
	}

	// Here 'ktmp' points to a matching entry, just increment counter

	ktmp->count++;

	return;
}



//  update_killedbyworld()
//
void update_killedbyworld(struct seen_player *seenp,
					  int wp,
					  struct global_state *gs)
{
	struct player_kill *ktmp;

	// Add or update entry for a death by <world> (1022) using weapon "wp"

	// First deal with the case where this is the first death entry

	if (seenp->death == NULL)
	{
		seenp->death = (struct player_kill *)malloc(sizeof(struct player_kill));
		if (seenp->death == NULL)
		{
			fprintf(gs->g_outf,"update_killedbyworld: malloc failed, exiting.\n");
			exit(-1);
		}
		memset((char *)(seenp->death), 0, sizeof(struct player_kill));

		(seenp->death)->count = 1;
		(seenp->death)->how = wp;
		(seenp->death)->who = NULL;

		return;
	}

	// Walk through the existing list to see if this <world>/weapon combination
	// has happened before. Note that in the death list, <world> is represented
	// by a NULL pointer in the ->who field.

	ktmp = seenp->death;
	while (ktmp != NULL)
	{
		if ((ktmp->how == wp) && (ktmp->who == NULL))
			break;
		ktmp = ktmp->next;
	}

	if (ktmp == NULL)
	{
		// Need to create a new entry

		ktmp = (struct player_kill *)malloc(sizeof(struct player_kill));
		if (ktmp == NULL)
		{
			fprintf(gs->g_outf,"Error in update_kill_list - memory allocation failed.\n");
			exit(-1);
		}
		memset((char *)(ktmp), 0, sizeof(struct player_kill));

		ktmp->count = 1;
		ktmp->how = wp;
		ktmp->who = NULL;

		ktmp->next = seenp->death;
		seenp->death = ktmp;

		return;
	}

	// Here 'ktmp' points to a matching entry, just increment counter

	ktmp->count++;

	return;
}

// player_in(tmpplay, ts)
//
// Initialize/Update the player inout time list
//

void player_in(struct seen_player *tp,
				int ts,
				int playmode,
				struct global_state *gs)
{
	struct player_inout *tmp1;


	tmp1 = (struct player_inout *)malloc(sizeof(struct player_inout));
	if (tmp1 == NULL)
	{
		// bad scene, we've got some wierd memory failure going on
		fprintf(gs->g_outf, "player_in: malloc failed, exiting.\n");
		exit(-1);
	}

	// Clear the new inout entry.
	memset((char *)tmp1, 0, sizeof(struct player_inout));

	// Set the in timestamp
	tmp1->in_ts = ts;
	tmp1->out_ts = -1;
	tmp1->playmode = playmode;

	// Place new entry at head of list

	tmp1->next = tp->inout;
	tp->inout = tmp1;

	return;
}


//
// player_out(tmpplay, ts)
//
// Updates the player in-out time list.
// Increment the seen_player entry's totaltime count
// ONLY if the player was NOT in spectate mode
//

void player_out(struct seen_player *tp,
				int ts)
{
	if (tp->inout == NULL)	// Player never reached the BEGIN state before leaving
		return;

	if ((tp->inout)->out_ts != -1) // Player is already listed as disconnected
		return;

	// Update timestamp of entry at head of list

	(tp->inout)->out_ts = ts;

	// Increment the total time this player has been in this particular game
	// by the length of this inout period, provided that this inout period
	// wasn't dedicated to spectating.

	if (tp->inout->playmode != 3)
		tp->totaltime += (ts - (tp->inout)->in_ts);

	return;
}


// c_add_ipaddr()
//
// Takes pointer to a seen_player entry, and adds the supplied
// ipaddress and port information to the list of ipaddr:port
// associated with this player during this game.

void c_add_ipaddr(struct seen_player *seenp,
				  struct in_addr *iaddr,
				  int port,
				  struct global_state *gs)
{

	struct cid_ipaddr *tmp_c_ipaddr;

	tmp_c_ipaddr = (struct cid_ipaddr *)malloc(sizeof(struct cid_ipaddr));
	if (tmp_c_ipaddr == NULL)
	{
		fprintf(gs->g_outf, "c_add_ipaddr: malloc() failed. Exiting.\n");
		exit(-1);
	}

	memset((char *)tmp_c_ipaddr, 0, sizeof(struct cid_ipaddr));
	memcpy(&(tmp_c_ipaddr->iaddr), iaddr, sizeof(struct in_addr));
	tmp_c_ipaddr->port = port;
	tmp_c_ipaddr->next = seenp->c_addr;
	seenp->c_addr = tmp_c_ipaddr;

	return;
}






// CHECK_FRAG_COUNTS()
//
// Walk through the seen_game list, and for each game and
// each player who was present at the end of each game,
// compare my calculated frag counts against what the server
// reported.
//
// This routine is just for development purposes, to verify
// my on-the-fly frag counting algorithm. It can also notice
// the few occasions where the server's logged 'kill' entries
// don't really reflect reality (sometimes players commit
// suicide before the rejoin right at the start of games,
// and the death is logged as though against client 0 but not counted
// against client 0). Such errors are rare though, and usually
// account for a mismatch of 1 frag count over many games.
//

void CHECK_FRAG_COUNTS(struct global_state *gs)
{
	struct  seen_games *glist;
	struct	seen_player *plist;
	struct player_inout *iolist;
	struct player_kill *pk;
	int frags;

	// Games are stored last->first hanging off glist. Walk the
	// list until we find the last entry, then step backwards up the
	// list, printing stats as we go

	if ((glist = gs->gamelist) == NULL)
		return;

	while (glist->next != NULL)
		glist = glist->next;

	// glist points to entry at end of list, chronologically
	// the first one parsed from the logfile

	printf("Checking frag counts\n");

	fprintf(gs->g_outf, "Checking frag counts\n");

	// Walk the seen_game list backwards (forwards chronologically)

	while (glist != NULL)
	{
		long starttime = glist->t_base + glist->t_start;
		char *timestr1 = ctime(&starttime);
		char *timestr2;
		int newline = 0;

		timestr2 = strchr(timestr1, '\n');
		if (timestr2 != NULL)
			*timestr2 = 0;


		// Step through all the players seen in this game
		plist = glist->c_all;
		while (plist != NULL)
		{
			if (plist->frags > 0)
			{
				// Step through the inout entries for this player,
				// summing the interim frag counts

				iolist = plist->inout;
				frags = 0;
				while (iolist != NULL)
				{
					frags += iolist->frags;
					iolist = iolist->next;
				}

				if (plist->inout->frags != plist->frags)
				{
					if (newline == 0)
						fprintf(gs->g_outf, "%s %s ", timestr1, glist->mapname);
					newline = 1;
					fprintf(gs->g_outf, ": %s mismatch. Seen %d, saw %d frags, server says %d\n",
						plist->playername, plist->numseen, frags, plist->frags);

					 // List the in/out times
					 iolist = plist->inout;
					 while (iolist != NULL)
					 {
						 fprintf(gs->g_outf, "    Joined %d:%02d Left %d:%02d Frags %d (ROUGHLY)\n",
							 iolist->in_ts / 60, iolist->in_ts % 60,
							 iolist->out_ts / 60, iolist->out_ts % 60,
							 iolist->frags);
						 iolist = iolist->next;
					 }
					 // List the kills

					 pk = plist->kill;
					 while (pk != NULL)
					 {
						 fprintf(gs->g_outf, "     Killed %s  %d times with %s\n",
							 pk->who, pk->count, weapname[pk->how]);
						 pk = pk->next;
					 }
					 // List the deaths
	
					 pk = plist->death;
					 while (pk != NULL)
					 {
						 if (pk->who != NULL)
							 fprintf(gs->g_outf, "     Killed by %s  %d times with %s\n",
								pk->who, pk->count, weapname[pk->how]);
						 else
							 fprintf(gs->g_outf, "     Killed by <WORLD>  %d times with %s\n",
								pk->count, weapname[pk->how]);
						 pk = pk->next;
					 }
				}
			}
			plist = plist->next;
		}
		glist = glist->prev;
	}

	return;
}



// process_unix_pings()
//
// Read in a file containing on each line the output of
// a unix ping command (in this case FreeBSD) and plot
// a histogram of the ping values.
//
// NOTE: This routine is a complete hack so I could post-process
// the ASCII output of long-running unix ping command (being
// used to characterize the ping times between two points
// as reported by 'ping')  Just interested to see how 'ping'
// compares to what the QuakeIII server is logging.

void process_unix_ping(struct global_names *gn,
						struct global_state *gs)

{
	FILE *infile;
	char oneline[MAXLINE];
	int pingvalue;
	char *ptr;
	int arrayptr = 0;
	int arraylimit = gs->max_ping / gs->ping_step;
	int pingcntr;
	int arraysize;

	printf("Processing a unix ping file - all other processing ignored.\n");
	fprintf(gs->g_outf, "Processing a unix ping file - all other processing ignored.\n");

	// Open the unix ping file

	if ((infile = fopen(gn->unixpingfile, "rb")) == NULL)
		{
		 fprintf(gs->g_outf, "Opening %s failed: %s\n",
			 gn->unixpingfile, strerror(errno));
		 exit(1);
		}

	// We're doing a histogram of ping times, so allocate space for the
	// array of 'float'

	gs->ping_histo = NULL;
	arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);
	gs->ping_histo = (float *)malloc(arraysize);
	if (gs->ping_histo == NULL)
	{
		printf("process_unix_pings: malloc failed, exiting\n");
		fprintf(gs->g_outf, "process_unix_pings: malloc failed, exiting\n");
		exit(-1);
	}

	memset((char *)gs->ping_histo, 0, arraysize);
	gs->ping_total = 0;
	pingcntr = 0;

	while (fgets(oneline, MAXLINE, infile) != NULL)
	{
		// Step back from end of the line, looking for "time=XXX ms"

		ptr = oneline + strlen(oneline) - 1;
		while (ptr > oneline)
		{
			if (strncmp(ptr, "time=\0", 5) == 0)
				break;
			ptr--;
		}

		if (ptr == oneline)
			fprintf(gs->g_outf, "Error - couldn't find time parameter!\n");


		// ptr now points to the text "time=XXX.YY ms" where
		// XXX.YY is a floating point value representing the ping
		// value in ms. Extract XXX as an integer.

		sscanf(ptr, "time=%d", &pingvalue);
		fprintf(gs->g_outf, "%d %d\n", pingcntr, pingvalue);
		pingcntr ++;

		add_ping_to_histo(pingvalue, 1.0, gs);
	}


	// Print out the histogram contents

	fprintf(gs->g_outf, "\nUnix Ping Histogram\n");
	pingcntr = 0;
	while (arrayptr < arraylimit)
	{
		fprintf(gs->g_outf, "%d %7.2f\n", pingcntr,
			(100.0*(gs->ping_histo)[arrayptr])/gs->ping_total);

		arrayptr++;
		pingcntr += gs->ping_step;
	}

	// Clean up the array space
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

	return;
}