// gmodstat_asnum.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Routines for performing lookups of the collected
// IP addresses and mapping to AS numbers as report
// by 'whois' on the RADB
//

// ++++ NOTE: As of gmodstat 0.1, the code here is incomplete.
// ++++ Yes, I'm re-working some earlier AS number lookup code
// ++++ that ended up being out of sync with other changes I made
// ++++ to gmodstat's internal structure. Will fold this all back
// ++++ in a later release of gmodstat.

#include "stdafx.h"
#include "gmodstat.h"


// push_new_asnum()
//
// Takes information about a new ipaddr->ASnum mapping
// and creates/updates appropriate gs->asnumbers and
// gs->dnslist entries
//
// Each gs->dnslist entry contains an IP address and
// optionally a domain name. It also contains a pointer
// to a gs->asnumbers entry containing an AS number,
// prefix, prefix length, and ASCII description field.
// There may be multiple gs->asnumbers entries for the
// same AS number if there were different prefix/prefixlen
// combinations associated with that AS.

void push_new_asnum(struct global_state *gs,
					unsigned int tmpaddr,
					int asnum,
					unsigned int prefix,
					int prefix_len,
					int multirec,
					int multiasnum,
					char *descr)
{
	struct hostaddrlist *tmphlist;
	struct asnumlist *tmpaslist;
	struct in_addr dummyaddr;

	memset((char *)(&dummyaddr), 0, sizeof(struct in_addr));
	dummyaddr.s_addr = tmpaddr;

	// Scan the existing gs->dnslist entries to see if this
	// ipaddr has already been seen before

	tmphlist = gs->dnslist;
	while (tmphlist != NULL)
	{
		if (tmphlist->ipaddr == tmpaddr)
			break;
		tmphlist = tmphlist->next;
	}

	// Scan existing AS number entries, see if this one's been
	// seen before (and flag if the ASCII description differs)

	tmpaslist = gs->asnumbers;
	while (tmpaslist != NULL)
	{
		if ((tmpaslist->asnum == asnum) &&
			(tmpaslist->prefix == prefix) &&
			(tmpaslist->prefix_len == prefix_len))
		{
			if (strcmp(tmpaslist->descr, descr) != 0)
			{
				printf("push_new_asnum: AS %d has multiple descriptions\n",
					asnum);
				printf("  Current: %s\n", tmpaslist->descr);
				printf("  New:     %s\n", descr);
				fprintf(gs->g_outf, "push_new_asnum: AS %d has multiple descriptions\n",
					asnum);
				fprintf(gs->g_outf, "  Current: %s\n", tmpaslist->descr);
				fprintf(gs->g_outf, "  New:     %s\n", descr);

			}
			break;
		}
		tmpaslist = tmpaslist->next;
	}

	// If the AS number hasn't been seen before, first we create
	// a new entry in the gs->asnumbers list

	if (tmpaslist == NULL)
	{
		if ((tmpaslist =
			 (struct asnumlist *)malloc(sizeof(struct asnumlist)))
			 == NULL)
		{
		  fprintf(gs->g_outf, "Memory allocation failure in push_new_asnum(), quitting.\n");
		  exit(-1);
		}

		memset((char *)tmpaslist, 0, sizeof(struct asnumlist));

		if ((tmpaslist->descr =
			 (char *)malloc(strlen(descr)+1))
			 == NULL)
		{
		  fprintf(gs->g_outf, "Memory allocation failure in push_new_asnum(), quitting.\n");
		  exit(-1);
		}
			
		tmpaslist->asnum = asnum;
		strcpy(tmpaslist->descr, descr);
		tmpaslist->next = gs->asnumbers;
		tmpaslist->prefix = prefix;
		tmpaslist->prefix_len = prefix_len;
		gs->asnumbers = tmpaslist;
		gs->asnum_cnt++;
	}


	// If there is no matching hostaddrlist for the ipaddr, create
	// a new one. Fill it with a pointer to the associated AS number
	// entry too.

	// If there *was* a matching hostaddrlist, update its AS number
	// information.

	if (tmphlist == NULL)
	{
		if ((tmphlist =
			 (struct hostaddrlist *)malloc(sizeof(struct hostaddrlist)))
			 == NULL)
		{
		  fprintf(gs->g_outf, "Memory allocation failure in push_new_asnum(), quitting.\n");
		  exit(-1);
		}

		memset((char *)tmphlist, 0, sizeof(struct hostaddrlist));
		tmphlist->name = NULL;		// Since we don't know the domain name, leave it null
		tmphlist->ipaddr = tmpaddr;
		tmphlist->asnum = tmpaslist;

		// Put this new entry at head of hlist

		tmphlist->next = gs->dnslist;
		gs->dnslist = tmphlist;

	}
	else
	{

		// If there's already a matching hostaddrlist entry, sanity
		// check that the asnum, prefix, and prefix_len are the same
		// (or update them if they haven't yet been initialized)

		if (tmphlist->asnum == NULL)
		{
			tmphlist->asnum	= tmpaslist;
			tmphlist->multirec = multirec;
			tmphlist->multiasnum = multiasnum;
		}
		else
		{
			if (tmphlist->asnum->asnum != asnum)
				  fprintf(gs->g_outf, "push_new_asnum: For %s, AS %d and AS %d don't match\n",
						inet_ntoa(dummyaddr), asnum, tmphlist->asnum->asnum);

			if (tmphlist->asnum->prefix != prefix)
				  fprintf(gs->g_outf, "push_new_asnum: For %s, prefix %d and %d don't match\n",
						inet_ntoa(dummyaddr), prefix, tmphlist->asnum->prefix);

			if (tmphlist->asnum->prefix_len != prefix_len)
				  fprintf(gs->g_outf, "push_new_asnum: For %s, prefix_len %d and %d don't match\n",
						inet_ntoa(dummyaddr), prefix_len, tmphlist->asnum->prefix_len);

		}
	}

}



//
// list_all_asnums()
//
// Dump the contents of hlist, only ONE entry per
// domains name/ipaddr pair, listing only the ASnumber
// related information.
//

void list_all_asnums(struct global_state *gs)
{
	struct hostaddrlist *tmphlist;
	struct asnumlist *tmpaslist;
	struct in_addr ta, pf;

	memset((char *)(&ta), 0, sizeof(struct in_addr));
	memset((char *)(&pf), 0, sizeof(struct in_addr));

	fprintf(gs->g_outf, "Dumping all unique ipaddr and AS numbers:\n");
	fprintf(gs->g_outf, "ipaddr\t\tASnum\tprefix\t\tplen\tdescription\n");

	// Walk through the hlist (where there's only one entry for
	// each ip address) and print each AS number info to the output file.

	tmphlist = gs->dnslist;
	while (tmphlist != NULL)
	{


		if (tmphlist->asnum != NULL)
		{
			ta.s_addr = tmphlist->ipaddr;

			pf.s_addr = tmphlist->asnum->prefix;

			fprintf(gs->g_outf, "%s\t%d",
				inet_ntoa(ta), tmphlist->asnum->asnum);
			fprintf(gs->g_outf, "\t%s\t%02d %d %d %s\n",
				inet_ntoa(pf), tmphlist->asnum->prefix_len,
				tmphlist->multirec, tmphlist->multiasnum,
				tmphlist->asnum->descr);
		}

		tmphlist = tmphlist->next;
	}

	fprintf(gs->g_outf, "\n");

	fprintf(gs->g_outf, "Dumping all unique AS number/prefix entries:\n");
	fprintf(gs->g_outf, "ASnum\tprefix\t\tplen\tdescription\n");

	// Walk through the hlist (where there's only one entry for
	// each ip address) and print each AS number info to the output file.

	tmpaslist = gs->asnumbers;
	while (tmpaslist != NULL)
	{
		pf.s_addr = tmpaslist->prefix;

		fprintf(gs->g_outf, "%d\t%s\t%d\t%s\n",
			tmpaslist->asnum, inet_ntoa(pf), tmpaslist->prefix_len,
			tmpaslist->descr);

		tmpaslist = tmpaslist->next;
	}

	fprintf(gs->g_outf, "\n");
	return;
}









