// gmodstat_clientlist.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Routines for manipulating, dumping, and analysing
// the seen_client list.
//
 
#include "stdafx.h"
#include "gmodstat.h"



// print_client_list()
//
// Sorts the list of seen_client entries in descending order
// of how many minutes the client appeared across
// all the log files. Then dumps the list to an output file,
// showing various stats (optionally including the addresses
// from which the playername appeared, and the common suffixes)
//

void print_client_list(struct global_state *gs,
						struct global_names *gn)

{	
	struct seen_client *tmpclist;
	char tempname[128];
	FILE *client_outf;
	
	sprintf(tempname, "clients-%s\0", gn->global_outfile);
	fprintf(gs->g_outf, "Dumping client list to %s\n",
		tempname);
	printf("Dumping client list to %s\n",
		tempname);
	if ((client_outf = fopen(tempname, "w")) == NULL)
	{
		printf("Opening output file %s failed, aborting.\n",
			tempname);
		fprintf(gs->g_outf, "Opening output file %s failed, aborting.\n",
			tempname);
		exit(-1);
	}	

	// Sort in descending order of played time

	sort_seen_client(gs);

	// Dump the sorted list

	tmpclist = gs->clientlist;

	while (tmpclist != NULL)
	{
		struct player_ipaddr *tmpip;
		if ((tmpclist->unnamedplayer == 0) || FLAG_SET(gs->stats, UNNAMED_PLAYER))
		{

			if (tmpclist->numseen > 1)
				fprintf(client_outf, "%03d games ",
					tmpclist->numseen);
			else
				fprintf(client_outf, "%03d game  ",
					tmpclist->numseen);

			fprintf(client_outf, "%04d min (%05.2f min/game) ",
				tmpclist->totaltime / 60,
				(float)(tmpclist->totaltime / 60)/(float)(tmpclist->numseen));

			if (tmpclist->numip > 1)
				fprintf(client_outf, "%03d addr/port pairs ",
					tmpclist->numip);
			else
				fprintf(client_outf, "%03d addr/port pair  ",
					tmpclist->numip);

			fprintf(client_outf, "%s\n", tmpclist->pname);


			// Print the ipaddr and domain names if available
			// (only for playernames with >= 2 associated ip addresses
			//  and only if there's a local DNS lookup table pointed to by hlist)

			if ((tmpclist->numip > 1) && (gs->dnslist != NULL))
			{
				tmpip = tmpclist->iplist;
				while (tmpip != NULL)
				{
					fprintf(client_outf, "\tAddr:port %s:%d\n",
						inet_ntoa(tmpip->iaddr), tmpip->port);
					tmpip = tmpip->next;
				}
			}

			// Print the common suffixes determined for this playername,
			fprintf(client_outf, "\t\t%s\n\n", tmpclist->suffix);

		}
		// Loop to next player name

		tmpclist = tmpclist->next;
	}

	fclose(client_outf);
	
	return;

}

// sort_seen_client()
//
// Dumb bubble sort on the seen_client list, in descending
// order of total played time.
//
// (Yeah, yeah... should be some wizzy sort algorithm, but
// bubble sort was simple and works just fine)
//
void sort_seen_client(struct global_state *gs)
{

	struct seen_client *tmpclist, *tmpclist2, *prevplist;
	int swap = 1;

	while (swap == 1)
	{
		tmpclist = gs->clientlist;
		prevplist = NULL;
		swap = 0;		// Made no bubble swaps yet
		
		while (tmpclist != NULL)
		{
		 if (tmpclist->next != NULL)
		 {
			
			 if ((tmpclist->totaltime) < (tmpclist->next->totaltime))
				
			 {
				 // swap
				swap = 1;
				if (prevplist == NULL)
				{
					tmpclist2 = tmpclist->next->next;
					gs->clientlist = tmpclist->next;
					tmpclist->next = tmpclist2;
					(gs->clientlist)->next = tmpclist;
				}
				else
				{
					tmpclist2 = tmpclist->next->next;
					prevplist->next = tmpclist->next;
					tmpclist->next = tmpclist2;
					(prevplist->next)->next = tmpclist;
				}
			 }

			 else if ((tmpclist->totaltime == tmpclist->next->totaltime) &&
				 (tmpclist->numip < tmpclist->next->numip))
			 { // swap
				swap = 1;

				if (prevplist == NULL)
				{
					tmpclist2 = tmpclist->next->next;
					gs->clientlist = tmpclist->next;
					tmpclist->next = tmpclist2;
					(gs->clientlist)->next = tmpclist;
				}
				else
				{
					tmpclist2 = tmpclist->next->next;
					prevplist->next = tmpclist->next;
					tmpclist->next = tmpclist2;
					(prevplist->next)->next = tmpclist;
				}

			 }
		 }
		 prevplist = tmpclist;
		 tmpclist = tmpclist->next;
		}
	}

	return;
}


//
// add_newclients()
//
// Takes a seen_player and updates (or creates) an associated
// seen_client entry in the global client list gs->clientlist
//
// A 'client' is defined as a unique <playername, suffix> tuple.
// This routine looks up the domain name of the first ipaddr in
// this seen_player entry, and searches for any pre-existing seen_client
// entries with matching <playername, domain_suffix>. If found, it
// adds the port numbers from the seen_player entry. If not found,
// a new seen_client entry is created.
//
// As a by-product, the first cid_ipaddr entry hanging off the
// seen_player entry will have its domain name field filled
// in if the ipaddr can be resolved from gs->dnslist (or NULL
// otherwise). Note that while fake suffixes are created for
// seen_client entries, they are not stored in the cid_ipaddr entry.
//
// The per-client pinglist is a copy of the pinglists hanging of
// each per-game seen_player entry.
//
// [This is called after every game during 'Shutdown' event processing]
//

void add_newclients(struct seen_player *seenp,
					struct global_state *gs)
{
	struct seen_client *tmpclient;	// Pointer into gs->clientlist
	struct seen_client *prevclient;
	struct hostaddrlist *tmphlist;	// Pointer into the ipaddr:name map
	struct cid_ipaddr *tmp_cip;
	char *fqdn;						// Pointer to resolved domain name
	char *suffix;					// Pointer to domain name suffix
	char dummyname[64];				// Tmp store for faked suffixes
	unsigned int x;
	int unresolvedflag = 0;
	int cnt;

	// Sanity check first

	if (seenp == NULL) return;			// Wierd
	if (gs == NULL) return;				// Also wierd
	if (seenp->c_addr == NULL) return;	// Slightly wierd, but no problem


	// From mid-Feb 2001 to end of May 2001 over 9000 playernames
	// were seen, many with multiple ipaddr. However, none of them
	// changed ipaddr during a single game. So, the list of
	// ipaddr:port pairs that is hanging off a per-game seen_player
	// entry is usually only one entry, and in the cases where there
	// are multiple entries it is because the port number changed
	// during the game. We'll make use of that assumption by only
	// doing a name resolution on the first entry, but just to be sure
	// we'll check and flag an error if the assumption isn't valid.

	tmp_cip = seenp->c_addr;
	x = (tmp_cip->iaddr).s_addr;
	while (tmp_cip != NULL)
	{
		if ((tmp_cip->iaddr).s_addr != x)
		{
			sprintf(dummyname,"%s\0", inet_ntoa((seenp->c_addr)->iaddr));
			fprintf(gs->g_outf, "add_newclients: Invalid assumption for %s, address %s different to %s\n",
				seenp->playername, dummyname, inet_ntoa(tmp_cip->iaddr));
		}
		tmp_cip = tmp_cip->next;
	}

	// Restore tmp_cip to point to the first entry again
	tmp_cip = seenp->c_addr;

	// Resolve the first address into a domain name, return pointer
	// to the found matching gs_dnslist (hostaddrlist) entry

	tmphlist = resolve_clientaddr(&(tmp_cip->iaddr), gs);

	// Okay, flag that we've attempted a name resolution
	// for this entry (some post-processing may be optimized
	// by not having to re-search gs->dnslist for this ipaddr)	
	// tmp_cip->name will be NULL if the domain
	// name couldn't be resolved from gs->dnslist

	tmp_cip->resolved = 1;	
	tmp_cip->name = NULL;
	tmp_cip->haddr = NULL;
	if (tmphlist != NULL)
	{
		tmp_cip->name = tmphlist->name;
		tmp_cip->haddr = tmphlist;
	}
	else
	{
		// This situation is worth noting, since it suggests the
		// gs->dnslist is out of date (there are ipaddr in the logfiles
		// that have no mapping in the hosts_file and we're not re-resolving
		// from gethostbyaddr() )
		fprintf(gs->g_outf, "add_newclients: gs->dnslist contains no entry for %s\n",
				inet_ntoa(tmp_cip->iaddr));
	}

	fqdn = tmp_cip->name;

	// If the ASCII name is null, create a fake domain suffix
	// based on the digits in the ip address. The fake suffix
	// is ".w-x-y-z.unresolved" for ip address w.x.y.z
	// Otherwise suffix is everything to the right of the
	// first "." found in the domain name.

	if (fqdn == NULL)
	{
		if (gs->faked_suffix == 4)
			sprintf(dummyname, ".%d-%d-%d-%d.unresolved\0",
				x & 0xFF, (x >> 8)&0xFF, (x >> 16)&0xFF, (x >> 24)&0xFF);
		else if (gs->faked_suffix == 3)
			sprintf(dummyname, ".%d-%d-%d.unresolved\0",
				x & 0xFF, (x >> 8)&0xFF, (x >> 16)&0xFF);
		else if (gs->faked_suffix == 2)
			sprintf(dummyname, ".%d-%d.unresolved\0",
				x & 0xFF, (x >> 8)&0xFF);
		else if (gs->faked_suffix == 1)
			sprintf(dummyname, ".%d.unresolved\0",
				x & 0xFF);
		else
		{
			fprintf(gs->g_outf, "add_newclients: Error, gs->faked_suffix is %d. Exiting.\n",
				gs->faked_suffix);
			exit (-1);
		}
		suffix = dummyname;
		unresolvedflag = 1;
	}
	else
	{
		// Domain name's suffix (non-hostname part) is to the right
		// of the first "."

		if ((suffix = strchr(fqdn, '.')) == NULL)
		{
			// Bad error (there should always be a "." in the resolved
			// domain name, but check in case of program error/failure)
			fprintf(gs->g_outf, "add_newclients: Error, there's no \".\" in %s.\n",
				fqdn);
			fprintf(gs->g_outf, "add_newclients: IP addr is %s\n",
				inet_ntoa(tmp_cip->iaddr));
			exit (-1);
		}
	}

	// Search the seen_client list for an existing
	// entry with same playername and domain suffix
	// (there should be only zero or one match)

	cnt = 0;
	prevclient = NULL;
	tmpclient = gs->clientlist;
	while (tmpclient != NULL)
	{
		if ((strcmp(tmpclient->pname, seenp->playername) == 0) &&
			(strcmp(tmpclient->suffix, suffix) == 0))
		{
			// Cool, this <playername,suffix> already seen. If it
			// is further than 32 entries from head of list, move it
			// to head of list. (Weak attempt to benefit from locality
			// in client lookups across games.)
			if (cnt > 31)
			{
				prevclient->next = tmpclient->next;
				tmpclient->next = gs->clientlist;
				gs->clientlist = tmpclient;
			}
			
			tmpclient->numseen++;

			// Add the ipaddr:port pair(s) from this seen_player
			// entry to the list hanging off the global seen_client
			// entry for this <playername,suffix>

			tmpclient->numip += push_new_clientports(tmpclient, seenp, gs);

			// Increment the cumulative count of this player's playing time
			tmpclient->totaltime += seenp->totaltime;

			// Prepend a copy of the pinglist from the supplied seen_player
			// entry to the pinglist hanging off this seen_client
			// entry.

			tmpclient->pingcnt += push_new_clientpings(tmpclient, seenp, gs);

			return;
		}

		prevclient = tmpclient;
		tmpclient = tmpclient->next;
		cnt++;
	}

	// Getting to here means this is the first time we've seen this
	// <playername,suffix>. Create a new entry for it.

	tmpclient = (struct seen_client *)malloc(sizeof(struct seen_client));
	if (tmpclient == NULL)
	{
		fprintf(gs->g_outf, "add_newclients: malloc failure, quitting.\n");
		exit(-1);
	}

	// Initialize simple fields in the seen_client entry

	memset((char *)tmpclient, 0, sizeof(struct seen_client));

	tmpclient->suffix = (char *)malloc(strlen(suffix)+1);
	if (tmpclient->suffix == NULL)
	{
		fprintf(gs->g_outf, "add_newclients: malloc failure, quitting.\n");
		exit(-1);
	}

	tmpclient->pname = (char *)malloc(strlen(seenp->playername)+1);
	if (tmpclient->pname == NULL)
	{
		fprintf(gs->g_outf, "add_newclients: malloc failure, quitting.\n");
		exit(-1);
	}

	tmpclient->numseen = 1;
	tmpclient->unresolved = unresolvedflag;
	tmpclient->numip = push_new_clientports(tmpclient, seenp, gs);
	tmpclient->totaltime = seenp->totaltime;
	strcpy(tmpclient->suffix, suffix);
	strcpy(tmpclient->pname, seenp->playername);

	// Prepend a copy of the pinglist from the supplied seen_player
	// entry to the pinglist hanging off this seen_client
	// entry.

	tmpclient->pingcnt = push_new_clientpings(tmpclient, seenp, gs);


	// If this is the UnnamedPlayer entry, set flag so we can
	// easily ignore the player during post-processing

	if (strcmp(tmpclient->pname, "UnnamedPlayer\0") == 0)
		tmpclient->unnamedplayer = 1;

	// Finally, add to the global seen_client list

	if (gs->clientlist == NULL)
		gs->clientlist = tmpclient; // First playername in the list
	else
	{
		tmpclient->next = (gs->clientlist)->next;
		(gs->clientlist)->next = tmpclient;
	}

	gs->clientlist_cnt++;	// Finally, increment the count of clientlist entries

	return;
}



// push_new_clientports()
//
// Copies over the list of one or more cid_ipaddr entries
// from a seen_player entry into a seen_client entry
//
//
int push_new_clientports(struct seen_client *tmpclient,
					   struct seen_player *seenp,
					   struct global_state *gs)
{
	struct player_ipaddr *paddr, *newpaddr;
	struct cid_ipaddr *tmp_cip;
	int newaddr = 0;
	int seen;

	// Walk through the one or more cid_ipaddr entries hanging
	// of this seen_player entry, adding them to the player_ipaddr
	// list off seen_client entry if they've not been seen before

	tmp_cip = seenp->c_addr;
	while (tmp_cip != NULL)
	{
		seen = 0;
		paddr = tmpclient->iplist;
		while (paddr != NULL)
		{
			if ((paddr->port == tmp_cip->port) &&
				((paddr->iaddr).s_addr == (tmp_cip->iaddr).s_addr))
				seen = 1;
			paddr = paddr->next;
		}

		if (seen == 0)
		{
			// Need to create a new player_ipaddr entry if we
			// didn't see a match already in the list

			newaddr++;

			newpaddr = (struct player_ipaddr *)malloc(sizeof(struct player_ipaddr));
			if (newpaddr == NULL)
			{
				// bad scene, we've got some wierd memory failure going on.		
				 fprintf(gs->g_outf, "push_new_clientports: malloc failure, quitting.\n");
				 exit(-1);
			}

			memset((char *)(newpaddr), 0, sizeof(struct player_ipaddr));
			newpaddr->port = tmp_cip->port;
			memcpy(&(newpaddr->iaddr), &(tmp_cip->iaddr), sizeof(struct in_addr));
			newpaddr->next = tmpclient->iplist;
			tmpclient->iplist = newpaddr;
		}
		else
		{
		//	fprintf(gs->g_outf, "NOTE: push_new_clientports() sees a repeat addr:port for %s\n",
		//		seenp->playername);
		}

		tmp_cip = tmp_cip->next;
	}

	return(newaddr);
}




// push_new_clientpings()
//
// Copies over the pinglist from a seen_player entry into a
// seen_client entry. (pinglist is created whenever ClientPing
// entries are found in the logfile. This only applies to my logfiles
// generated before September 2001. Superseded by CPhisto2 entries.)
//
// 061901: Although not very efficient, the simple approach is
// to malloc() a copy of each pinglist entry in the seen_player list.
//
// 082501: malloc() a single block of space for all
// pinglist entries, then copy over one at a time.
//
//
int push_new_clientpings(struct seen_client *tmpclient,
					   struct seen_player *seenp,
					   struct global_state *gs)
{
	struct player_ping *seenping, *newping;
	struct player_ping *pinglistbase;

	if ((seenping = seenp->pinglist) == NULL)
		return(0);
	
	// malloc() enough space for seenp->pingcnt entries.
	// Step through the seen_player pinglist and create a new
	// linked list of player_ping entries hanging off the
	// seen_client entry. As a by-product, reverses
	// the order of the seen_player's pinglist entries - i.e. restores
	// them in ascending chronological order.

	pinglistbase = (struct player_ping *)malloc(seenp->pingcnt * sizeof(struct player_ping));
	if (pinglistbase == NULL)
	{
		fprintf(gs->g_outf,"push_new_clientpings: malloc failed, seenp->pingcnt %d, exiting!.\n",
			seenp->pingcnt);
		exit(-1);
	}

	newping = pinglistbase;
	while (seenping != NULL)
	{
		newping->ping = seenping->ping;
		newping->tstamp = seenping->tstamp;
		newping->next = tmpclient->pinglist;
		tmpclient->pinglist = newping;

		seenping = seenping->next;
		newping = newping + 1;		// Increments in units of sizeof(struct player_ping)
	}

	return(seenp->pingcnt);
}



// client_stats()
//
// The primary routine for printing out how many NAT clients, how
// many come from 'home' domains, how many minutes total played by
// clients who individually played more than 'gametime' minutes,
// and how many clients from unresolved IP addresses.
//
// Each entry in gs->clientlist is a separate client, already differentiated
// by <playername, suffix> while parsing the logfiles. All we need to do is
// count each entry, check if any ports hanging off it are 'odd', and note
// if the seen_client entry is due to unresolved IP address or not.
//
// Dumps output to gs->g_outf3 (and optionally appends domain name usage
// information to "domain_trees.txt")
//

void client_stats(int gametime,
				struct global_state *gs)
{
	int num_clients;		// Unique clients seen while walking the list

	int numodd;				// Possible NAT boxes seen while walking the list
	int numodd_notresolved;	// Possible NAT boxes of clients due to unresolved ip domain names
	int numodd_home;		// Possible NAT boxes ('odd ports') of clients believed to
							// be from "home domains"

	int num_notresolved;	// Number of clients counted due to unresolved addresses

	int raw_num_suffixes;	// Total resolved suffixes pushed onto local domain name tree
	int num_suffixes;		// Resolved suffixes falling under the "valid_domains" list that
							// were pushed onto local domain name tree (or all domains, if
							// valid_domains wasn't set).
	int norm_num_suffixes;	// 'num_suffixes' counted across normalized suffix tree

	int numhome;			// Num. of suffixes believed to match 'home' domains
	int norm_numhome;		// numhome across normalized suffix tree


	int time_total;			// Sum of the playing times of every client >= min_gametime

	int n;

	struct player_ipaddr *tmpaddr;
	struct seen_client *tmpclist;
	struct dname_list *pname_dnametop;	// Tree of domain name fragments
	struct dname_list *tmpdname;
	struct hostaddrlist *tmpvlist;

	// Walk through the seen_client list, each entry representing a unique
	// client (as determined by the <playername, suffix> argument)
	// Count the clients with 'odd' ports, clients due to unresolved
	// addresses (who had a fake suffix created for them), and clients
	// who were resolved to a real domain name falling under the list
	// of 'home' domains in gs->fragmentlist
	//
	// NOTE: We ignore the UnnamedPlayer, unless specifically requested
	// to include it with the "include_unnamedplayer" option.

	num_clients = 0;
	numodd = 0;
	numodd_notresolved = 0;
	num_suffixes = 0;
	raw_num_suffixes = 0;
	norm_num_suffixes = 0;
	num_notresolved = 0;
	numodd_home = 0;
	numhome = 0;
	time_total = 0;

	pname_dnametop = NULL;

	tmpclist = gs->clientlist;

	while (tmpclist != NULL)
	{
		int tmpodd = 0;

		// Only count this client entry if:
		//	- it has played for MORE than gametime (or LESS than gametime
		// if MAX_GAMETIME is set)
		//  - it isn't the UnnamedPlayer (unless UNNAMED_PLAYER is set).

		if ((FLAG_UNSET(gs->stats, MAX_GAMETIME) && (tmpclist->totaltime >= gametime)) ||
			(FLAG_SET(gs->stats, MAX_GAMETIME) && (tmpclist->totaltime <= gametime)))
		 if ((tmpclist->unnamedplayer == 0) || FLAG_SET(gs->stats, UNNAMED_PLAYER))
		 {
			// Increment the number of seen_client entries, and the number
			// of them due to unresolved addresses
			num_clients++;
			if (tmpclist->unresolved == 1)
				num_notresolved++;

			// Add to the running total of played time

			time_total += tmpclist->totaltime;
			//fprintf(gs->g_outf, "tmpclist->totaltime is %d for %s\n",
			//		tmpclist->totaltime, tmpclist->pname);
		
			// Step through each addr:port entry, and note if any
			// of them are 'odd' (evidence of NAT)

			tmpaddr = tmpclist->iplist;
			while (tmpaddr != NULL)
			{
				tmpodd |= IS_ODD_PORT(tmpaddr->port);
				tmpaddr = tmpaddr->next;
			}

			numodd += tmpodd;
			if (tmpclist->unresolved == 1)
				numodd_notresolved += tmpodd;

			// Check if this domain suffix falls under one of
			// the 'home' domains listed in gs->fragmentlist

			tmpvlist = gs->fragmentlist;
			while (tmpvlist != NULL)
			{
				if (is_name_in_dname(tmpclist->suffix, tmpvlist->name))
				{
					numhome++;
					numodd_home += tmpodd;
					break;
				}
				else
					tmpvlist = tmpvlist->next;
			}

			// Build inverted tree of suffixes for resolved addresses
			// (ignore unresolved addresses, even though a faked suffix
			// has been created for it)
			// Note that this tree will count suffixes multiple
			// times if they've appeared under multiple clients.

			if (tmpclist->unresolved == 0)
				push_domain_name(gs, &pname_dnametop, tmpclist->suffix);

		 }
		tmpclist = tmpclist->next;
	}

	// Count the number of distinct client&suffix seen
	// and the number of these belonging to a 'home' domain


	sort_dname(&pname_dnametop);	// Sort in descending order of counts

	// Count the total number of entries in the suffix tree by summing
	// the top level 'count' values

	raw_num_suffixes = 0;
	tmpdname = pname_dnametop;
	while (tmpdname != NULL)
	{
		raw_num_suffixes += tmpdname->count;
		tmpdname = tmpdname->next;
	}

	if (FLAG_SET(gs->dom_stats, DUMP_DTREE))
	{
		if ((gs->g_outf2 = fopen("domain_trees.txt\0", "a")) == NULL)
		{
			printf("client_stats(): Couldn't open domain_trees.txt for appending. Exiting.\n");
			fprintf(gs->g_outf, "client_stats(): Couldn't open domain_trees.txt for appending. Exiting.\n");
			exit(-1);
		}
		// Dump the suffix tree as it currently exists, use the value in
		// 'num_suffixes' when calculating percentages for each branch of tree

		printf("Dumping domain name suffix tree...\n");
		dump_dname(pname_dnametop, 0, raw_num_suffixes, gametime, gs); // Print results
		
		fclose(gs->g_outf2);
	}

	num_suffixes = raw_num_suffixes;
	if (gs->limitlist != NULL)
	{
		// If we're only interested in a subset of the suffix tree
		// then re-count 'num_suffixes' by calling count_home_domains and
		// supplying the limlist as set of domains we're interested in.
		//
		// NB. limlist MUST be a SUPERSET of fragmentlist for this to work
		// (limlist is set by the "valid_domains" config option)
		
		num_suffixes = count_home_domains(pname_dnametop, gs->limitlist, gs->stats);
	}

	// Now recalculate, but normalize the counts along each branch of the
	// tree so we only count each suffix, not each distinct client&suffix
	// (since pname_dnametop is orginally constructed by push_domain_name()
	// to count each instance of client&suffix, rather than just per suffix)

	n = normalize_dname(&pname_dnametop);

	// A "normalized" count how many of the suffixes fall under the suffix list
	// hanging of gs->fragmentlist

	norm_numhome = count_home_domains(pname_dnametop, gs->fragmentlist, gs->stats);

	if (gs->limitlist != NULL)
	{
		// Minor hack, if we're only interested in a subset of the suffix tree
		// then calculate 'norm_num_suffixes' by calling count_home_domains and
		// supplying the limlist as set of domains we're interested in.
		norm_num_suffixes = count_home_domains(pname_dnametop, gs->limitlist, gs->stats);
	}
	else
	{
		// Otherwise, walk through the first level entries in pname_dnametop
		// and count the number of NORMALIZED subentries to obtain total number
		// of suffixes (normalized)
		tmpdname = pname_dnametop;
		while (tmpdname != NULL)
		{
			norm_num_suffixes += tmpdname->count;
			tmpdname = tmpdname->next;
		}
	}

	// Print out the results in either human readable or machine readable form

	if (FLAG_SET(gs->stats, HUMAN_READABLE))
	{
	 fprintf(gs->g_outf3, "Estimated %d clients", num_clients);

	 if (FLAG_SET(gs->stats, MAX_GAMETIME))
		 fprintf(gs->g_outf3, " <= %3d min ", gametime/60);
	 else
		 fprintf(gs->g_outf3, " >= %3d min ", gametime/60);


	 fprintf(gs->g_outf3, "of which %d (%4.1f%%) had unusual ports.\n",
			 numodd, (100*numodd)/(float)num_clients);

 	 fprintf(gs->g_outf3, "%d (%4.1f%%) clients from unresolved addr, %d (%4.1f%%) had unusual ports. \n",
			 num_notresolved, (100*num_notresolved)/(float)num_clients,
			 numodd_notresolved, (100*numodd_notresolved)/(float)num_notresolved);

 	 fprintf(gs->g_outf3, "%4.1f%% of the unusual ports came from clients of unresolved addresses.\n",
			(100*numodd_notresolved)/(float)numodd);		
		
	 fprintf(gs->g_outf3, "Saw %d resolved client&suffix pairs before applying 'validdomains' filter\n",
		 raw_num_suffixes);

	 fprintf(gs->g_outf3, "Of %d filtered client&suffix pairs, %d (%4.1f%%) were HOME domains\n",
			num_suffixes, numhome,
			100*(float)(numhome)/(float)(num_suffixes));

	 fprintf(gs->g_outf3, "Of %d unique/filtered suffixes, %d (%4.1f%%) were HOME domains\n",
			norm_num_suffixes, norm_numhome,
			100*(float)(norm_numhome)/(float)(norm_num_suffixes));

	 fprintf(gs->g_outf3, "%d (%4.1f%%) clients in HOME domains had unusual ports.\n",
			numodd_home, 100*(float)(numodd_home)/numhome);
	 fprintf(gs->g_outf3, "These clients contributed %5d seconds playing time\n",
		 time_total);

	 fprintf(gs->g_outf3, "\n");
	}
	else
	{
		// Print results as a string of numbers per line for direct pasting
		// into a spread sheet or similar. Each line contains the following results
		//
		// min_gametime,
		// number of unique clients,
		// number of unique clients with odd ports (possible NAT)
		// number of clients from unresolved ipaddr,
		// number of odd ports from clients due to unresolved ipaddr
		// number of resolved client&suffixes,
		// number of resolved client&suffixes from "valid" domains,
		// number of resolved client&suffixes from HOME domains,
		// number of unique/resolved suffixes,
		// number of unique/resolved suffixes from HOME domains,
		// number of resolved client&suffixes from HOME domains with ODD ports (possible NAT)
		// total number of seconds playing time contributed by these clients

		
	 fprintf(gs->g_outf3, "%3d, ",
			 gametime/60);

	 fprintf(gs->g_outf3, "%5d, %5d, ",
			 num_clients, numodd);

 	 fprintf(gs->g_outf3, "%5d, ",
			 num_notresolved);
		
	 if (num_notresolved == 0)
		 fprintf(gs->g_outf3, "0, ");
	 else
		 fprintf(gs->g_outf3, "%5d, ",
			 numodd_notresolved);
		
  	 fprintf(gs->g_outf3, "%5d, ",
			raw_num_suffixes);

	 fprintf(gs->g_outf3, "%5d, %5d, ",
			num_suffixes, numhome);

	 fprintf(gs->g_outf3, "%5d, %5d, ",
			norm_num_suffixes, norm_numhome);

	 fprintf(gs->g_outf3, "%5d, %5d\n",
			numodd_home, time_total);
	}

	return;
}

