// gmodstat_clients.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Routines for parsing specific logfile tokens
// (such as "ClientBegin:", "Item:", etc..)
//

#include "stdafx.h"
#include "gmodstat.h"



//
// clientconnect()
//
// Parse a ClientConnect line to establish a new initial
// Client ID in the current game.
//

struct client_id *clientconnect(char *p,
								int ts,
								struct global_state *gs)
{
	struct seen_games *gl = gs->gamelist;
	struct client_id *newcid;
	int pnum;
	char ipaddr_str[TOKENLEN];
	struct in_addr iaddr;
	int udpport;
	int is_bot = 0;
	int is_seen = 0;
	int is_addr = 0;


	// Sanity check that a game as already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "clientconnect(): At %d:%02d gamelist is NULL\n",
		ts/60, (ts % 60));
		return(NULL);
	}

	// Extract client ID number and possibly IPaddr from command line

	memset(ipaddr_str, 0, TOKENLEN);
	sscanf(p, "%d  %[^\n\t\r]", &pnum, ipaddr_str);  // Line may end with one or other of
												// "\n" (newline) "\r" (carriage return)
												// or "\t" (tab) depending on post-processing.

	// If this log file was generated by a QIII server running gja's
	// mod, the Connect line's 2nd paramater is non-null string of
	// characters, either an IPaddr:port, "bot", or "seen". The string
	// will be "seen" if the Connect is part of a game starting a new map
	// and the client had already Connect'ed during a previous game. Typically
	// the IPaddr:port is only supplied the first time a client connects, or
	// if there's some network disruption and the connection comes back mid-game.

	if (strlen(ipaddr_str) > 0)
	{
		if (strcmp(ipaddr_str, "bot\0") == 0)
		{
			is_bot = 1;
		}
		else if (strcmp(ipaddr_str, "seen\0") == 0)
		{
			is_seen = 1;
		}
		else
		{
			int tmp;
			char dottedquad[32];

			// Should be an ASCII string of form "w.x.y.z:nn"
			// Extract the IP address and 16 bit port number

			tmp = sscanf(ipaddr_str, "%[^:] : %d",
					dottedquad, &udpport);

			if ((tmp == 0) || (tmp == EOF))
				{
				 fprintf(gs->g_outf,
					 "clientconnect(): Zero fields matched, or EOF (code %d) parsing IP address .\n", tmp);
				 return(NULL);
				}

			memset((char *)(&iaddr), 0, sizeof(struct in_addr));
			if ((iaddr.s_addr = inet_addr(dottedquad)) == INADDR_NONE)
				{
				 fprintf(gs->g_outf,
					 "clientconnect(): inet_addr(%s) failed parsing IP address. port %d\n",
					 dottedquad, udpport);
				 return(NULL);
				}

			is_addr = 1;
		}
	}


	// Step through the existing client_id list to verify there's no
	// match to an existing client_id number

	newcid = gl->c_now;
	while (newcid != NULL)
	{
		if (newcid->c_id == pnum)
		{
			// Getting here means there's already a client_id entry with
			// matching client ID number, ie. client has joined and not yet
			// disconnected. In this case, we silently update the timestamp
			// (on the assumption that the client hasn't started playing yet)

			// Flag a message if there's been an unexpected change in
			// the ipaddr:port information
			// Note: Over XXXX clients during Feb/Mar/Apr 2001 I saw this event
			// occur only twice, and in each case the IP address didn't change
			// only the port number (and in each case both before and after port
			// numbers were 'odd' ports).

			if (is_addr == 1)
				if (iaddr.s_addr != (newcid->iaddr).s_addr)
				{
				 // But the server thinks its new and different
				 fprintf(gs->g_outf,"clientconnect(): At %d:%02d client_id %d already exists in state %d.\n",
				 ts/60, (ts % 60), pnum, newcid->state);
				 fprintf(gs->g_outf, "              Orig IP: %s:%d   New IP: %s:%d\n",
					inet_ntoa(newcid->iaddr), newcid->port, inet_ntoa(iaddr), udpport);

				 // Add this new ipaddress:port information to the seen_player entry
				 // associated with this existing client, and update the current
				 // client_id field's ipaddr:port information
				 c_add_ipaddr(newcid->c_info, &iaddr, udpport, gs);
				 newcid->port = udpport;
				 memcpy(&(newcid->iaddr), &iaddr, sizeof(struct in_addr));
				}

			newcid->state = C_CONN; // Reset state flag
			newcid->s_time = ts; // Update the "seen" timestamp
			return(newcid); // Do nothing else but return as though all is normal
		}

		newcid = newcid->next;
	}

	// Okay, build a new client_id entry

	if ((newcid = (struct client_id *)malloc(sizeof(struct client_id))) == NULL)
	{
		 fprintf(gs->g_outf, "clientconnect(): malloc failure, exiting.\n");
		 exit(-1);
	}

	// Initialize the new client_id entry.

	memset((char *)newcid, 0, sizeof(struct client_id));

	newcid->next = gl->c_now; // Add client_id entry to current game.
	gl->c_now = newcid;

	newcid->state = C_CONN;
	newcid->s_time = ts;
	newcid->c_id = pnum;

	// If there was an additional parameter to "ClientConnect", process it

	if ((is_addr | is_seen | is_bot) == 1)
	{
		// Now finally, either copy in the w.x.y.z:nn (IPaddr:port) text, or
		// if "seen" then grab the w.x.y.z:nn (IPaddr:port) text
		// from the same clientID entry of the previous game

		if (is_seen == 1)
		{
			struct client_id *prevcid;

			// Go find the same clientID in the previous game, copy its
			// IPaddr:port info into the ipaddr string.
			if (gl->next == NULL)
				{
				 // Unlikely error, there's "seen" but no previous game.
				 fprintf(gs->g_outf,"clientconnect(): At %d:%02d there's NO previous game.\n",
				 ts/60, (ts % 60));
				}

			if ((prevcid = findclient_id(gl->next->c_now, newcid->c_id)) == NULL)
				{
				 // But the server thinks its new and different
				 fprintf(gs->g_outf,"clientconnect(): At %d:%02d client_id %d not found in previous game.\n",
				 ts/60, (ts % 60), pnum);
				}
			else
			{
				newcid->port = prevcid->port;
				memcpy(&(newcid->iaddr), &(prevcid->iaddr), sizeof(struct in_addr));
			}
		}



		if(is_addr == 1)
		{
			newcid->port = udpport;
			memcpy(&(newcid->iaddr), &iaddr, sizeof(struct in_addr));
		}

		newcid->is_bot = is_bot;

	}
	
	return(newcid);
}


// clientuserinfochanged()
//
// Parse the rest of the ClientUserInfoChanged line to
// extract player name and associate it with previously
// seen Client ID.
//
// This event can occur even while the client is already
// connected, for example when shifting from playing to
// spectating mode.
//

struct client_id *clientuserinfochanged(char *p,
								int ts,
								struct global_state *gs)
{
	struct seen_games *gl = gs->gamelist;
	char playername[TOKENLEN];
	int pnum;
	int playmode;				// 0 for "playing", 3 for "spectating"
	struct client_id *tmpcid;
	struct seen_player *tmpplay;

	// Sanity check that a game as already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "clientuserinfochanged(): At %d:%02d: gamelist is NULL.\n",
						ts/60, (ts % 60));
		return(NULL);
	}

	// clientuserinfochanged can occur even after a game finishes but before
	// the actual "shutdown"


	// Extract player information
	// Assume name is first parameter after the client ID, and
	// parameter after the name is "\t\N" where N is an integer
	// representing playing (0) or spectating (3). [For team
	// based games, 1 represents RED team and 2 is BLUE team.]

	memset((char *)playername, 0, TOKENLEN);
	sscanf(p, " %d n\\%64[^\\]\\t\\%d", &pnum, playername, &playmode);

	// Look for match to player number in current clientID list

	if ((tmpcid = findclient_id(gl->c_now, pnum)) == NULL)
	{
		// A previous ClientConnect has not been seen, so this is an error.
		fprintf(gs->g_outf, "clientuserinfochanged(): At %d:%02d: %s has no client ID.\n",
						ts/60, (ts % 60), playername);
		return(NULL);
	}

	// tmpcid now points to the client_id entry with the same client ID number
	// as the client whose state is now being changed.

	// If this client_id is already associated with a seen_player entry,
	// it might be that the player's name is being changed in-game,
	// or it is moving between playing and spectating modes, or this
	// is simply a benign repeat of a previous ClientUserinfoChanged event.

	if (tmpcid->c_info != NULL)
	{
		// Copy the supplied playername into the seen_player entry, and
		// forget the old name ever existed (necessary for keeping rational
		// sense of kills, deaths, etc). Also copy over the new 'playmode'
		// into the client_id entry (this will be used on the subsequent
		// ClientBegin to appropriately fill the next player_inout entry).

		tmpcid->playmode = playmode;

		if ((strlen(playername) != strlen((tmpcid->c_info)->playername))
			|| (strcmp (playername, (tmpcid->c_info)->playername) != 0))
		{
			// Log the fact that this name has been changed (even though we
			// dont currently keep a list of the previous names used by this
			// player, I guess this could be done later)
			//
			// As for tracking playernames, the global seen_client list only
			// records the player's final name when each game ends
			// (during 'Shutdown' processing), so we don't really mind
			// name changes in-game (typically players change not long
			// after they log in, and are happy after that).

			(tmpcid->c_info)->namechanged++;
   			//fprintf(gs->g_outf, "clientuserinfochanged(): At %d:%02d: Name being changed from %s to %s\n",
   			//	ts/60, (ts % 60), (tmpcid->c_info)->playername, playername);
		}

		strcpy((tmpcid->c_info)->playername, playername);
		return(tmpcid);
	}

	// Okay, so there's an existing client_id entry but it isn't
	// active yet (not pointing to a seen_player entry).
	// Search the seen_player name list for a matching name and ipaddr:port
	// (has this guy appeared before in this particular game?) and if not,
	// create a new one.
	//
	// (NOTE: This algorithm CAN handle concurrent clients from the same
	// ipaddr with same name that only differ in the port number. The QuakeIII
	// server does allow multiple clients differing only in the port number,
	// and will assign them different client_id numbers for concurrent play
	// even though the visible ASCII playername is identical.)

	tmpplay = gl->c_all;

	while (tmpplay != NULL)
	{
		if ((strcmp (playername, tmpplay->playername) == 0) && (tmpplay->c_addr != NULL))
			if (((tmpcid->iaddr).s_addr == (tmpplay->c_addr->iaddr).s_addr) &&
				(tmpcid->port == tmpplay->c_addr->port))
			{
				// A player with this playername has already been in this game (joined,
				// left, reconnecting) from the same IPaddr:port
				// Update counters, link it back to the client_id entry,
				// add this ipaddr:port to the seen_player list (although this is
				// probably redundant?)
				tmpplay->numseen++;
				tmpcid->playmode = playmode;
				tmpcid->c_info = tmpplay;
				tmpcid->state = C_INFO;
				tmpcid->s_time = ts;
				if (tmpcid->is_bot == 0)
					c_add_ipaddr(tmpplay, &(tmpcid->iaddr), tmpcid->port, gs);

				return(tmpcid);
			}

		tmpplay = tmpplay->next;	
	}

	// No player with this player name and IPaddr:port has been seen in this game
	// until now, so create a seen_player entry and link to client_id entry.

	tmpplay = (struct seen_player *)malloc(sizeof(struct seen_player));

	if (tmpplay == NULL)
	{
		 fprintf(gs->g_outf, "clientuserinfochanged(): malloc failure, exiting.\n");
		 exit(-1);
	}

	// Zero the new seen_player entry and initialize some fields

	memset((char *)tmpplay, 0, sizeof(struct seen_player));

	tmpplay->next = gl->c_all; // Add seen_player entry to current game.
	gl->c_all = tmpplay;

	strcpy(tmpplay->playername, playername);
	tmpplay->numseen = 1;	// This entry counts as first seen
	tmpplay->frags = -1;
	tmpplay->ping = -1;

	// Add first ipaddr:port pair to this seen_player entry
	// (if this isn't a bot, which has no ipaddr)

	if (tmpcid->is_bot == 0)
		c_add_ipaddr(tmpplay, &(tmpcid->iaddr), tmpcid->port, gs);

	// Adjust the associated client_id entry

	tmpcid->playmode = playmode;	// Store for later use by ClientBegin
	tmpcid->state = C_INFO;
	tmpcid->s_time = ts;
	tmpcid->c_info = tmpplay;


	return(tmpcid);
}






// clientbegin()
//
// Parse the rest of the ClientBegin line to find which Client ID
// in the current game has moved to playing state.
//
// Note: If the routine is called when there's already an active
// player_inout entry associated with the seen_player (i.e. its
// end timestamp is still -1) then in general we close out that
// player_inout entry and push a new one on. However, if the active
// inout entry has a *start* time of -1 we simply update the start
// time with our current timestamp and exit [since the top inout
// entry must have been created by clientkill() routine forced to
// preemptively establish an inout entry due to a race condition
// in the way the server somtimes creates log entries]
//

struct client_id *clientbegin(char *p,
								int ts,
								struct global_state *gs)
{
	struct seen_games *gl = gs->gamelist;
	struct client_id *tmpcid;
	int pnum;

	// Sanity check that a game as already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "clientbegin(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return(NULL);
	}

	sscanf(p, "%d", &pnum); // Extract client id

	// Sanity check that a game hasn't already finished
	if (gl->t_end != -1)
		return(NULL);

	// Search for existing matching client id
	if ((tmpcid = findclient_id(gl->c_now, pnum)) == NULL)
	{
		fprintf(gs->g_outf, "clientbegin(): At %d:%02d Client %d doesnt exist\n",
			ts/60, (ts % 60), pnum);
		return(NULL);
	}

	if (tmpcid->c_info == NULL)
	{
		fprintf(gs->g_outf, "clientbegin(): At %d:%02d Client %d has no seen_player entry\n",
			ts/60, (ts % 60), pnum);
		return(NULL);
	}

	// The client may already be active, and this ClientBegin
	// might be due to the client shifting between playing to spectating
	// states. If already active, close out the previous inout entry
	// and start a new one.

	if (tmpcid->c_info->inout != NULL)
	{
		if (tmpcid->c_info->inout->in_ts == -1)
		{
			// The current inout entry was partially installed
			// during some other event processing. Just update the
			// timestamp and return.
			tmpcid->c_info->inout->in_ts = ts;
			return(tmpcid);
		}

		if (tmpcid->c_info->inout->out_ts == -1)
			player_out(tmpcid->c_info, ts); // Close the previous inout entry
	}

	tmpcid->state = C_BEGIN;
	tmpcid->s_time = ts;
	// Update the player inout time list
	player_in(tmpcid->c_info, ts, tmpcid->playmode, gs);
	
	return(tmpcid);
}




//
// clientdisconnect()
//
// Parse a Disconnect line to find which Client ID
// in the current game has moved to non-playing state
//

struct seen_player *clientdisconnect(char *p,
								int ts,
								struct global_state *gs)
{
	struct seen_games *gl = gs->gamelist;
	struct client_id *tmpcid, *tmp2cid;
	struct seen_player *tmpplayer;
	int pnum;

	// Sanity check that a game as already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "clientdisconnect(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return(NULL);
	}

	// Disconnects can occur even after a game finishes but before
	// the actual "shutdown", so go ahead regardless whether or not
	// the game is finished.

	// NOTE: Disconnects can also occur for 'orphaned' client_id entries
	// e.g. a player connects on one client_id, executes the changeinfo,
	// but then reconnects at a lower client_id number because someone
	// else just left. in this case the higher client_id is orphaned, and
	// the server will eventually time it out and issue a ClientDisconnect.
	// we must be careful not to corrupt a valid playername that *was*
	// associated with the orphaned client_id but is now being served by
	// a different, and still active, client_id entry. the problem is having
	// two client_id entries point to a single seen_player entry.
	// [This case is rare, but did occur in my feb/mar/apr 2001 trials.]
	
	sscanf(p, "%d", &pnum); // Extract client id

	tmpcid = gl->c_now;
	tmp2cid = NULL;
	while (tmpcid != NULL)	// Search for existing matching client id
	{
		if (tmpcid->c_id == pnum)
		{	// Found match
			
			// Update timestamp on matching seen_player entry

			tmpplayer = tmpcid->c_info;
			if (tmpplayer != NULL)
			{
				player_out(tmpplayer, ts); // Update player inout times
			}
			else // Only an error if there *should* have been seen_player info
				if(tmpcid->state != C_CONN)
				{
					fprintf(gs->g_outf,"clientdisconnect(): At %d:%02d, no player for %d",
						ts/60, (ts % 60), pnum);
					fprintf(gs->g_outf, " client state was %d\n", tmpcid->state);
				}

			// Remove client_id from game list
			if (tmp2cid == NULL)
			{
				gl->c_now = tmpcid->next;
			}
			else
			{
				tmp2cid->next = tmpcid->next;
			}

			free(tmpcid);

			return(tmpplayer);
		}

		// Step to next client_id in list
		tmp2cid = tmpcid;
		tmpcid = tmpcid->next;
	}
	
	fprintf(gs->g_outf, "clientdisconnect(): At %d:%02d, %d doesnt exist\n",
			ts/60, (ts % 60), pnum);

	return(NULL);
}




// clientsay()
//
// Parse the rest of the Say line, and if the ClientID is
// greater than a specified parameter store the said text
//

struct client_id *clientsay(char *p,
								int ts,
								struct global_state *gs,
								int saycid)
{
	struct seen_games *gl = gs->gamelist;
	struct client_id *tmpcid;
	struct seen_player *tmpplayer;
	struct seen_say *tmpsay;
	char *tmptext;
	char playername[TOKENLEN];
	int numchar;

	if (saycid == -1)
		return(NULL);

	// Sanity check that a game has already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "clientsay(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return(NULL);
	}

	sscanf(p, "%[^:] %n", playername, &numchar); // Extract player name
	p += numchar;	// Move pointer past the player name

	// Two step search, first find playername in the seen_player list
	// and then find the client_id entry pointing to that seen_player entry

	tmpplayer = gl->c_all;
	while (tmpplayer != NULL)
	{
		if (strcmp(playername, tmpplayer->playername) == 0)
			break;
		tmpplayer = tmpplayer->next;
	}

	if (tmpplayer == NULL)
	{
		fprintf(gs->g_outf, "clientsay(): At %d:%02d, %s hasn't played in this game\n",
			ts/60, (ts % 60), playername);
		return(NULL);
	}

	// Now we have the seen_player entry, find the client_id entry

	tmpcid = gl->c_now;
	while (tmpcid != NULL)
	{
		if (tmpcid->c_info == tmpplayer)
			break;
		tmpcid = tmpcid->next;
	}

	if (tmpcid == NULL)
	{
		fprintf(gs->g_outf, "clientsay(): At %d:%02d, %s isn't currently playing this game\n",
			ts/60, (ts % 60), playername);
		return(NULL);
	}
	
	// Now, check to see if the client ID is in the acceptable range

	if (tmpcid->c_id < saycid)
		return(NULL);

	// Create entry and space for text actually "said" and store it in
	// the associated seen_game entry

	if ((tmpsay = (struct seen_say *)malloc(sizeof(struct seen_say))) == NULL)
	{
		 fprintf(gs->g_outf, "clientsay(): At %d:%02d, Memory allocation failure, quitting.\n",
			 ts/60, (ts % 60));
		 exit(-1);
	}

	// Initialize the new seen_say entry
	memset((char *)tmpsay, 0, sizeof(struct seen_say));

	// Allocate space to store copy of the text that was said
	if ((tmptext = (char *)malloc(strlen(p))) == NULL)
	{
		 fprintf(gs->g_outf, "clientsay(): At %d:%02d, Memory allocation failure, quitting.\n",
			 ts/60, (ts % 60));
		 exit(-1);
	}

	// Copy over the said text
	strcpy(tmptext, p);

	tmpsay->saytext = tmptext;
	tmpsay->tstamp = ts;
	tmpsay->who = tmpplayer;

	tmpsay->next = gl->say;
	if (gl->say != NULL)
		(gl->say)->prev = tmpsay;
	gl->say = tmpsay;

	return(tmpcid);
}



// clientkill()
//
// Parse the rest of the Kill line to find who was killed, who
// was the killer, and the method of death.
//
// Note there is a race condition I observed a few times over
// a 5 month period - the server may log a client having killed
// someone else in between the client's ClientUserinfoChanged and
// ClienBegin. This is wierd, but true. So, we do not check to confirm
// that the client has a valid "inout" list.
//

struct client_id *clientkill(char *p,
								int ts,
								struct global_state *gs)
{
	struct seen_games *gl = gs->gamelist;
	struct client_id *killercid, *killedcid;
	int killer, killed, weapon;

	// Sanity check that a game as already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "clientkill(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return(NULL);
	}

	// Extract two client IDs and weapon number from command line

	sscanf(p, "%d %d %d", &killer, &killed, &weapon);

	// Find and verify killed party's client_id list entry

	if ((killedcid = findclient_id(gl->c_now, killed)) == NULL)
	{
		fprintf(gs->g_outf,"clientkill(): At %d:%02d, killed's client_id %d doesn't appear in local list.\n",
			ts/60, ts%60, killed);
		return(NULL);
	}

	if (killedcid->c_info == NULL)
	{
		fprintf(gs->g_outf,"clientkill(): At %d:%02d, killed's client_id %d has no playername.\n",
			ts/60, ts%60, killed);
		return(NULL);
	}

	if (killer != 1022)
	{
		// Assuming the killer is not <world> (killer ID = 1022),
		// find and verify killer's client_id list entry

		if ((killercid = findclient_id(gl->c_now, killer)) == NULL)
		{
			fprintf(gs->g_outf,"clientkill(): At %d:%02d, killer's client_id %d doesn't appear in local list.\n",
				ts/60, ts%60, killer);
			return(NULL);
		}

		if (killercid->c_info == NULL)
		{
			fprintf(gs->g_outf,"clientkill(): At %d:%02d, killer's client_id %d has no playername.\n",
				ts/60, ts%60, killer);
			return(NULL);
		}

		if (killercid->c_info->inout == NULL)
		{
			char fakeline[32];
			// This is the race condition. The ClientBegin will occur fairly soon
			// after the Kill event in the logfile, but in order to account for
			// frags accurately we need to preemptively initate ClientBegin processing.
			// So, fake the ClientBegin event and set start time to -1 so that the real
			// ClientBegin knows to just update the timestamp and nothing else.

			sprintf(fakeline, " %d\0", killer);
			clientbegin(fakeline, ts, gs);
			killercid->c_info->inout->in_ts = -1;
		}

		if (killercid->c_info->inout->out_ts != -1)
			{
				fprintf(gs->g_outf,"clientkill(): At %d:%02d, killer %s (%d,%d) has supposedly left game!.\n",
					ts/60, ts%60, killercid->c_info->playername, killer, killercid->c_id);
				return(NULL);
			}

	}
	else
	{
		// Killer was <world>, in which case we register the death against
		// the killed player as a decrement of the killed player's frag
		// count and return.

		((killedcid->c_info)->inout)->frags--;
		update_killedbyworld(killedcid->c_info, weapon, gs);
		return(NULL);
	}


	if (killer == killed)
	{
		// This is a case of suicide (or stupidity). Either way,
		// the player has their frag count decremented. I'll log
		// the event as a death rather than a kill for this player.

		((killedcid->c_info)->inout)->frags--;

		update_kill_list(&((killedcid->c_info)->death), killercid->c_info, weapon, gs);
	}
	else
	{
		// Okay, a real kill. Credit the killer, increment killer's
		// frag count, and log the death against the killed player.

		update_kill_list(&((killercid->c_info)->kill), killedcid->c_info, weapon, gs);

		((killercid->c_info)->inout)->frags++;

		update_kill_list(&((killedcid->c_info)->death), killercid->c_info, weapon, gs);
	}

	return(NULL);
}





// clientitem()
//
// Parses the "Item: N <string>" events and create/maintain a list
// of collected items per-game per-player
//
void clientitem(char *p, int ts, struct global_state *gs)
{
	struct client_id *itemcid;
	struct seen_player *player;
	struct player_item *tmpitem;
	int client;
	char item[32];		// Item names are less than 32 characters

	// Sanity check that a game as already begun, and space allocated.
	if (gs->gamelist == NULL)
	{
		fprintf(gs->g_outf, "clientitem(): At %d:%02d gamelist is NULL\n",
			ts/60, ts%60);
		return;
	}

	// Extract client ID and item name from command line

	sscanf(p, "%d %s", &client, item);

	// Find the client's client_id list entry

	if ((itemcid = findclient_id(gs->gamelist->c_now, client)) == NULL)
	{
		fprintf(gs->g_outf,"clientitem(): At %d:%02d, client_id %d not in local list.\n",
			ts/60, ts%60, client);
		return;
	}

	// Sanity check the client is actually an active player
	if ((player = itemcid->c_info) == NULL)
	{
		fprintf(gs->g_outf,"clientitem(): At %d:%02d, client_id %d has no playername.\n",
			ts/60, ts%60, client);
		return;
	}

	// Note that Item events can occur *after* the end of a game (the "Exit"
	// event) and (oddly) between the client's ClientUserinfoChanged and
	// ClientBegin events. So it doesn't make any sense to check if the game
	// is over or if there's a valid inout entry. [This might change if
	// item collection was ever to be associated with a specific inout
	// entry, in which case a 'fake' ClientBegin would need to be performed,
	// as is done with clientkill() processing.]

	// With a valid seen_player pointer, update the list of collected
	// items. First search to see if the item's already in the list,
	// and update counter. Otherwise, add new entry.

	tmpitem = player->items;
	while (tmpitem != NULL)
	{
		if (strcmp(tmpitem->itemname, item) == 0)
			break;
		tmpitem = tmpitem->next;
	}

	player->itemcnt++;

	if (tmpitem != NULL)
	{
		tmpitem->itemcnt++;
		return;
	}

	// Getting to here, need to create and add a new item

	if ((tmpitem = (struct player_item *)malloc(sizeof(struct player_item))) == NULL)
	{
		fprintf(gs->g_outf,"clientitem(): At %d:%02d, malloc failed, exiting!.\n",
			ts/60, ts%60);
		exit(-1);
	}

	if ((tmpitem->itemname = (char *)malloc(strlen(item)+1)) == NULL)
	{
		fprintf(gs->g_outf,"clientitem(): At %d:%02d, malloc failed, exiting!.\n",
			ts/60, ts%60);
		exit(-1);
	}

	tmpitem->itemcnt = 1;
	strcpy(tmpitem->itemname, item);
	tmpitem->next = player->items;
	player->items = tmpitem;

	return;
}


