// gmodstat_dns.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Routines for reverse lookups of client IP addresses,
// storing and analysing the resolved domain names, etc.
//

#include "stdafx.h"
#include "gmodstat.h"



// resolve_clientaddr()
//
// Locates the hostaddrlist entry corresponding to a supplied
// ipaddress in struct in_addr format.
//
// If the ipaddr doesn't exist in gs->dnslist, or the ipaddr exists
// but there's a null domain name, optionally retry a gethostbyaddr()
// and add/update gs->dnslist with the result.
//
struct hostaddrlist *resolve_clientaddr(struct in_addr *iaddr,
										struct global_state *gs)
{
	struct hostent *hostname = NULL;
	struct hostaddrlist *tmphlist, *prevhlist;
	unsigned int x = iaddr->s_addr;
	int cnt;

	// Walk the existing gs->dnslist for a matching ipaddr

	cnt = 0;
	prevhlist = NULL;
	tmphlist = gs->dnslist;
	while (tmphlist != NULL)
	{
		if (tmphlist->ipaddr == x)
			break;
		else
		{
			prevhlist = tmphlist;
			tmphlist = tmphlist->next;
			cnt++;
		}
	}

	// If a match found:
	//	If the name isn't null, return now.
	//  If the name is null, optionally re-attempt a resolution
	//  and update the gs->dnslist entry before returning.

	if (tmphlist != NULL)
	{
		// Hack optimization - if this entry was more than 32 elements
		// from the top of the list, move it to the top. Ensures the
		// search makes use of locality - IP addresses most recently seen
		// will be nearer the top of the list.
		if (cnt > 31)
		{
			prevhlist->next = tmphlist->next;
			tmphlist->next = gs->dnslist;
			gs->dnslist = tmphlist;
		}
		
		if (tmphlist->name != NULL)
			return(tmphlist);

		// Name is null - attempt a re-resolution if RESOLVE_AGAIN is set
		// and tmphlist->trieddns is unset (If tmphlist->trieddns is set,
		// we've already had a go at this ipaddr and couldn't resolve
		// in before, so don't try again.

		if ((FLAG_SET(gs->dom_stats, RESOLVE_AGAIN)) && (tmphlist->trieddns == 0))
		{
			hostname = gethostbyaddr((char *)(&x), sizeof(int), AF_INET);
			tmphlist->trieddns = 1;	// Set to show we've now tried at least once
									// for this particular ipaddr

			if (hostname == NULL)
			{
				if (FLAG_SET(gs->dom_stats, DUMP_RESOLVES))
					fprintf(gs->g_outf, "%s\tFailed: %d\n",
						inet_ntoa(*iaddr), WSAGetLastError());

			}
			else
			{
				if (hostname->h_name != NULL)
				{
					// We resolved a name. Update the entry pointed to
					// by tmphlist, case we're called again for the same ipaddr

					tmphlist->name = (char *)malloc(strlen(hostname->h_name)+1);
					if (tmphlist->name == NULL)
					{
						 fprintf(gs->g_outf, "resolve_clientaddr: malloc failure, quitting.\n");
						 exit(-1);
					}
					strcpy(tmphlist->name, hostname->h_name);
				}
			}
		}

		// Return pointer to hostaddrlist entry, regardless of whether we managed
		// to actually re-resolve a domain name

		return(tmphlist);
	}

	// If there's no gs->dnslist entry for this ipaddr, create one, mark
	// it as having had a resolution attempted, and then attempt gethostbyaddr()

	// If RESOLVE_MISSING is not set, just log the error (which suggests
	// we need to update the file being used to resolve addresses).

	if (FLAG_UNSET(gs->dom_stats, RESOLVE_MISSING))
	{
		// Print this error regardless of wether the
		// DUMP_RESOLVES flag is set, since we need to know
		// that our resolution data may be out of date.
		fprintf(gs->g_outf, "resolve_clientaddr: WARNING %s not found in hostdomain_file list\n",
			inet_ntoa(*iaddr));
		return(NULL);
	}

	// Create a new, blank hostaddrlist entry and add it
	// to gs->dnslist

	if ((tmphlist =
		 (struct hostaddrlist *)malloc(sizeof(struct hostaddrlist)))
		 == NULL)
	{
		fprintf(gs->g_outf, "resolve_clientaddr: malloc failure, quitting.\n");
		exit(-1);
	}

	memset((char *)tmphlist, 0, sizeof(struct hostaddrlist));
	tmphlist->ipaddr = x;
	tmphlist->next = gs->dnslist;
	gs->dnslist = tmphlist;
	tmphlist->trieddns = 1;

	// Attempt the new gethostbyaddr()

	hostname = gethostbyaddr((char *)(&x), sizeof(int), AF_INET);
	if (hostname == NULL)
	{
		if (FLAG_SET(gs->dom_stats, DUMP_RESOLVES))
			fprintf(gs->g_outf, "%s\tFailed: %d\n",
				inet_ntoa(*iaddr), WSAGetLastError());
		return(tmphlist);
	}

	if (hostname->h_name == NULL)
	{
		// Perhaps an odd situation, the gethostbyaddr() returned
		// okay but there's no name? Log "Failed" anyway.
		if (FLAG_SET(gs->dom_stats, DUMP_RESOLVES))
			fprintf(gs->g_outf, "%s\tFailed: %d\n",
				inet_ntoa(*iaddr), WSAGetLastError());


		return(tmphlist); // Returning now leaves the name field null
	}
	
	// Check for a rare error case where gethostbyaddr() returns
	// a non-null string, but the string isn't a meaningful domain
	// name. One such case I saw once in July 2001 returned a domain
	// name with only one component (i.e. only a host part, no non-host
	// part). In this case, consider the lookup to have failed.
	
	if (strchr(hostname->h_name, '.') == NULL)
	{
		// Bad error (there should always be a "." in the resolved
		// domain name, but check in case of program error/failure)
		if (FLAG_SET(gs->dom_stats, DUMP_RESOLVES))
			fprintf(gs->g_outf, "%s\tFailed: 0\n",
				inet_ntoa(*iaddr));

		return(tmphlist); // Returning now leaves the name field null
	}


	// Create space for the newly resolved domain name, copy
	// it over, and return.

	tmphlist->name = (char *)malloc(strlen(hostname->h_name)+1);
	if (tmphlist->name == NULL)
	{
		 fprintf(gs->g_outf, "resolve_clientaddr: malloc failure, quitting.\n");
		 exit(-1);
	}
	strcpy(tmphlist->name, hostname->h_name);

	return(tmphlist);

}


// list_all_domain_names()
//
// Dump the contents of a hostaddrlist, one entry per
// domainname/ipaddr pair. The output is suitable for
// later use by gmodstat as a source of pre-defined
// ipaddr->name mappings.
//

void list_all_domain_names(struct global_names *gn,
							struct global_state *gs)
{
	struct hostaddrlist *hlist = gs->dnslist;
	struct hostaddrlist *tmphlist;
	struct in_addr ta;
	FILE *ipnames_outf;
	char tempname[128];
	
	sprintf(tempname, "ipnames-%s\0", gn->global_outfile);
	fprintf(gs->g_outf, "Dumping ipaddr->domain name mappings to %s\n",
		tempname);
	if ((ipnames_outf = fopen(tempname, "w")) == NULL)
	{
		printf("Opening output file %s failed, aborting.\n",
			tempname);
		fprintf(gs->g_outf, "Opening output file %s failed, aborting.\n",
			tempname);
		exit(-1);
	}	

	// Walk through the hlist (where there's only one entry for
	// each ip address) and print each entry to the output file.

	tmphlist = hlist;
	while (tmphlist != NULL)
	{
		ta.s_addr = tmphlist->ipaddr;

		if (tmphlist->name != NULL)
			fprintf(ipnames_outf, "%s\tName: %s\n",
				inet_ntoa(ta), tmphlist->name);
		else
			fprintf(ipnames_outf, "%s\tFailed: 0\n",
				inet_ntoa(ta));

		tmphlist = tmphlist->next;
	}

	fprintf(ipnames_outf, "\n");
	fclose(ipnames_outf);
	return;
}


//
// dump_dname()
//
// Recursively walk the dname_list, printing out the subdomains
//
// 2nd param is indentation level (no. of spaces)
// 3rd param is total number of entries one level up in
// the hierarchy (so percentages are calculated relative to
// the 3rd param)
// 4th param is a unique integer to place at the front of each
// printed line.
//
// This routine assumes it should dump its output to the
// file pointed to by gs->g_outf2. It is an error for this
// stream pointer to be null.
//
void dump_dname(struct dname_list *ptr,
				int tab,
				int total,
				int mtime,
				struct global_state *gs)
{
	char tabs[128];
	int temptab = tab*2;

	tabs[0] = 0;

	while (temptab > 0)
	{
		strcat(tabs, "  \0");
		temptab--;
	}

	while (ptr != NULL)
	{
		fprintf(gs->g_outf2, "+%dMin Level %01d: %s%s  (%d times, %2.1f%%)",
			mtime/60, tab, tabs, ptr->nfrag, ptr->count,
			100*(float)ptr->count/(float)total);
		if (ptr->leafnum > 0)
			fprintf(gs->g_outf2, " [%d subdomains]\n", ptr->leafnum);
		else
			fprintf(gs->g_outf2, "\n");


		if (ptr->leaf != NULL)
			dump_dname(ptr->leaf, tab+1, ptr->count, mtime, gs);
		ptr = ptr->next;
	}

}



//
// push_domain_name()
//
// Take a single supplied domain name, and add new entry(s)
// to the supplied dname_list tree
//

void push_domain_name(struct global_state *gs,
					  struct dname_list **dlist,
					  char *name)
{
	char *nameptr;
	struct dname_list *tmplist, *prevtmplist;

	// Recursively step back through the supplied domain name
	// and build/add to the dname_list tree of domain name components.
	// Top level is ".com", etc, next level "home.com", etc.

	// Main outer loop steps backwards looking for "." field
	// separators. When one is found, the ascii string to the
	// right of the "." is entered into the dname_list tree

	nameptr = name + strlen(name);
	prevtmplist = NULL;

	while (nameptr >= name)
	{
		if (*nameptr == '.')
		{
			// Found a fragment, now insert into dname_list tree
			if (*dlist == NULL)
			{
				// Need to create an initial dname_list
				// entry for this fqdn fragment

				*dlist = (struct dname_list *)malloc(sizeof(struct dname_list));
				if (*dlist == NULL)
				{
				 fprintf(gs->g_outf, "Malloc failure in push_domain_name(), quitting.\n");
				 exit(-1);
				}

				// Zero the first entry's fields
				memset((char *)*dlist, 0, sizeof(struct dname_list));

				// Allocate space for the name fragment, copy it over
				(*dlist)->nfrag = (char *)malloc(1+strlen(nameptr));
				if ((*dlist)->nfrag == NULL)
				{
				 fprintf(gs->g_outf, "Malloc failure in push_domain_name(), quitting.\n");
				 exit(-1);
				}
				strcpy((*dlist)->nfrag, nameptr);

				// Finally, increment the leaf counter of the parent
				// dname_list entry (except if this entry is already at
				// the top level, where there is no parent)
				if (prevtmplist != NULL)
					prevtmplist->leafnum++;
			}

			// Scan through the existing dname_list (including, perhaps
			// redundantly, the entry we might have just created above).

			tmplist = *dlist;
			while (tmplist != NULL)
			{
				// Use "_stricmp()" so that upper/lower case names
				// are considered equivalent
				if (stricmp(nameptr, tmplist->nfrag) == 0)
					break;
				tmplist = tmplist->next;
			}

			if (tmplist == NULL)
			{
				// Getting to here means need to create a dname_list
				// entry for this fqdn fragment

				if ((tmplist = (struct dname_list *)malloc(sizeof(struct dname_list))) == NULL)
				{
				 fprintf(gs->g_outf, "Malloc failure in push_domain_name(), quitting.\n");
				 exit(-1);
				}

				// Zero the first entry's fields
				memset((char *)tmplist, 0, sizeof(struct dname_list));

				// Allocate space for the name fragment, copy it over
				if ((tmplist->nfrag = (char *)malloc(1+strlen(nameptr))) == NULL)
				{
				 fprintf(gs->g_outf, "Malloc failure in push_domain_name(), quitting.\n");
				 exit(-1);
				}
				strcpy(tmplist->nfrag, nameptr);

				// Insert this new dname_list entry at head of list
				tmplist->next = *dlist;
				*dlist = tmplist;

				// Finally, increment the leaf counter of the parent
				// dname_list entry (except if this entry is already at
				// the top level, where there is no parent)
				if (prevtmplist != NULL)
					prevtmplist->leafnum++;
			}

			// here we have a valid dname_list entry, increment
			// the counter and modify *dlist to now point one
			// level down the tree

			tmplist->count++;			// Incremement entry count of parent
			prevtmplist = tmplist;		// Remember the parent
			dlist = &(tmplist->leaf);	// Move to next child/leaf

		}

		nameptr--;	// Move back one character in supplied fqdn
	}


}


//
// sort_dname()
//
// Recursively walk a dname_list, ordering the list of
// entries at each level in descending sequence of
// "->count" field values, then "->leafnum" values, and
// finally, alphabetical order of (sub)domain names.
//
//
void sort_dname(struct dname_list **ptr)
{
	struct dname_list *tmp, *tmp2, *prevtmp;
	int swap = 1;

	if (*ptr == NULL)
		return;

	// First use a dumb bubble sort on the level pointed
	// to by *ptr

	while (swap == 1)
	{
		tmp = *ptr;
		prevtmp = NULL;
		swap = 0;		// Made no bubble swaps yet
		while (tmp != NULL)
		{
		 if (tmp->next != NULL)
		 {
			 if (tmp->count < tmp->next->count)
			 {
				 // swap if address count is smaller
				swap = 1;

				if (prevtmp == NULL)
				{
					tmp2 = tmp->next->next;
					*ptr = tmp->next;
					tmp->next = tmp2;
					(*ptr)->next = tmp;
				}
				else
				{
					tmp2 = tmp->next->next;
					prevtmp->next = tmp->next;
					tmp->next = tmp2;
					(prevtmp->next)->next = tmp;
				}
			 }
			 else if ((tmp->count == tmp->next->count) &&
				 (tmp->leafnum < tmp->next->leafnum))
			 {
				 // or swap if counts match, but number of subdomains
				 // is smaller
				swap = 1;

				if (prevtmp == NULL)
				{
					tmp2 = tmp->next->next;
					*ptr = tmp->next;
					tmp->next = tmp2;
					(*ptr)->next = tmp;
				}
				else
				{
					tmp2 = tmp->next->next;
					prevtmp->next = tmp->next;
					tmp->next = tmp2;
					(prevtmp->next)->next = tmp;
				}
			 }
			 else if ((tmp->count == tmp->next->count) &&
				 (tmp->leafnum == tmp->next->leafnum) &&
				 (stricmp(tmp->nfrag, tmp->next->nfrag) > 0))
			 {
				 // or swap if counts match, number of subdomains
				 // matches, but the domain name itself is smaller
				 // (will sort in ascending alphabetical order if we
				 // get to here)
				swap = 1;

				if (prevtmp == NULL)
				{
					tmp2 = tmp->next->next;
					*ptr = tmp->next;
					tmp->next = tmp2;
					(*ptr)->next = tmp;
				}
				else
				{
					tmp2 = tmp->next->next;
					prevtmp->next = tmp->next;
					tmp->next = tmp2;
					(prevtmp->next)->next = tmp;
				}

			 }

		 }
		 prevtmp = tmp;
		 tmp = tmp->next;
		}
		// If we get to here and 'swap' is zero, no bubble swaps
		// were performed and the sorting is complete
	}


	// Finally, walk the list just sorted and recursively
	// sort the sublevels
	
	tmp = *ptr;
	while (tmp != NULL)
	{
		sort_dname(&(tmp->leaf));
		tmp = tmp->next;
	}

	return;
}



//
// normalize_dname()
//
// Recursively walk a dname_list, normalizing the "->count"
// fields to adjust for cases where the last leaf on every
// branch SHOULD have appeared only once but was counted
// multiple times.
//
//
int normalize_dname(struct dname_list **ptr)
{
	int decrement;
	struct dname_list *tmpd;

	decrement = 0;
	tmpd = *ptr;
	while (tmpd != NULL)
	{
		int dec2;

		if (tmpd->leaf == NULL)
			dec2 = (tmpd->count - 1);
		else
			dec2 = normalize_dname(&(tmpd->leaf));

		decrement += dec2;
		tmpd->count -= dec2;

		tmpd = tmpd->next;
	}

	return(decrement);
}



//
// count_home_domains()
//
// Walk through a list of top and mid level domain name fragments,
// and count the number of IP addresses whose names resolved to
// something under each fragment. (e.g. if the fragments are a list of
// high level domains believed to represent home users, then the
// result provides a lower bound on the number of home users seen
// in the logfile.)
//

int count_home_domains(struct dname_list *nametop,
					   struct hostaddrlist *hlist,
					   int global_stats)
{
	struct dname_list *tmpdname;
	int numtotal, numhome;
	struct hostaddrlist *tmphlist;

	if (hlist == NULL)
		return(0);	// No fragments specified, nothing to count

	if (nametop == NULL)
		return(0);	// No domain names, nothing to count either

	// Now walk through the list pointed to by hlist, and walk through
	// the tree at nametop to find each fragment in the hlist list.
	// Then count the number of branch leaf nodes under matching entries.

	tmphlist = hlist;
	numtotal = nametop->count;
	numhome = 0;

	while (tmphlist != NULL)
	{
		// Walk the domain name tree for a match to tmphlist->name
		tmpdname = find_dname_match(nametop, tmphlist->name);

		if (tmpdname != NULL)
			numhome += tmpdname->count;	// Add the number of leaf nodes under this fragment

		tmphlist = tmphlist->next;
	}
	return(numhome);


}

// find_dname_match()
//
// Search the pre-existing domain name tree for a matching entry
// to a particular domain name suffix fragment (e.g. ".sfba.home.com")

struct dname_list *find_dname_match(struct dname_list *nametop, char *name)
{
	char *nameptr;
	struct dname_list **dlist, *tmplist;

	// Main outer loop steps backwards looking for "." field
	// separators. When one is found, the dname_list is searched
	// for a match to the ascii string to the right of the "."
	// Then we go one level down if there's still characters left
	// in the 'name' parameter. Return a pointer to the last dname_list
	// entry that matched.

	if ((name == NULL) || (nametop == NULL))
		return(NULL);

	nameptr = name + strlen(name);
	dlist = &nametop;

	while (nameptr >= name)
	{
		if (*nameptr == '.')
		{
			// Search for this component of the fragment in the current
			// level of the dname_list tree

			tmplist = *dlist;
			while (tmplist != NULL)
			{
				// Use "stricmp()" so that upper/lower case names
				// are considered equivalent
				if (stricmp(nameptr, tmplist->nfrag) == 0)
					break;
				tmplist = tmplist->next;
			}

			if (tmplist == NULL)
			{
				// Getting to here means there's no match at this level
				// in the dname tree for the contents of 'name'.
				// Might as well return NULL

				return(NULL);
			}

			// here we have a valid dname_list entry that matches
			// the 'name' text to the right of 'nameptr'. Move
			// down a level and continue searching.

			dlist = &(tmplist->leaf);	// Move to next child/leaf
		}
		nameptr--;	// Move back one character in supplied fqdn
	}

	// Getting to here means we've exhausted the components of the
	// supplied 'name' (moving right to left) and are still pointing
	// to the last level in the dname_list tree that matched. Return
	// pointer to this level.

	return(tmplist);

}


// create_hlist()
//
// Build a memory-resident ipaddr:hostname mapping
// list from the ascii text in file pointed to by "rfile"
// Ignore redundant entries in the ascii file (e.g. same ipaddr
// turning up multiple times results in only one memory-resident
// mapping).
//

struct hostaddrlist *create_hlist(struct global_state *gs,
								FILE *rfile)
{
	struct hostaddrlist *hlist = NULL;
	struct hostaddrlist *tmphlist;
	char dname[256];
	char ipnum[32];
	char token[32];
	unsigned int tmpaddr;
	int entries = 0;
	int redundant = 0;

	if (rfile == NULL)
		return (NULL);

	// If rfile is non-null it points
	// to a file containing ipaddr:name mappings.
	// Read in that file to a local hostaddrlist list, which
	// will be used instead of gethostbyaddr() to resolve
	// ipaddr to domain names.

	// Each entry in the rfile is a text line containing
	// three components: w.x.y.z  <token>  <result>
	// (w.x.y.z is the numeric ip address. if <token> is
	// "Name:" then <result> is the domain name. if <token>
	// is "Failed:" then <result> is a numeric error code
	// so we create a NULL name mapped to that ipaddress.

	while (fscanf(rfile, "%s %s %s\n", ipnum, token, dname) > 0)
	{
		// printf("%05d   %s  %s \n", entries, ipnum, dname);
		
		// First scan the existing entries to see if this new
		// entry is redundant

		tmpaddr = inet_addr(ipnum);	// Munge ascii address into a 4 byte
									// unsigned int in network-byte order
									// (it will be stored in this fashion in
									// the hostaddrlist entry)
		if (FLAG_SET(gs->dom_stats, CHK_HLIST_UNIQ))
		{
    		tmphlist = hlist;
    		while (tmphlist != NULL)
    		{
    			if (tmphlist->ipaddr == tmpaddr)
    				break;

    			tmphlist = tmphlist->next;
    		}
		}
		else
			tmphlist = NULL;

		if (tmphlist == NULL)
		{
			// Getting to here means this is a new ipaddr,
			// so create a new hostaddrlist entry

			if ((tmphlist =
				 (struct hostaddrlist *)malloc(sizeof(struct hostaddrlist)))
				 == NULL)
			{
			  fprintf(gs->g_outf, "create_hlist(): malloc failure, exiting.\n");
			  exit(-1);
			}

			memset((char *)tmphlist, 0, sizeof(struct hostaddrlist));

			if (strcmp(token, "Name:\0") == 0)
			{
			 if ((tmphlist->name = (char *)malloc(strlen(dname)+1)) == NULL)
			 {
			  fprintf(gs->g_outf, "create_hlist(): malloc failure, exiting.\n");
			  exit(-1);
			 }

			 strcpy(tmphlist->name, dname);
			}
			else
				tmphlist->name = NULL;

			tmphlist->ipaddr = tmpaddr;

			// Put this new entry at head of hlist

			tmphlist->next = hlist;
			hlist = tmphlist;

			entries++;
		}
		else
		{
			redundant++;

			if ((tmphlist->name == NULL) &&
				(strcmp(token, "Name:\0") == 0))
			{
					fprintf(gs->g_outf,
					"create_hlist(): Redundant entry: %s %s replaces previous FAILED status.\n",
					ipnum, dname);

					if ((tmphlist->name = (char *)malloc(strlen(dname))) == NULL)
					{
					 fprintf(gs->g_outf,
						 "create_hlist(): malloc failure, exiting.\n");
					 exit(-1);
					}
					strcpy(tmphlist->name, dname);

			}

			if (tmphlist->name != NULL)
			{
				if (stricmp(tmphlist->name, dname) != 0)
					fprintf(gs->g_outf,
					"create_hlist(): Redundant entry: %s %s was already %s, left unchanged.\n",
					ipnum, dname, tmphlist->name);
			}
		}

	}

	printf("%d unique entries (%d redundant entries).\n",
		entries, redundant);
	fprintf(gs->g_outf, "%d unique entries (%d redundant entries).\n",
		entries, redundant);

	return(hlist);
}



//
// add_to_hlist()
//
// Add to a memory-resident ipaddr:hostname mapping
// list from the ascii text in file pointed to by "rfile"
// Ignore redundant entries in the ascii file (e.g. same ipaddr
// turning up multiple times results in only one memory-resident
// mapping) unless the redudant entry provides more information
// (e.g. previous one was "Failed:" but now we have an address)
//
// File format is assued to be a sequence of ANSWER lines
// from the unix command "dig" executed on a sequence of
// ipaddresses that previously "Failed" gethostbyaddr().
//
void add_to_hlist(struct global_state *gs,
					FILE *file,
					struct hostaddrlist **hlist)
{
	struct hostaddrlist *tmphlist;
	char dname[256];
	char ipnum[64], tmpipnum[64];
	char time[32];
	char iclass[32];
	char type[32];
	unsigned int tmpaddr;
	int entries = 0;
	int redundant = 0;
	int w,x,y,z;

	if (file == NULL)
		return;

	// If file is non-null it points
	// to a file containing ipaddr:name mappings.
	// Read in that file to a local data tree, which
	// will be used instead of gethostbyaddr() to resolve
	// names.

	// Build hostaddrlist table
	// Each entry in the file is a text line containing
	// five components: "z.y.x.w.in-addr.arpa <time> IN PTR <domainname>"
	// (w.x.y.z is the numeric ip address)

	while (fscanf(file, "%s %s %s %s %[^\n\r]", ipnum, time, iclass, type, dname) > 0)
	{

		// Extract the ip number for which this domain name applies
		sscanf(ipnum, "%d.%d.%d.%d.in-addr.arpa.", &w, &x, &y, &z);
		sprintf(tmpipnum, "%d.%d.%d.%d\0", z, y, x, w);
		tmpaddr = inet_addr(tmpipnum);	// Munge ascii address into a 4 byte
										// unsigned int in network-byte order
										// (it will be stored in this fashion in
										// the hostaddrlist entry)

		// Scan the existing entries to see if this new
		// entry is redundant
		tmphlist = *hlist;
		while (tmphlist != NULL)
		{
			if (tmphlist->ipaddr == tmpaddr)
				break;

			tmphlist = tmphlist->next;
		}

		if (tmphlist == NULL)
		{
			// Getting to here means this is a new ipaddr,
			// so create a new hostaddrlist entry

			if ((tmphlist =
				 (struct hostaddrlist *)malloc(sizeof(struct hostaddrlist)))
				 == NULL)
			{
			  fprintf(gs->g_outf, "add_to_hlist(): malloc failure, exiting.\n");
			  exit(-1);
			}

			memset((char *)tmphlist, 0, sizeof(struct hostaddrlist));

			if ((tmphlist->name = (char *)malloc(strlen(dname))) == NULL)
			{
			  fprintf(gs->g_outf, "add_to_hlist(): malloc failure, exiting.\n");
			  exit(-1);
			}

			strcpy(tmphlist->name, dname);

			tmphlist->ipaddr = tmpaddr;

			// Put this new entry at head of hlist

			tmphlist->next = *hlist;
			*hlist = tmphlist;

			entries++;
		}
		else
		{
			redundant++;

			if (tmphlist->name == NULL)
			{
					fprintf(gs->g_outf,
					"add_to_hlist(): %s %s replaces previous FAILED status.\n",
					tmpipnum, dname);

					if ((tmphlist->name = (char *)malloc(strlen(dname))) == NULL)
					{
					 fprintf(gs->g_outf,
						 "add_to_hlist(): malloc failure, exiting.\n");
					 exit(-1);
					}
					strcpy(tmphlist->name, dname);
			}

			if (tmphlist->name != NULL)
			{
				if (stricmp(tmphlist->name, dname) != 0)
					fprintf(gs->g_outf,
					"add_to_hlist(): %s %s left unchanged as %s.\n",
					ipnum, dname, tmphlist->name);
			}
		}

	}

	printf("%d unique entries (%d redundant entries).\n",
		entries, redundant);
	fprintf(gs->g_outf, "add_to_hlist() saw %d unique entries and %d redundant entries.\n\n",
		entries, redundant);

	return;
}




