// gmodstat_graphstats.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Routines for dumping various per-game statistics such
// as aggregate playing time.
//

#include "stdafx.h"
#include "gmodstat.h"



// create_ptime_histo_onegame()
//
// Takes pointer to a particular seen_game entry, and adds the
// end-of-game total played time of each player to a histogram
//
// (Ignores bots)
//

void create_ptime_histo_onegame(int hour,
					  struct seen_games *glist,
					  struct global_state *gs)
{
	struct	seen_player *plist;
	int yesflag;

	((gs->playt_histo[hour]).gamecnt)++;

	plist = glist->c_all;	// Step through all the players seen in this game
	while (plist != NULL)
	{
		// If there's one or more IP addresses this player entry is NOT a bot

		if (plist->c_addr != NULL)
		{
			// If ONLY_UNRESOLVED is set then only count this player's
			// contribution if its addresses were not resolved to a
			// domain name.
			//
			// Otherwise, let is_in_region() decide if the resolvable
			// domain names are in the desired set of domains.

			yesflag = 0;

			if ((FLAG_SET(gs->graph_stats, ONLY_UNRESOLVED)) &&
				(is_in_unresolved(plist, gs)))
					yesflag = 1;
			if ((FLAG_UNSET(gs->graph_stats, ONLY_UNRESOLVED)) &&
				(is_in_region(plist, gs)))
					yesflag = 1;

			if (yesflag == 1)
			{
				// Add this player's totaltime to the histogram

				(gs->playt_histo[hour]).playtime += (long)plist->totaltime;
				(gs->ptime_total) += (long)plist->totaltime;
			}
		}

		plist = plist->next;
	}

	return;

}


//
// create_selected_ptime_histo()
//
// Walks through the seen_games list, and creates
// a histogram of total played time vs time for clients
// who fall under specified conditions (eg. a specific
// domain, multiple domains, or unresolved addresses)
//
// The histogram is dumped to gs->g_outf2.
//
// If HOUR_OF_DAY flag is set, the histogram buckets
// are 30 minutes wide from 0 to 23.5 hours. Otherwise,
// histogram buckets are 60 min wide, from 0 to 168 hours
// (i.e. a week)
//
// This routine is not called directly, rather it is
// called from create_ptime_histo() which sets up the
// correct output files gs->g_outf2 for each iteration.
//

void create_selected_ptime_histo(struct global_state *gs,
								struct global_names *gn)
{
	struct  seen_games *glist;
	int week;
	char separator;
	int arraysize;
	int arrayptr;
	int arraylimit;
	int divider;
	
	if (gs->g_outf2 == NULL)
	{
		printf("create_selected_ptime_histo(): gs->g_outf2 not set!?\n");
		fprintf(gs->g_outf, "create_selected_ptime_histo(): gs->g_outf2 not set!?\n");
		return;
	}
	
   	if (FLAG_SET(gs->graph_stats, HOUR_OF_DAY))
   	{
   		divider = 2;
   		arraylimit = 48;	// 48 30-min intervals in a day
   	}
   	else
   	{
   		divider = 1;
   		arraylimit = 168;	// 168 hours in a week
   	}

   	arraysize = (arraylimit) * sizeof(struct playt);
	
	// Games are stored last->first hanging off glist. Walk the
	// list until we find the last entry. (We then walk backwards
	// up the list, which is chronologically forward in time)

	if ((glist = gs->gamelist) == NULL)
		return;
	while (glist->next != NULL)
		glist = glist->next;

	printf("Generating played time histogram\n");
	week = 0;

	if (FLAG_SET(gs->graph_stats, XGRAPH_OUT))
		separator = ' ';
	else
		separator = ',';

	// allocate space for the array of 'struct playt'

	gs->ptime_total = 0;
	if ((gs->playt_histo = (struct playt *)malloc(arraysize))
		 == NULL)
	{
		printf("create_selected_ptime_histo(): malloc failed, exiting\n");
		fprintf(gs->g_outf, "create_selected_ptime_histo(): malloc failed, exiting\n");
		exit(-1);
	}
	else
	{
		// Zero all the entries in the array
		memset((char *)(gs->playt_histo), 0, arraysize);
	}

	// Begin stepping through the game list

	while (glist != NULL)
	{
		int gamelen = glist->t_end - glist->t_start;
		long realstart = glist->t_base + glist->t_start;	//Absolute time stamp
		long realmiddle = realstart + (gamelen/2);
		struct tm *tmnow;
		int hour;

		// The time of this game is calculated from the mid-point
		// of the game ('realmiddle' is halfway between absolute start
		// and end times)
		
		tmnow = localtime(&realmiddle);
		if (FLAG_SET(gs->graph_stats, HOUR_OF_DAY))
			hour = 2*tmnow->tm_hour + (tmnow->tm_min / 30);
		else
			hour = 24*(tmnow->tm_wday) + tmnow->tm_hour;

		create_ptime_histo_onegame(hour, glist, gs);

		glist = glist->prev;
	}

	
	// Finished walking backwards through the seen_game list
	// Now dump appropriate results

	arrayptr = 0;
	while (arrayptr < arraylimit)
	{
		fprintf(gs->g_outf2, "%f %c %7.2f\n", (arrayptr/(float)24)/divider, separator,
			100.0*((gs->playt_histo)[arrayptr].playtime)/(float)gs->ptime_total);

		arrayptr++;
	}

	// Clean up the array space, in case we're called again
	if (gs->playt_histo != NULL)
		free(gs->playt_histo);

	return;
}


//
// create_ptime_histo()
//
// Calls create_selected_ptime_histo() once or multiple
// times to generate one or more histograms of played time
// vs time. If GRAPH_MANY flag is set then histograms are
// generated for clients in each of the domains listed
// in gs->fragementlist. Otherwise, a single histogram is
// generated for clients who are members of any of the
// domains listed in gs->fragementlist (or optionally, all
// domains).
//

void create_ptime_histo(struct global_state *gs,
								struct global_names *gn)
{
   	struct hostaddrlist *old_frglist;
   	struct hostaddrlist *nextfrg;
   	struct hostaddrlist tmpfragment;
   	char outfname[64];

	// Sanity checks
	
	if (gs->dnslist == NULL)
	{
		printf("create_ptime_histo: there's no dns mapping list, exiting.\n");
		fprintf(gs->g_outf, "create_ptime_histo: there's no dns mapping list, exiting.\n");
		return;
	}
	
	// If we're supposed to be doing histograms only for some specific
	// domains (either individually or as a group) then flag error if there's
	// nothing hanging off gs->fragmentlist
	
	if ((gs->fragmentlist == NULL) &&
		(FLAG_SET(gs->graph_stats, GRAPH_MANY) || FLAG_SET(gs->graph_stats, INCLUDE_REGIONS)))
	{
		printf("create_ptime_histo: no target domain(s) specified, exiting.\n");
		fprintf(gs->g_outf, "create_ptime_histo: no target domain(s) specified, exiting.\n");
		return;
	}

	// If GRAPH_MANY is set, begin to loop through all entries in
	// gs->fragmentlist, doing histograms for each one. Otherwise,
	// do one histogram (that will include clients from the domains
	// gs->fragementlist, or all clients)
	
	if (FLAG_SET(gs->graph_stats, GRAPH_MANY))
	{
    	// Decouple the original gs->fragmentlist, and replace it with a
    	// single entry.

    	old_frglist = gs->fragmentlist;
    	nextfrg = old_frglist;
    	gs->fragmentlist = &tmpfragment;
    	memset((char *)gs->fragmentlist, 0, sizeof(struct hostaddrlist));

    	while (nextfrg != NULL)
    	{
    		// Establish new output file
    		sprintf(outfname,"PT%s-ptimehisto.txt\0", nextfrg->name);
    		gs->g_outf2 = fopen(outfname, "w");
    		if (gs->g_outf2 == NULL)
    		{
    			printf("Opening output file %s failed, aborting.\n",
    				outfname);
    			exit(1);
    		}
    		
    		printf("Creating playedtime histogram for domain %s (file %s)\n",
    					nextfrg->name, outfname);
    		fprintf(gs->g_outf, "Creating playedtime histogram for domain %s (file %s)\n",
    					nextfrg->name, outfname);

    		// Modify gs->fragmentlist appropriately (adjust the name pointer
    		// to point to the real name in nextfrg->name, and leave the next
    		// pointer NULL so it looks like a fragmentlist with only one entry).

    		gs->fragmentlist->name = nextfrg->name;

    		// Call create_selected_ptime_histo() to parse the seen_game list with
    		// a focus on the single entry in gs->fragmentlist

    		create_selected_ptime_histo(gs, gn);
    		fclose(gs->g_outf2);
    		
    		nextfrg = nextfrg->next;
    	}

    	// Finished, clean up and return

    	gs->fragmentlist = old_frglist;
	}
	else
	{
		// This is the single-histogram case. Open one file, then
		// call create_selected_ptime_histo(). But first, list
		// the specific domains (if applicable).
		
    	if (FLAG_UNSET(gs->graph_stats,ONLY_UNRESOLVED) &&
    			(gs->fragmentlist != NULL))
    	{
    		printf("Creating playedtime histogram for multiple domains\n");
    		fprintf(gs->g_outf, "Creating playedtime histogram for multiple domains\n");
    		if (FLAG_SET(gs->graph_stats, INCLUDE_REGIONS))
    		{
    			printf("Include list: ");
    			fprintf(gs->g_outf, "Include list: ");
    		}
    		if (FLAG_SET(gs->graph_stats, EXCLUDE_REGIONS))
    		{
    			printf("Exclude list: ");
    			fprintf(gs->g_outf, "Exclude list: ");
    		}
    		if (FLAG_SET(gs->graph_stats, EXCLUDE_REGIONS) ||
    			FLAG_SET(gs->graph_stats, INCLUDE_REGIONS))
    		{
    			nextfrg = gs->fragmentlist;
    			while (nextfrg != NULL)
    			{
    				printf("%s", nextfrg->name);
    				fprintf(gs->g_outf, "%s", nextfrg->name);
    				nextfrg = nextfrg->next;
    				if (nextfrg != NULL)
    				{
    					printf(", ");
    					fprintf(gs->g_outf, ", ");
    				}
    			}
    			printf("\n");
    			fprintf(gs->g_outf, "\n\n");
    		}
    	}
    	else
    	{
       		printf("Creating playedtime histogram for all domains\n");
    		fprintf(gs->g_outf, "Creating playedtime histogram for all domains\n");
	   	}

	   	// Open output file
   		sprintf(outfname,"PT0-ptimehisto.txt\0");
   		gs->g_outf2 = fopen(outfname, "w");
   		if (gs->g_outf2 == NULL)
   		{
   			printf("Opening output file %s failed, aborting.\n",
   				outfname);
   			exit(1);
   		}
   		else
   			fprintf(gs->g_outf, "Dumping histogram to %s\n", outfname);

   		create_selected_ptime_histo(gs, gn);
    	fclose(gs->g_outf2);

	}
	
	
	return;
}




//
// is_in_region()
//
// Looks up the domain names of the ip addresses from which
// a specified seen_player entry was seen from, and returns
// 1 if any were from the 'homedomain_file' list,
// otherwise 0.
//
// NOTE: Using the logfiles from Feb.2001 to end of April 2001
// I saw 698 player entries where the player had multiple IP
// address/port pairs in a single game. However, NONE of these instances
// of multiple IP address/port pairs represented a different IP
// address. Thus we make the following SIMPLIFICATION: We determine
// whether the player's ipaddress falls under the appropriate domain
// by looking ONLY at the FIRST ip address hanging off the seen_player
// entry.
//
// NOTE2: The above simplification is limited to *WITHIN* any
// given game. Over a period of many games, playernames are
// seen with different ipaddresses too (although often related
// by common domain name suffix). But for our purposes here,
// ipaddr doesn't change within a game.
//

int is_in_region(struct seen_player *plist, struct global_state *gs)
{

	struct cid_ipaddr *tmp_cip;
	int nameflag;
	unsigned int x;
	struct hostaddrlist *tmphlist;	// Pointer into the ipaddr:name map
	struct hostaddrlist *tmpvlist;	// Pointer into valid domains list

	// If we're not including only certain regions, always return TRUE
	// (since we're clearly including everyone)

	if (FLAG_UNSET(gs->graph_stats, INCLUDE_REGIONS))
		return(1);

	// Resolve first ipaddr in the cip_addr list, and return TRUE
	// it falls under a domain listed in the gs->fragmentlist table
	// (populated by 'homedomain_file' option)

	tmp_cip = plist->c_addr;
	if (tmp_cip == NULL)
		return(0);	// This error case shouldn't occur, but if it does
					// we'll treat the player as NOT belonging to the
					// set of home domains

	// A simple optimization - if we've already gone through the
	// seen_game table resolving player ipaddr to domain names,
	// skip the name lookup on subsequent passes

	if (tmp_cip->resolved == 0)
	{
	// Attempt to resolve the first ipaddr into a domain name
	// (we assume any other entries are for different ports, which
	// dont concern us during this phase of the analysis)

		x = (tmp_cip->iaddr).s_addr;
		tmphlist = gs->dnslist;
		nameflag = 0;
		while ((tmphlist != NULL) && (nameflag == 0))
		{
			if (tmphlist->ipaddr == x)
				nameflag = 1;
			else
				tmphlist = tmphlist->next;
		}

		// Okay, flag that we've attempted a name resolution
		// for this entry, and tmp_cip->name is NULL if the
		// name couldn't be resolved (for whatever reason)

		tmp_cip->resolved = 1;
		tmp_cip->name = NULL;
		if (nameflag == 1)
			tmp_cip->name = tmphlist->name;
	}

	// If we actually found a real domain name,
	// see if it falls under home domains

	if (tmp_cip->name != NULL)
	{
		// Step through the valid domains list, and see
		// if any of the domain name suffixes match the
		// current ipaddresses 'name'. If so, return TRUE
		// immediately.

		tmpvlist = gs->fragmentlist;
		while (tmpvlist != NULL)
		{
			if (is_name_in_dname(tmp_cip->name, tmpvlist->name))
			{
				return (1);
			}
			tmpvlist = tmpvlist->next;

		}
	}

	// Falling through to here means either no matching ipaddr
	// was found, or the matched address had no domain name falling
	// under the 'home domains' list

	return(0);
}

//
// Returns 1 if the entire contents of 2nd parameter
// match exactly the right hand suffix of the first parameter
// Otherwise returns 0


int is_name_in_dname(char *name, char *dname)
{
	int index;

	// If 2nd parameter is longer, there's clearly no match
	// so return 0
	if (strlen(name) < strlen(dname))
		return(0);

	// If both parameters are same length and they dont match,
	// return 0
	if ((strlen(name) == strlen(dname)) &&
		(stricmp(name, dname) != 0))
		return(0);

	index = strlen(name) - strlen(dname);

	if (stricmp(name+index, dname) != 0)
		return(0);

	return(1);
}




// is_in_unresolved()
//
// Looks up the domain names of the ip addresses from which
// a specified seen_player entry was seen from, and returns
// 1 if the domain name was unresolved, or 0 if it was.
//
//
// NOTE: Using the logfiles from Feb.2001 to end of April 2001
// I saw 698 player entries where the player had multiple IP
// address/port pairs in a single game. However, NONE of these instances
// of multiple IP address/port pairs represented a different IP
// address. Thus we make the following SIMPLIFICATION: We determine
// whether the player's ipaddress falls under the appropriate domain
// by looking ONLY at the FIRST ip address hanging off the seen_player
// entry.
//

int is_in_unresolved(struct seen_player *plist, struct global_state *gs)
{

	struct cid_ipaddr *tmp_cip;
	int nameflag;
	unsigned int x;
	struct hostaddrlist *tmphlist;	// Pointer into the ipaddr:name map

	// Resolve first ipaddr in the cip_addr list, and return TRUE
	// if it cannot be resolved to an ascii domain name.

	tmp_cip = plist->c_addr;
	if (tmp_cip == NULL)
		return(0);	// This error case shouldn't occur, but if it does
					// we'll treat the player as NOT belonging to the
					// set of unresolved ipaddresses

	// A simple optimization - if we've already gone through the
	// seen_game table resolving player ipaddr to domain names,
	// skip the name lookup on subsequent passes

	if (tmp_cip->resolved == 0)
	{
	// Attempt to resolve the first ipaddr into a domain name
	// (we assume any other entries are for different ports, which
	// dont concern us during this phase of the analysis)

		x = (tmp_cip->iaddr).s_addr;
		tmphlist = gs->dnslist;
		nameflag = 0;
		while ((tmphlist != NULL) && (nameflag == 0))
		{
			if (tmphlist->ipaddr == x)
				nameflag = 1;
			else
				tmphlist = tmphlist->next;
		}

		// Okay, flag that we've attempted a name resolution
		// for this entry, and tmp_cip->name is NULL if the
		// name couldn't be resolved (for whatever reason)

		tmp_cip->resolved = 1;
		tmp_cip->name = NULL;
		if (nameflag == 1)
			tmp_cip->name = tmphlist->name;
	}

	if (tmphlist->name == NULL)
			return(1);

	return(0);
}

//
// For a given seen_player entry, flag if the player has multiple
// addresses, list the addresses, and identify if the addresses
// are different or all the same. Use this routine just for debugging
// and/or verifying validity of the "one addr per player per game" assumption
//

void dump_check_multiple_addr(struct seen_player *plist, struct global_state *gs)
{

	struct cid_ipaddr *tmp_cip;
	int nameflag, seen_unresolved, seen_resolved;
	int numentries, numdiffentries;
	unsigned int x, prev_addr;
	struct hostaddrlist *tmphlist;	// Pointer into the ipaddr:name map

	seen_unresolved = 0;
	seen_resolved = 0;
	numentries = 0;
	numdiffentries = -1;
	prev_addr = 0;

	tmp_cip = plist->c_addr;
	while (tmp_cip != NULL)
	{
		numentries++;
		if (prev_addr != (tmp_cip->iaddr).s_addr)
			numdiffentries++;

		prev_addr = (tmp_cip->iaddr).s_addr;
		tmp_cip = tmp_cip->next;
	}

	if (numentries == 1)
		return;

	// Okay, so there's multiple addresses. Lets resolve them
	// if possible, and print them out, if any of them differ from
	// each other.

	printf("is_in_unresolved: %d addresses, %d different, for player %s\n",
		numentries, numdiffentries, plist->playername);
	fprintf(gs->g_outf, "is_in_unresolved: %d addresses, %d different, for player %s\n",
		numentries, numdiffentries, plist->playername);

	if (numdiffentries == 0)
		return;

	tmp_cip = plist->c_addr;
	while (tmp_cip != NULL)
	{

		// Attempt to resolve the ipaddr into a domain name

		x = (tmp_cip->iaddr).s_addr;
		tmphlist = gs->dnslist;
		nameflag = 0;
		while ((tmphlist != NULL) && (nameflag == 0))
		{
			if (tmphlist->ipaddr == x)
			{
				nameflag = 1;
			}
			else
			{
				tmphlist = tmphlist->next;
			}
		}

		if (nameflag == 0)
		{
			printf("is_in_unresolved: Unresolved %s\n", inet_ntoa(tmp_cip->iaddr));
			fprintf(gs->g_outf, "is_in_unresolved: Unresolved %s\n",
					inet_ntoa(tmp_cip->iaddr));
			seen_unresolved |= 1;	// There's no entry, definitely unresolved
		}

		if (nameflag == 1)
		{
			// There's an entry, but check for null name.

			if (tmphlist->name == NULL)
			{
				printf("is_in_unresolved: Unresolved %s\n", inet_ntoa(tmp_cip->iaddr));
				fprintf(gs->g_outf, "is_in_unresolved: Unresolved %s\n",
						inet_ntoa(tmp_cip->iaddr));
				seen_unresolved |= 1;
			}
			else
			{
				printf("is_in_unresolved: %s to %s\n",
						inet_ntoa(tmp_cip->iaddr), tmphlist->name);
				fprintf(gs->g_outf, "is_in_unresolved: %s to %s\n",
						inet_ntoa(tmp_cip->iaddr), tmphlist->name);
				seen_resolved |= 1;
			}
		}

		tmp_cip = tmp_cip->next;
		numentries++;
	}


	if ((seen_resolved == 1) && (seen_unresolved == 1))
	{
		printf("is_in_unresolved: Seen both resolved and unresolved for player %s\n",
		plist->playername);
		fprintf(gs->g_outf, "is_in_unresolved: Seen both resolved and unresolved for player %s\n",
		plist->playername);
	}

	return;
}



// is_right_client()
//
// Returns 1 if the seen_player entry has a playername
// matching gs->playername, and ipaddr suffix matching
// gs->playersuffix. Otherwise returns 0.

int is_right_client(struct seen_player *seenp,
					struct global_state *gs)
{
	struct cid_ipaddr *tmp_cip;
	char *fqdn;
	char *suffix;
	char dummyname[64];
	unsigned int x;
	struct hostaddrlist *tmphlist;

	// Exit quickly if the ASCII names dont match

	if (strcmp(seenp->playername, gs->playername) != 0)
		return(0);

	// Resolve the first ip address into a domain name, then
	// compare to the supplied gs->playersuffix.

	tmp_cip = seenp->c_addr;
	if (tmp_cip == NULL)
		return(0);	// This error case shouldn't occur, but if it does
					// then this entry is NOT the right client


	// A simple optimization - if we've already gone through the
	// seen_game table resolving player ipaddr to domain names,
	// skip the name lookup on subsequent passes (a hold over from
	// earlier versions of gmodstat where ipaddr translations might be
	// deferred to later)

	x = (tmp_cip->iaddr).s_addr;
	if (tmp_cip->resolved == 0)
	{
	// Attempt to resolve the first ipaddr into a domain name
	// (we assume any other entries are for different ports, which
	// dont concern us during this phase of the analysis)

		//x = (tmp_cip->iaddr).s_addr;
		tmphlist = gs->dnslist;
		while (tmphlist != NULL)
		{
			if (tmphlist->ipaddr == x)
				break;
			else
				tmphlist = tmphlist->next;
		}

		// Okay, flag that we've attempted a name resolution
		// for this entry, and tmp_cip->name is NULL if the
		// name couldn't be resolved (for whatever reason)

		tmp_cip->resolved = 1;
		tmp_cip->name = NULL;

		if (tmphlist != NULL)
		{
			tmp_cip->name = tmphlist->name;
			tmp_cip->haddr = tmphlist;
		}
		else
		{
			// This situation is worth noting, since it suggests the
			// gs->dnslist is out of date (there are ipaddr in the logfiles
			// that have no mapping in the hosts_file and we're not re-resolving
			// from gethostbyaddr() )
			fprintf(gs->g_outf, "is_right_client: gs->dnslist contains no entry for %s\n",
					inet_ntoa(tmp_cip->iaddr));
		}
	}

	fqdn = tmp_cip->name;

	// If the ASCII name is null, created a fake domain suffix
	// based on the digits in the ip address. The fake suffix
	// is ".w-x-y-z.unresolved" for ip address w.x.y.z
	// Otherwise suffix is everything to the right of the
	// first "." found in the domain name.

	if (fqdn == NULL)
	{
		if (gs->faked_suffix == 4)
			sprintf(dummyname, ".%d-%d-%d-%d.unresolved\0",
				x & 0xFF, (x >> 8)&0xFF, (x >> 16)&0xFF, (x >> 24)&0xFF);
		else if (gs->faked_suffix == 3)
			sprintf(dummyname, ".%d-%d-%d.unresolved\0",
				x & 0xFF, (x >> 8)&0xFF, (x >> 16)&0xFF);
		else if (gs->faked_suffix == 2)
			sprintf(dummyname, ".%d-%d.unresolved\0",
				x & 0xFF, (x >> 8)&0xFF);
		else if (gs->faked_suffix == 1)
			sprintf(dummyname, ".%d.unresolved\0",
				x & 0xFF);
		else
		{
			fprintf(gs->g_outf, "is_right_client: Error, gs->faked_suffix is %d. Exiting.\n",
				gs->faked_suffix);
			exit (-1);
		}
		suffix = dummyname;
	}
	else
	{
		// Domain name's suffix (non-hostname part) is to the right
		// of the first "."

		if ((suffix = strchr(fqdn, '.')) == NULL)
		{
			// Bad error (there should always be a "." in the resolved
			// domain name, but check in case of program error/failure)
			fprintf(gs->g_outf, "is_right_client: Error, there's no \".\" in %s Exiting.\n",
				fqdn);
			exit (-1);
		}
	}

	// Compare the found (or faked) suffix against the supplied
	// suffix in gs->playersuffix

	if (strcmp(suffix, gs->playersuffix) == 0)
		return(1);

	return(0);
}








