// gmodstat_otherevents.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Collection of in-game events that are not related to
// normal client actions: Shutdown, Exit, Initgame, CPhisto2
//
//
 
#include "stdafx.h"
#include "gmodstat.h"



// initgame()
//
// Extract the map name, initialize and add a new
// seen_game entry to the supplied seen_games list
//

struct seen_games *initgame(char *p,
							int ts,
							struct seen_games **gl,
							struct global_state *gs)
{
	struct seen_games *newg;
	char delimit[] = "\\mapname\\\0"; // Escape codes needed for "\mapname\"
	int delimit_len = 9;			// Number of characters to match from log line

	newg = (struct seen_games *)malloc(sizeof(struct seen_games));

	if (newg == NULL)
	{
		fprintf(gs->g_outf, "initgame(): At %d:%02d, malloc failure, exiting.\n",
				ts/60, (ts % 60));
		 exit(-1);
	}

	// Build the new seen_games entry
	memset((char *)newg, 0, sizeof(struct seen_games));

	newg->next = *gl;		// First entry in the gamelist is always current game.
	if ((*gl) != NULL)
		(*gl)->prev = newg;	// This new game will become the "prev" of
							// the previously 'current' game
	*gl = newg;
	newg->t_start = ts;		// Game begins at current tstamp value
	newg->t_end = -1;		// An end of "-1" means game still underway.
	newg->max_c_id = -1;

	// Find where the delimiter "\mapname\" occurs
	while ((strncmp(p, delimit, delimit_len) != 0) && (*p != 0))
		p++;
	// Grab the ascii map name (assume max of 64 chars)
	sscanf(p, "\\mapname\\%64[^\\]", newg->mapname);

	// Set the map's base timestamp entry appropriately
	newg->t_base = gs->time_base;

	return(newg);
}



//
// exitgame()
//
// Game is now over. Mark the player entries with the game stop
// time, and set game end time.
//
// NOTE: Clients may still join the server at this point, and will
// be considered valid clients (with client IDs already assigned,
// and IP address already 'seen') when the game performs shutdown/restart.
// (For example, it is possible for clients to "say" things between the
// Exit and Shutdown - need to treat these as valid in-game comments)
//
// Most importantly, between Exit and Shutdown the server will log the
// final player rankings in a descending sequence of "Score:" lines.
//

struct seen_games *exitgame(char *p,
							int ts,
							struct global_state *gs)
{
	struct seen_player *tmpplayer;
	struct seen_games *gl = gs->gamelist;

	// Sanity check that a game as already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "exitgame(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return(NULL);
	}

	gl->t_end = ts;

	// Ensure every player's inout list is stamped with the final
	// exit time if they haven't been so stamped already.
	tmpplayer = gl->c_all;
	while (tmpplayer != NULL)
	{
		player_out(tmpplayer, ts);
		tmpplayer = tmpplayer->next;
	}

	return(gl);
}



//
// gamescore()
//
// Game is over, we read in a few 'score' lines that tell
// us the final stats according to the server. Note these
// only apply to the clients remaining at the end of the game.
// Clients who joined, then left before the end will not have
// any 'score' info recorded in their seen_player entry.
//


struct seen_games *gamescore(char *p,
							int ts,
							struct global_state *gs)
{
	struct seen_games *gl = gs->gamelist;
	struct client_id *tmpcid;
	int frags, ping, cid;
	char name[TOKENLEN];


	// Sanity check that a game as already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "gamescore(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return(NULL);
	}

	// Sanity check that the game as also finished ("Exit:" has been seen)
	if (gl->t_end == -1)
	{
		fprintf(gs->g_outf, "gamescore(): At %d:%02d, game isn't finished\n",
			ts/60, (ts%60));
		return(NULL);
	}

	// Read in the rest of the score line

	sscanf(p, "%d ping: %d  client: %d %[^\n\r]", &frags, &ping, &cid, name);

	// Make sure the client ID matches a current client
	tmpcid = findclient_id(gl->c_now, cid);
	if (tmpcid == NULL)
	{
		fprintf(gs->g_outf,"gamescore(): At %d:%02d, client_id %d doesn't appear in local list.\n",
			ts/60, (ts%60), cid);
		return(NULL);
	}

	// Make sure there's a seen_player associated with this client ID
	if (tmpcid->c_info == NULL)
	{
		fprintf(gs->g_outf,"gamescore(): At %d:%02d, client_id %d has no seen_player link.\n",
			ts/60, (ts%60), cid);
		return(NULL);
	}

	// Unlikely, but log if the ascii playername doesn't match
	if (strcmp(name, (tmpcid->c_info)->playername) != 0)
		fprintf(gs->g_outf,"gamescore(): At %d:%02d, %s doesn't match %s for client_id %d.\n",
			ts/60, (ts%60), name, (tmpcid->c_info)->playername, cid);

	// Okay, set the frag and ping counts

	(tmpcid->c_info)->frags = frags;
	(tmpcid->c_info)->ping = ping;

	// Increment the "seen clients" counter

	gl->max_c_id++;

	return(gl);
}



// shutdowngame()
//
// Game is over, clean up the local state, copy over various
// client information into the global seen_clients list.
//

struct seen_games *shutdowngame(char *p,
								int ts,
								struct global_state *gs)
{
	struct seen_player *tmpplayer;
	struct seen_games *gl = gs->gamelist;

	// Sanity check that a game as already begun, and space allocated.
	if (gl == NULL)
	{
		fprintf(gs->g_outf, "shutdowngame(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return(NULL);
	}

	// It is possible for ShutdownGame to be called before
	// an 'Exit:' token is seen (manual shutdown), so
	// explicitly perform 'Exit' clean up.

	if (gl->t_end == -1)
	{	// Game is being shutdown in mid-flight, call exitgame()
		exitgame(p, ts, gs);
	}

	// Weed out the seen_player entries of people who didn't
	// collect enough items per minute to be considered valid.
	// (this minimum value is set in the config file)

	weed_out_idleplayers(gl, gs);

	// Sort and rank seen_player entries by fragcount, and
	// store the rankings in each seen_player entry

	sort_by_frags(gl, gs);

	// Sort and rank seen_player entries by median ping, and
	// store the rankings in each seen_player entry
	// (NB: This is only relevant pre-gmod1.0)

	sort_by_pings(gl, gs);

	// Count the number of remaining seen_player entries in this game,
	// and add them to the seen_client list for post-analysis

	gl->max_p_name = 0;
	tmpplayer = gl->c_all;
	while (tmpplayer != NULL)
	{
		gl->max_p_name++; // Count this player

		// While we're here, sort the list of items this
		// player picked up, in descending order of frequency
		// with which each item was picked up.

		sort_items(tmpplayer);

		// Log this client to the global seen_clients list

		if ((tmpplayer->c_addr != NULL) && FLAG_UNSET(gs->stats, NO_CLIENT_CALC))
			add_newclients(tmpplayer, gs);
			
		tmpplayer = tmpplayer->next;
	}

	// Retain the client_id list, we need its history to
	// carry over IP addresses of clients who first joined in
	// previous games

	return(gl);
}



// clientpinghisto()
//
// Parses a "CPhisto2:" entry, which represents a histogram of
// server-estimated ping times (for a particular client) since
// the previous CPhisto2 event. CPhisto2 entries are generated
// by QuakeIII servers running gmod1.0 or later.
//
// Format is "CPhisto2: ID Low Hi lowerrs hierrs tdelta string"
// where
//  ID is clientID
//  Low is the lowest bucket in this interval (ms)
//  Hi is the highest bucket in this interval (ms)
//  lowerr is number of ping samples = 0ms (wierd but true)
//  hierr is number of ping samples > 998ms (mostly 999ms)
//  tdelta number of milliseconds since last histogram
//  string is the histogram, encoded in printable ASCII. (See
// 		the definition of extract_one_phisto() for the discussion
//		of the encoding schemes.)
//
// The <string> is assumed to be less than 1024 characters.
//
// By default, gmod1.0 generates CPhisto2 lines every 2000
// client frames.
//
void clientpinghisto(char *p, int ts, struct global_state *gs)
{
	struct client_id *pingedcid;
	struct seen_player *player;
	struct player_cphisto *tmpcph;
	int client, low, hi, lerr, herr, tdelta;
	char cphisto[MAXLINE];		// No more than 1024 chars in compressed
								// histo string

	// Sanity check that a game as already begun, and space allocated.
	if (gs->gamelist == NULL)
	{
		fprintf(gs->g_outf, "clientpinghisto(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return;
	}

	// If the gmod version hasn't been set yet, this logfile is
	// probably one of the unreleased versions of gmod (call it "0.8")
	// that generated CPhisto2 but did not log its version
	// number at the start of each logfile with "ModVersion:" line.
	
	if (gs->gmod_ver < (float)1.0)
		gs->gmod_ver = (float)0.8;
		
	// Extract client ID and ping value from command line

	sscanf(p, "%d %d %d %d %d %d %[^\r\n]",
		&client, &low, &hi, &lerr, &herr, &tdelta, cphisto);
		
	// Find the pinged client's client_id list entry

	pingedcid = findclient_id(gs->gamelist->c_now, client);
	if (pingedcid == NULL)
	{
		fprintf(gs->g_outf,"clientpinghisto(): At %d:%02d, client_id %d not in local list.\n",
			ts/60, ts%60, client);
		return;
	}

	// Sanity check the client is actually an active player
	if ((player = pingedcid->c_info) == NULL)
	{
		fprintf(gs->g_outf,"clientpinghisto(): At %d:%02d, client_id %d has no playername.\n",
			ts/60, ts%60, client);
		return;
	}
	
	// Note that incase CPhisto events can occur *after* the end of a game,
	// i.e. between the "Exit" (conclusion of one game) and "Shutdown" (beginning
	// of the next game).
	
	if (player->inout == NULL)
	{
		// Only complain about this case if the game is currently running.
		// If the game has finished then silently ignore, since we don't
		// really care to record CPhisto information covering the non-realtime
		// periods between actual games.
		
		if (gs->gamelist->t_end == -1)
			fprintf(gs->g_outf,"clientpinghisto(): At %d:%02d, client %s hasn't joined game yet!?.\n",
				ts/60, ts%60, player->playername);
		return;
	}

	// Only flag a "supposedly left game" error if the game is still in progress
	// (i.e. the "end of game" timestamp still hasn't been filled in).

	if ((player->inout->out_ts != -1) && (gs->gamelist->t_end == -1))
	{
		fprintf(gs->g_outf,"clientpinghisto(): At %d:%02d, client %s (%d,%d) has supposedly left game!.\n",
			ts/60, ts%60, player->playername, client, pingedcid->c_id);
		return;
	}

	// With a valid client ID pointer, update the list of accumulated ping histograms
	// in the associated seen_player entry

	// Make space for the new player_cphisto entry
	if ((tmpcph = (struct player_cphisto *)malloc(sizeof(struct player_cphisto)))
		 == NULL)
	{
		fprintf(gs->g_outf,"clientpinghisto(): At %d:%02d, malloc failed, exiting!.\n",
			ts/60, ts%60);
		exit(-1);
	}

	// Make space for the compressed ping histo string
	if ((tmpcph->cphisto = (char *)malloc(1+strlen(cphisto)))
		 == NULL)
	{
		fprintf(gs->g_outf,"clientpinghisto(): At %d:%02d, malloc failed, exiting!.\n",
			ts/60, ts%60);
		exit(-1);
	}
	
	strcpy(tmpcph->cphisto, cphisto);
	tmpcph->tstamp = (long)ts + gs->time_base;
	tmpcph->low = low;
	tmpcph->hi = hi;
	tmpcph->interval = tdelta;
	tmpcph->lerr = lerr;
	tmpcph->herr = herr;
	tmpcph->next = player->cph;
	player->cph = tmpcph;
	player->cphcnt++;

	return;
}



// ++++++ Supporting routines for preceding in-game
// ++++++ event routines


// sort_items()
//
// Sort the 'items' list hanging off a seen_player entry in
// descending order of count, and then ascii item name.
// Uses a brain dead bubble sort, but who cares since the
// list will be short.
//
void sort_items(struct seen_player *seenp)
{
	struct player_item *tmpitem, *tmpitem2, *previtem;
	int swap = 1;

	while (swap == 1)
	{
		tmpitem = seenp->items;
		previtem = NULL;
		swap = 0;		// Made no bubble swaps yet
		
		while (tmpitem != NULL)
		{
		 if (tmpitem->next != NULL)
		 {
			
			 if ((tmpitem->itemcnt) < (tmpitem->next->itemcnt))
				
			 {
				 // swap
				swap = 1;
				if (previtem == NULL)
				{
					tmpitem2 = tmpitem->next->next;
					seenp->items = tmpitem->next;
					tmpitem->next = tmpitem2;
					(seenp->items)->next = tmpitem;
				}
				else
				{
					tmpitem2 = tmpitem->next->next;
					previtem->next = tmpitem->next;
					tmpitem->next = tmpitem2;
					(previtem->next)->next = tmpitem;
				}
			 }

			 else if ((tmpitem->itemcnt == tmpitem->next->itemcnt) &&
				 (strcmp(tmpitem->itemname, tmpitem->next->itemname) < 0))
			 { // swap
				swap = 1;

				if (previtem == NULL)
				{
					tmpitem2 = tmpitem->next->next;
					seenp->items = tmpitem->next;
					tmpitem->next = tmpitem2;
					(seenp->items)->next = tmpitem;
				}
				else
				{
					tmpitem2 = tmpitem->next->next;
					previtem->next = tmpitem->next;
					tmpitem->next = tmpitem2;
					(previtem->next)->next = tmpitem;
				}

			 }
		 }
		 previtem = tmpitem;
		 tmpitem = tmpitem->next;
		}
	}

	return;
}


// sort_by_frags()
//
// Sorts the seen_player entries of a particular game,
// in order of descending frag count, then assigns each
// entry a numeric ranking - 1 for highest frag count,
// 2 for 2nd highest, etc. Updates the seen_game entry
// gl->ranktotal with the number of seen_player entries
// finally seen (may differ from gs->max_p_name because
// some seen_players were previously eliminated due to
// being idle)
//
// Players of equivalent frag counts are sorted in ascending
// playedtime (higher frags per minute are better)
//
// NOTE: The bots are NOT ranked. Their rank is set to "-1"
// and they are not counted by gl->ranktotal

void sort_by_frags(struct seen_games *gl,
					struct global_state *gs)
{
	struct seen_player *tmpplayer, *tmpplayer2, *prevplayer;
	int totalrank;
	int swap = 1;

	// First, a dumb bubble sort

	while (swap == 1)
	{
		tmpplayer = gl->c_all;
		prevplayer = NULL;
		swap = 0;		// Made no bubble swaps yet
		
		while (tmpplayer != NULL)
		{
		 if (tmpplayer->next != NULL)
		 {
			 if (frag_count(tmpplayer) < frag_count(tmpplayer->next))
				
			 {
				// swap because of fragcount
				swap = 1;
				if (prevplayer == NULL)
				{
					tmpplayer2 = tmpplayer->next->next;
					gl->c_all = tmpplayer->next;
					tmpplayer->next = tmpplayer2;
					(gl->c_all)->next = tmpplayer;
				}
				else
				{
					tmpplayer2 = tmpplayer->next->next;
					prevplayer->next = tmpplayer->next;
					tmpplayer->next = tmpplayer2;
					(prevplayer->next)->next = tmpplayer;
				}
			 }

			 else if ((frag_count(tmpplayer) == frag_count(tmpplayer->next)) &&
				 (tmpplayer->totaltime > tmpplayer->next->totaltime))
			 {
				// swap because of totaltime
				swap = 1;

				if (prevplayer == NULL)
				{
					tmpplayer2 = tmpplayer->next->next;
					gl->c_all = tmpplayer->next;
					tmpplayer->next = tmpplayer2;
					(gl->c_all)->next = tmpplayer;
				}
				else
				{
					tmpplayer2 = tmpplayer->next->next;
					prevplayer->next = tmpplayer->next;
					tmpplayer->next = tmpplayer2;
					(prevplayer->next)->next = tmpplayer;
				}

			 }
		 }
		 prevplayer = tmpplayer;
		 tmpplayer = tmpplayer->next;
		}
	}

	// Walk the list, assign frag rank numbers to non-bot players

	totalrank = 0;
	tmpplayer = gl->c_all;
	while (tmpplayer != NULL)
	{
		if (tmpplayer->c_addr != NULL)
		{
			totalrank++;
			tmpplayer->fragrank = totalrank;
		}
		else
			tmpplayer->fragrank = -1;

		tmpplayer = tmpplayer->next;
	}

	gl->ranktotal = totalrank;

	return;
}

// frag_count()
//
// Takes a pointer to a seen_player entry, and
// returns a total frag count by walking the
// player_inout list
//

int frag_count(struct seen_player *tmpplayer)
{
	int tmpfrags;
	struct player_inout *iolist;

	// Calculate the frag count from inout list
	// rather than end-of-game number.

	iolist = tmpplayer->inout;
	tmpfrags = 0;
	while (iolist != NULL)
	{
		tmpfrags += iolist->frags;
		iolist = iolist->next;
	}

	return(tmpfrags);
}




// sort_by_pings()
//
// Sorts the seen_player entries of a particular game,
// in order of descending median ping, then assigns each
// entry a numeric ranking - 1 for lowest ping value,
// 2 for 2nd lowest, etc. Updates the seen_game entry
// gl->ranktotal with the number of seen_player entries
// finally seen (may differ from gs->max_p_name because
// some seen_players were previously eliminated due to
// being idle)
//
// Players of equivalent median ping value are sorted in descending
// pingcnt (preference to players with more ping samples)
//
// NOTE: The bots are NOT ranked. Their rank is set to "-1"
// and they are not counted by gl->ranktotal


void sort_by_pings(struct seen_games *gl,
					struct global_state *gs)
{
	struct seen_player *tmpplayer, *tmpplayer2, *prevplayer;
	int totalrank;
	int swap = 1;

	// Calculate and store the median ping for each seen_player
	// entry based on the pinglist

	tmpplayer = gl->c_all;
	while (tmpplayer != NULL)
	{
		// Only call median_ping() for non-bot players
		if (tmpplayer->c_addr != NULL)
			tmpplayer->medianping = median_ping(gs, tmpplayer);
		else
			tmpplayer->medianping = 0;
		tmpplayer = tmpplayer->next;
	}

	// A dumb bubble sort

	while (swap == 1)
	{
		tmpplayer = gl->c_all;
		prevplayer = NULL;
		swap = 0;		// Made no bubble swaps yet
		
		while (tmpplayer != NULL)
		{
		 if (tmpplayer->next != NULL)
		 {
			 if (tmpplayer->medianping > tmpplayer->next->medianping)
				
			 {
				// swap because of median ping
				swap = 1;
				if (prevplayer == NULL)
				{
					tmpplayer2 = tmpplayer->next->next;
					gl->c_all = tmpplayer->next;
					tmpplayer->next = tmpplayer2;
					(gl->c_all)->next = tmpplayer;
				}
				else
				{
					tmpplayer2 = tmpplayer->next->next;
					prevplayer->next = tmpplayer->next;
					tmpplayer->next = tmpplayer2;
					(prevplayer->next)->next = tmpplayer;
				}
			 }

			 else if ((tmpplayer->medianping == tmpplayer->next->medianping) &&
				 (tmpplayer->pingcnt < tmpplayer->next->pingcnt))
			 {
				// swap because of pingcnt
				swap = 1;

				if (prevplayer == NULL)
				{
					tmpplayer2 = tmpplayer->next->next;
					gl->c_all = tmpplayer->next;
					tmpplayer->next = tmpplayer2;
					(gl->c_all)->next = tmpplayer;
				}
				else
				{
					tmpplayer2 = tmpplayer->next->next;
					prevplayer->next = tmpplayer->next;
					tmpplayer->next = tmpplayer2;
					(prevplayer->next)->next = tmpplayer;
				}

			 }
		 }
		 prevplayer = tmpplayer;
		 tmpplayer = tmpplayer->next;
		}
	}

	// Walk the list, assign ping rank numbers

	totalrank = 0;
	tmpplayer = gl->c_all;
	while (tmpplayer != NULL)
	{
		if (tmpplayer->c_addr != NULL)
		{
			totalrank++;
			tmpplayer->pingrank = totalrank;
		}
		else
			tmpplayer->pingrank = -1;

		tmpplayer = tmpplayer->next;
	}

	gl->ranktotal = totalrank;

	return;
}



// median_ping()
//
// Takes a pointer to a seen_player entry, and
// returns the median ping value.
//
// The player's logged CPhisto2 information will
// generate median ping value over the game, if there
// is a valid cph list.
//
// The player's logged ClientPing information (for
// pre-gmod1.0 logfiles) will be used if there's
// a valid pinglist.
//
// 102501: Fixed serious bug that caused median ping
// values to be scaled by factor 1/gs->ping_step (kinda
// bad!)
//

int median_ping(struct global_state *gs,
				struct seen_player *tmpplayer)
{
	float pinghisto[1000];
	float pingtotal;
	int arrayptr;
	float sum;
	int tmpsize, tmpmax;

	// Zero all the entries in the array

	pingtotal = 0;
	memset((char *)pinghisto, 0, 1000*sizeof(float));
	
	// Temporarily reset the global ping bucket size and maximum
	// ping values.
	
	tmpsize = gs->ping_step;
	tmpmax = gs->max_ping;
	gs->ping_step = 1;
	gs->max_ping = 999;


	if (tmpplayer->cph != NULL)
	{
    	// Step through the CPhisto2 entries for this player,
    	// adding to aggregate histo.
		struct player_cphisto *cphistolist;
		float tmpcph[1000];
		
		cphistolist = tmpplayer->cph;
		while (cphistolist != NULL)
		{
			int low, mid, hi;
           	int n;
						
			// Extract the compressed ping histo into a full
			// sized array of 1000 buckets. Ignore the histogram
			// if the number of errored samples (lerr and herr)
			// exceeds 1% of the number of actual samples.
			n = extract_one_phisto(cphistolist, &(tmpcph[0]), &low, &mid, &hi, gs);
   			if ((float)(cphistolist->lerr + cphistolist->herr) > (float)0.01*n)
   			{
   				//fprintf(gs->g_outf, "median_ping: skipping a cphisto, too many errors.\n");
   				//fprintf(gs->g_outf, "     lerr/herr %d/%d, numpings %d.\n",
   				//		cphistolist->lerr, cphistolist->herr, n);							
   			}
   			else
   			{
     			// Add each element of this cphisto histogram to the aggregate
    			// histogram in pinghisto[]
    			n = 0;			
            	while (n <= 999)
            	{	
            		if (tmpcph[n] != 0)
            			new_add_ping_to_histo(n, tmpcph[n], gs, pinghisto, &(pingtotal));	
            		n++;
            	}
           	}

			cphistolist = cphistolist->next;
		}
	}
	else if (tmpplayer->pinglist != NULL)
	{
    	// Step through the ClientPing entries for this player,
    	// adding to aggregate histo.

		struct player_ping *pinglist;
		
    	pinglist = tmpplayer->pinglist;
    	while (pinglist != NULL)
    	{
			new_add_ping_to_histo(pinglist->ping, (float)1.0, gs, pinghisto, &(pingtotal));	
    		pinglist = pinglist->next;
    	}
	}
	
	// Restore the global ping bucket size and maximum
	// ping values.
	
	gs->ping_step = tmpsize;
	gs->max_ping = tmpmax;

	// Find the ping value below which 50% of ping samples
	// exist

	arrayptr = 0;
	sum = (float)0.0;
	while (arrayptr < 999)
	{
		sum += pinghisto[arrayptr]/pingtotal;

		if (sum >= 0.5)
			return(arrayptr);

		arrayptr++;
	}

	return(-1);
}





// weed_out_idleplayers()
//
// Weed out the seen_player entries of people who didn't
// collect enough items per minute to be considered valid,
// or who accumulated zero minutes (zero-time players
// aren't even useful for collecting IP address/port statistics
// since nothing can be learned from their existence).
//

void weed_out_idleplayers(struct seen_games *gl,
						struct global_state *gs)
{
	struct seen_player *tmpplayer;
	struct seen_player *prevplayer;

	tmpplayer = gl->c_all;
	prevplayer = NULL;
	while (tmpplayer != NULL)
	{
		int dropflag;

		if (tmpplayer->totaltime == 0)
		{
			dropflag = 1;
		}
		else if ((60*tmpplayer->itemcnt/(float)tmpplayer->totaltime) < gs->itemratio)
		{
			dropflag = 1;
		//	fprintf(gs->g_outf, "shutdowngame at %d:%02d: dropped %s, too idle (itemratio %f)\n",
		//		ts/60, (ts%60), tmpplayer->playername,
		//		(60*tmpplayer->itemcnt/(float)tmpplayer->totaltime));
		}
		else
			dropflag = 0;

		if (dropflag == 1)
		{

			if (prevplayer == NULL)
			{
				gl->c_all = tmpplayer->next;
				free(tmpplayer);
				tmpplayer = gl->c_all;
			}
			else
			{
				prevplayer->next = tmpplayer->next;
				free(tmpplayer);
				tmpplayer = prevplayer->next;
			}
		}
		else
		{
			prevplayer = tmpplayer;
			tmpplayer = tmpplayer->next;
		}
	}

	return;
}
