// gmodstat_parsing.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
//Pull each line from the logfile, parse timestamp and
// call appropriate token handling routines.
//

#include "stdafx.h"
#include "gmodstat.h"



//
// token_stats()
//
// Updates a linked list of tokens and the number of times we've seen
// each token in the file. Called each time a token is
// parsed from a logfile line. Creates a new entry if the token hasn't
// been seen before. Tokens are arbitrary ASCII strings.
//

struct seen_cmdtoken *token_stats(struct global_state *gs,
								  char *token,
								  int tstamp,
								  struct seen_cmdtoken **tl)
{
	struct seen_cmdtoken *tmpaddr;

	// step through the existing token list and see if this one
	// has already been seen

	tmpaddr = *tl;
	while (tmpaddr != NULL)
	{
		if (strcmp (token, tmpaddr->cmdtoken) == 0)
		{
			// Token has been seen before, update the count and return.
			tmpaddr->numseen++;
			return(tmpaddr);
		}
		tmpaddr = tmpaddr->next;	
	}

	// Create memory for new token entry.
	
	if ((tmpaddr = (struct seen_cmdtoken *)malloc(sizeof(struct seen_cmdtoken)))
		 == NULL)
	{
		fprintf(gs->g_outf, "Memory allocation failure in token_stats, quitting.\n");
		exit(-1);
	}

	// Clear the new token entry.
	memset((char *)tmpaddr, 0, sizeof(struct seen_cmdtoken));

	tmpaddr->next = *tl; // Make tmpaddr the first entry in the list.
	*tl = tmpaddr;

	strcpy(tmpaddr->cmdtoken, token);

	tmpaddr->numseen = 1;
	tmpaddr->tstamp = tstamp;

	return(tmpaddr);


}

//
// parse_line2()
//
// Extracts timestamp and first token (command), then calls
// post-processing functions depending on the token seen.
//

int parse_line2(FILE *in_f,
			   int *ts,
			   struct global_state *gs)
{
	char	oneline[MAXLINE];	// Hope 1024 bytes is longer than any likely
								// log file line
 	char tmptoken[64];
	struct seen_cmdtoken *token;
	char *ptr;
	int tstamp;
	int ss;


	// First characters are the xx:yy timestamp, xx minutes
	// and yy seconds since begin of logfile. Return total
	// integer number of seconds since begin of logfile.

	if ((tstamp = parse_time2(gs, in_f)) < 0)
		return(tstamp);		// Parsing the timestamp encountered problem

	*ts = tstamp;

	// Read token and rest of line into separate string.
	// (This is a hack - QIII logfiles contain one specific line
	// that breaks the "xx:yy <token> <stuff> format. It is a line
	// of "-----" characters by themselves as the token. fscanf()
	// copies at most 32 chars into 'tmptoken', so the second field
	// copies the remaining "----" chars into oneline, retaining correct
	// line boundaries.)
	
	ss = fscanf(in_f, "%32s %[^\n\r]", tmptoken, oneline);

	if ((ss == 0) || (ss == EOF))
	{
		fprintf(gs->g_outf, "parse_token error: Zero fields matched, or EOF (code %d).\n", ss);
		return(NULL); // Error if nothing read at this point
	}

	ptr = oneline;
	
	token = token_stats(gs, tmptoken, tstamp, &(gs->cmdlist));
 	if (token == NULL)
		return(0);

	// Perform token-specific processing on rest of the line

	if (strcmp(token->cmdtoken, "InitGame:\0") == 0)
		initgame(ptr, tstamp, &(gs->gamelist), gs);

	if (strcmp(token->cmdtoken, "ClientConnect:\0") == 0)
		clientconnect(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "ClientUserinfoChanged:\0") == 0)
		clientuserinfochanged(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "ClientBegin:\0") == 0)
		clientbegin(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "ClientDisconnect:\0") == 0)
		clientdisconnect(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "say:\0") == 0)
		clientsay(ptr, tstamp, gs, gs->saycid);

	if (strcmp(token->cmdtoken, "Kill:\0") == 0)
		clientkill(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "Exit:\0") == 0)
		exitgame(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "score:\0") == 0)
		gamescore(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "ShutdownGame:\0") == 0)
		shutdowngame(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "Item:\0") == 0)
		clientitem(ptr, tstamp, gs);
		
	// Below are tokens I've added to the QuakeIII server
	// logfile output (either server mods or logfile massaging)

	if (strcmp(token->cmdtoken, "BaseTime:\0") == 0)
		setbasetime(ptr, tstamp, gs);

	if (strcmp(token->cmdtoken, "CPhisto2:\0") == 0)
		clientpinghisto(ptr, tstamp, gs);
		
	if (strcmp(token->cmdtoken, "ModVersion:\0") == 0)
		setgmodversion(ptr, tstamp, gs);

	// Modification pre-gmod1.0, and only used for parsing
	// my older logfiles.
	
	if (strcmp(token->cmdtoken, "ClientPing:\0") == 0)
		clientping(ptr, tstamp, gs);
	
	return(0);
}



//
// parse_time2()
//
// Take ASCII xx:yy timestamp and extract min and sec values.
// Returns integer number of seconds since log start
//

int parse_time2(struct global_state *gs,
				FILE *in_f)
{
	int min = 0, sec = 0, numchar, ss;

	ss = fscanf(in_f, "%d : %d %n", &min, &sec, &numchar);

	if ((ss == 0) || (ss == EOF))
	{
		fprintf(gs->g_outf, "parse_time error: Zero fields matched, or EOF (code %d).\n", ss);
		return(-2); // Error if nothing read at this point
	}

	if (numchar < 5)
	{
		fprintf(gs->g_outf, "parse_time error: Only %d characters scanned.\n", numchar);
		return(-1); // Also an error for less than 5 ascii characters
					// to have been read while parsing the time stamp
	}

	// Hack to fix a bug in the QuakeIII logfile format. When the minutes
	// since the game began exceeds 9999 (four digits) the seconds value
	// is squashed from 2 digits to only 1, the left most digit representing
	// "tens of seconds". Scale it up before handing back to the calling
	// routine.

	if (min > 9999)  sec = sec * 10;

	return(sec + (min*60));
}

