// gmodstat_pingstuff.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Routines for analysing ClientPing and CPhisto2 data from
// each QuakeIII logfile
//

#include "stdafx.h"
#include "gmodstat.h"




// single_client_phisto()
//
// If the SINGLE_PHISTOS flag was set ('single_client_phisto X Y' option)
// then walk through every game, and generate either an aggregate ping
// histogram (over all games) or print the in-game ping histograms
// logged from "CPhisto2:" entries in the logfile. Do this only for
// playername X whose IP addr have prefix Y
// (i.e. specific client defined by playername and suffix).  Clients
// with unresolved playernames are represented by suffix of the
// form ".w-x.unresolved"
//
// The ping data used here is collected from CPhisto2: lines in
// the logfile, which themselves represent histograms of ping values
// covering 20->50s of playing time (2000 client frames each).
//
// Use "do_phisto_total" or "do_phisto_pergame" to create aggregate histograms
// (over all games, or per game) rather than dumping each CPhisto2 line's
// uncompressed histogram.
//
// Use the "ping_histo_range N M" config option to specify bucket
// size N (ms) and maximum ping value of M (ms). (defaults are
// 1ms bucket, 998ms maximum ping)

void single_client_phisto(struct global_state *gs)
{
	struct  seen_games *glist;
	struct	seen_player *plist;
	struct player_cphisto *cphistolist;
	int totalpings;
	int totalphistos;
	int totalgames;
	char outfname[64];
	float tmpcph[1000];		// Temporary ping histo 0->999ms
	int arraysize;


	// Games are stored last->first hanging off glist. Walk the
	// list until we find the last entry, then step backwards up the
	// list, printing stats as we go

	if ((glist = gs->gamelist) == NULL)
		return;

	while (glist->next != NULL)
		glist = glist->next;

	// Establish new output file
	sprintf(outfname,"SCPH-%s.txt\0", gs->playername);
	if ((gs->g_outf2 = fopen(outfname, "w")) == NULL)
	{
		printf("single_client_phisto(): Opening output file %s failed, aborting.\n",
			outfname);
		exit(1);
	}
    		
	// glist points to entry at end of list, chronologically
	// the first one parsed from the logfile

	printf("Generating single_client_phisto for %s/%s to file %s\n",
		gs->playername, gs->playersuffix, outfname);

	fprintf(gs->g_outf, "Generating single_client_phisto for %s/%s to file %s\n",
		gs->playername, gs->playersuffix, outfname);

	// If we're doing final or intermediate aggregate histograms, allocate
	// space for the array of 'float'

	gs->ping_histo = NULL;
	gs->ping_total = 0;
	
	if (FLAG_SET(gs->graph_stats, PHISTO_TOTAL) ||
		FLAG_SET(gs->graph_stats, PHISTO_GAME))
	{
		arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);
		gs->ping_histo = (float *)malloc(arraysize);
		if (gs->ping_histo == NULL)
		{
			printf("single_client_phisto: malloc failed, exiting\n");
			fprintf(gs->g_outf, "single_client_phisto: malloc failed, exiting\n");
			exit(-1);
		}
		else
		{
			// Zero all the entries in the array
			memset((char *)gs->ping_histo, 0, arraysize);
		}
	}

	// Walk the seen_game list backwards (forwards chronologically)
	totalpings = totalphistos = totalgames = 0;

	while (glist != NULL)
	{
		if ((gs->maptime == -1) || (glist->t_start == gs->maptime))
		{
			// Step through all the players seen in this game
			plist = glist->c_all;
			while (plist != NULL)
			{
				if (is_right_client(plist, gs))
				{
					totalphistos += plist->cphcnt;
					totalgames++;

					// Step through the CPhisto entries for this client,
					// adding to histo or printing to outfile as appropriate
					cphistolist = plist->cph;

					while (cphistolist != NULL)
					{
						int low, mid, hi;
						int numpings;
						
						// Extract the compressed ping histo into a full
						// sized array of 1000 buckets. Ignore the histogram
                        // if the number of errored samples (lerr and herr)
                        // exceeds 1% of the number of actual samples.
						numpings = extract_one_phisto(cphistolist, &(tmpcph[0]), &low, &mid, &hi, gs);
						if ((float)(cphistolist->lerr + cphistolist->herr) > (float)0.01*numpings)
						{
							//fprintf(gs->g_outf, "single_client_phisto: skipping a cphisto, too many errors.\n");
							//fprintf(gs->g_outf, "     lerr/herr %d/%d, numpings %d.\n",
							//		cphistolist->lerr, cphistolist->herr, numpings);
						}
						else
						{
    						totalpings += numpings;
    		
    						if (FLAG_SET(gs->graph_stats, PHISTO_TOTAL) ||
    							FLAG_SET(gs->graph_stats, PHISTO_GAME))
    						{
                            	int n = 0;
    							// If we're looking for an aggregate histogram output, add each
    							// element of this cphisto histogram to the aggregate
    							// histogram hanging off gs->ping_histo, using the ping
    							// bucket size specified in the config file (may be larger
    							// than the 1ms bucket sizes reported in CPhisto2: lines)							
                            	while (n <= 999)
                            	{	
                            		if (tmpcph[n] != 0)
                            			new_add_ping_to_histo(n, tmpcph[n], gs, gs->ping_histo, &(gs->ping_total));	
                            		n++;
                            	}
    						}
    						else if (FLAG_SET(gs->graph_stats, PHISTO_MEDIAN))					
    						{
    							fprintf(gs->g_outf2, "%d  %d\n",
    								cphistolist->tstamp, mid);
     						}
    						else if (FLAG_SET(gs->graph_stats, PHISTO_JITTER))					
    						{
    							fprintf(gs->g_outf2, "%d  %d\n",
    								cphistolist->tstamp, hi-low);
     						}
    						else						
    						{
    							int n = 0, x = 0;
    							float cumulative = 0.0;
    							// If we're not doing an aggregate histogram, dump
    							// the in-game histogram represented by the
    							// current player_cphisto entry. Trim off the left
    							// and right hand extremes of the histogram where
    							// buckets tend to zero (cumulative is 0 or close
    							// to 99% respectively). Plot a minimum of 100 buckets
    							// in any case. Buckets are 1ms wide in this case (the
    							// fixed bucket size for CPhisto2: data).
    							fprintf(gs->g_outf2, "\n\"%d\n", cphistolist->tstamp);
                            	while ((n <= 999) && ((cumulative < 99.0) || (x < 100)))
                            	{
                            		cumulative += 100*tmpcph[n] / (float)numpings;
                            		if (cumulative > 0.0)
                            		{
                            			x++;
                            			fprintf(gs->g_outf2, "%d %7.3f\n",
                            				 n, 100*tmpcph[n] / (float)numpings);
                            		}
                            		n++;
                            	}
     						}
 						}

						cphistolist = cphistolist->next;
					}
					
					// We've scanned every CPhisto2 entry associated with
					// this player in this game. If we're doing per-game aggregate
					// histos, and the player saw more than TWO CPhisto2 entries,
					// then dump the current aggregate (and reset the aggregate
					// even if only one or two CPhisto2 line was seen, before analysing
					// next game).
					//
					if (FLAG_SET(gs->graph_stats, PHISTO_GAME))
					{
                		int arrayptr = 0;
                		int arraylimit = gs->max_ping / gs->ping_step;
                		int pingcntr = 0;
	               		float current, cumulative;
	               		
	               		if (plist->cphcnt > 2)
	               		{
                    		fprintf(gs->g_outf2, "\n\"Start %d\n",
                    			glist->t_start + glist->t_base);

                    		// Print out the histogram contents, up to the ping value below
                    		// which 99% of all samples were seen (used to suppress zeros
                    		// out to 998ms when most players show pings under 200ms)
                			cumulative = 0.0;
                    		while ((arrayptr < arraylimit) && (cumulative <= 99.0))
                    		{
                    			current = ((float)100.0*(gs->ping_histo)[arrayptr])/gs->ping_total;
                    			cumulative += current;
                    			if (cumulative > 0.0)
                    				fprintf(gs->g_outf2, "%d %7.2f\n", pingcntr, current);
                    			arrayptr++;
                    			pingcntr += gs->ping_step;
                    		}
                		}
					
						memset((char *)gs->ping_histo, 0, arraysize);
						gs->ping_total = 0;
					}
					
				}
				plist = plist->next;
			}
		}
		glist = glist->prev;
	}

	if (totalpings == 0)
	{
    	fprintf(gs->g_outf, "Warning - Saw no pinghisto samples for >%s< >%s<\n",
    		gs->playername, gs->playersuffix);	
    	printf("Warning - Saw no pinghisto samples for >%s< >%s<\n",
    		gs->playername, gs->playersuffix);	
	}
	else
	{
    	fprintf(gs->g_outf, "Saw %d pinghistos covering %d samples (%d games) for >%s< >%s<\n",
    		totalphistos, totalpings, totalgames, gs->playername, gs->playersuffix);

    	// If we were collecting a histogram, dump its contents

    	if (FLAG_SET(gs->graph_stats, PHISTO_TOTAL))
    	{
    		int arrayptr = 0;
    		int arraylimit = gs->max_ping / gs->ping_step;
    		int pingcntr = 0;
    		int lowping, midping, highping;
    		float current, cumulative;

    		// Find the lower, median and upper ping values that encompass
    		// N% of ping samples.
    		
			new_find_ping_hilow(gs, gs->ping_histo, gs->ping_total,
					&lowping, &midping, &highping);
		
    		fprintf(gs->g_outf, "     %7.2f elements, lower %d, median %d, upper %d\n",
    			gs->ping_total, lowping, midping, highping);
    			
    		fprintf(gs->g_outf2, "\"SCPH-%s\n", gs->playername);

    		// Print out the histogram contents, up to the ping value below
    		// which 99% of all samples were seen (used to suppress zeros
    		// out to 998ms when most players show pings under 200ms)
			cumulative = 0.0;
    		while ((arrayptr < arraylimit) && (cumulative <= 99.0))
    		{
    			current = ((float)100.0*(gs->ping_histo)[arrayptr])/gs->ping_total;
    			cumulative += current;
    			if (cumulative > 0.0)
    				fprintf(gs->g_outf2, "%d %7.2f\n", pingcntr, current);
    			arrayptr++;
    			pingcntr += gs->ping_step;
    		}
    	}
	}

	// Clean up the array space, if used
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

	if (gs->g_outf2 != NULL)
		fclose(gs->g_outf2);

	return;
}


// extract_one_phisto()
//
// Takes a pointer to a player_phisto entry and extracts the
// compressed ping histogram into an array of 1000 'float'
// buckets. Returns the total number of ping samples across
// the histogram.
//
// The histogram is extracted from the compressed ASCII string
// stored with each player_cphisto entry. The string begins with
// a ">" character (to ensure it begins with a non-space char)
// and then continues:
//
//  For gmod version 1.0:
//
//   Repeated "XY" pairs of ASCII characters, or "+nn%" indicating
//   the previous bucket's value is repeated in the next nn
//   buckets (mostly used for suppressing adjacent buckets
//   with value of zero when there's bi/multi-modal distribution
//   of ping values).
//
//   The "XY" pairs use base64, with X being
//   the 64s column and Y being the 1s column. The ASCII encoding
//   adds 32 (code for " ") to the base64 digit. This way each
//   bucket can count up to 4095 using just two ASCII characters.
//
//  For gmod version 1.1 and greater:
//
//   Repeated "XY" pairs of ASCII characters, or "znn%" indicating
//   the previous bucket's value is repeated in the next nn
//   buckets (mostly used for suppressing adjacent buckets
//   with value of zero when there's bi/multi-modal distribution
//   of ping values).
//
//   The "XY" pairs use base64, with X being
//   the 64s column and Y being the 1s column. The ASCII encoding
//   adds 33 (code for " ") to the base64 digit. This way each
//   bucket can count up to 4095 using just two ASCII characters.
//
//  (gmod version 1.0 had a bug in that the "+nn%" escape sequence
//	 was ambiguous for buckets whose value was above 703, since the
//   base64, adjust character for 704 was "+". Verison 1.1 and later
//	 use "z" as the escape character, and also chose to make all the
//	 encoded characters non-space by adding 33 rather than 32.)

int extract_one_phisto(struct player_cphisto *cph,
						float *tmparray,
						int *low,
						int *mid,
						int *hi,
						struct global_state *gs)
{
	char *index;
	int bucket;
	int total;
	int tmpcnt;
	int tmpstep, tmpmax;
	int offset;
	char esc_char;

	*low = *mid = *hi = 0;
	memset((char *)tmparray, 0, 1000*sizeof(float));
	
	// Unlikely error, but just in case.
	if (cph->cphisto == NULL)
		return(0);
		
	if (gs->gmod_ver < (float)0.8)
		return(0);
		
	if (gs->gmod_ver < (float)1.1)
	{
		// For gmod1.0 (or the implied gmod0.8 which was never released)
		offset = 32;
		esc_char = '+';	
	}
	else
	{
		// For gmod1.1 or later
		offset = 33;
		esc_char = 'z';	
	}

	total = 0;
	tmpcnt = 0;
	index = cph->cphisto;
	bucket = cph->low;
	index++;	// Skip the initial non-space marker
	
	// Scan the compressed histo string until null terminator
	while (*index != 0)
	{
		// Ugly hack here to side-step the error in gmod1.0 where
		// "+" was used as the escape character when it could also
		// appear as a legitimate base64+32 digit. Only treat
		// this as an escaped compression code if there's a one, two,
		// or three digit number immediately following it.
		if ((*index == esc_char) &&
				((isdigit(index[1]) && (index[2] == '%')) ||
				 (isdigit(index[1]) && isdigit(index[2]) && (index[3] == '%')) ||
				 (isdigit(index[1]) && isdigit(index[2]) && isdigit(index[3]) && (index[4] == '%'))))
				
		{
			// Extract the digits up to the next "%" to
			// represent how many times to repeat the previous
			// bucket's count value.
			int multiplier, chars = 0;
			
			index++;
			sscanf(index,"%d%%%n", &multiplier, &chars);
			index += chars;
			while (multiplier > 0)
			{
				tmparray[bucket] += tmpcnt;
				total += tmpcnt;
				bucket++;
				multiplier--;			
			}
		}
		else
		{	
			// Extract a single 2-digit base64 ASCII value representing
			// the number of elements at the current bucket value
			tmpcnt = (((*index) - offset)*64 + (*(index+1) - offset));
			tmparray[bucket] += tmpcnt;
			total += tmpcnt;
			index += 2;
			bucket++;
		}
	}
	
	tmpstep = gs->ping_step;
	tmpmax = gs->max_ping;
	gs->ping_step = 1;
	gs->max_ping = 999;
	if ((tmpcnt = new_find_ping_hilow(gs, tmparray, (float)total, low, mid, hi))
			 != 0)
		fprintf(gs->g_outf, "extract_one_phisto: Error %d with new_find_ping_hilow\n",
			tmpcnt);
	gs->ping_step = tmpstep;
	gs->max_ping = tmpmax;
	
	return(total);

}


// new_add_ping_to_histo()
//
// Takes an integer ping time and increments a counter in the
// appropriate histogram bucket corresponding to that ping value.
// [ping times accumulate in the bucket X if they fall
// within -0.5*ping_step to +0.4999*ping_step of X. e.g. if ping_step
// is 10, then a ping time of 22 would add to the bucket for ping == 20,
// whereas a ping time of 26 would add to the bucket for ping == 30.]

void new_add_ping_to_histo(int newping,
					  float incr,
					  struct global_state *gs,
					  float *array,
					  float *counter)
{
	double arrayptr_f1, arrayptr_f2;
	int arrayptr;

	// If the supplied ping value is out of range,
	// silently return

	if ((newping < 0) || (newping > gs->max_ping))
		return;

	// Calculate the index into the array, rounding up or
	// down as appropriate (e.g. 5.50 becomes 6, 5.49 becomes 5)
	// This means ping times accumulate in the bucket X if they
	// fall within -0.5*ping_step to +0.4999*ping_step of X

	arrayptr_f1 = ((float)newping / (float)gs->ping_step);
	arrayptr_f2 = floor(0.5 + arrayptr_f1);
	arrayptr = (int)arrayptr_f2;

	// update that entry

	(array)[arrayptr] += incr;

	// Increment the running total counter
	*counter += incr;

	return;

}

// new_find_ping_hilow()
//
// Scans the supplied array and finds the lower and upper ping
// values that cover N% of the total number of ping samples.
// Also finds the median ping value below which 50% of samples
// are found.
//
// Returns -1 if for some reason it couldn't calculate all
// three bounds (lower, median, and upper).
//

int new_find_ping_hilow(struct global_state *gs,
					 float *array,
					 float count,
					 int *lping,
					 int *mping,
					 int *hping)
{
	int arrayptr = 0;
	int arraylimit = gs->max_ping / gs->ping_step;
	float sum = 0.0;
	float lowersum, highersum;

	// Wierd error check

	if (array == NULL)
		return(-2);

	lowersum = (1-(gs->jitterwidth))/(float)2.0;
	highersum = 1-lowersum;

	*lping = -1;
	*mping = -1;
	*hping = -1;

	while (arrayptr < arraylimit)
	{
		sum += ((array)[arrayptr])/count;

		if ((sum >= lowersum) && (*lping == -1))
			*lping = arrayptr*gs->ping_step;

		if ((sum >= (float)0.5) && (*mping == -1))
			*mping = arrayptr*gs->ping_step;

		if ((sum >= highersum) && (*hping == -1))
		{
			*hping = arrayptr*gs->ping_step;
			return(0);
		}

		arrayptr++;
	}

	fprintf(gs->g_outf, "new_find_ping_hilow: Error - ended without setting values.\n");
	return(-1);

}





// topN_client_phistos()
//
// Called if TOPN_PHISTOS flag was set.
//
// Call single_client_phisto() for every client that accumulated
// more than gs->clienttime total playing time.
//
// See single_client_phisto() for additional options.
//

void topN_client_phistos(struct global_state *gs)
{
	struct seen_client *tmpclist;
	struct global_state *tmpgs;
	
	// Sort the client list in descending order of number of
	// games played

	sort_seen_client(gs);
	
	// Create a copy of the global state struct 'gs', so we can
	// mess with the parameters passed to single_client_phisto().

	if ((tmpgs = (struct global_state *)malloc(sizeof(struct global_state)))
		== NULL)
		{
			printf("topN_client_phistos: malloc failed, exiting.\n");
			fprintf(gs->g_outf, "topN_client_phistos: malloc failed, exiting.\n");
			exit(-1);
		}
		
	memcpy(tmpgs, gs, sizeof(struct global_state));

	// Walk the seen_clients list

	tmpclist = gs->clientlist;
	while ((tmpclist != NULL) && (tmpclist->totaltime >= gs->clienttime))
	{
		if ((tmpclist->unnamedplayer == 0) || FLAG_SET(gs->stats, UNNAMED_PLAYER))
		{
    		// Set the parameters for single_client_phisto() to be
    		// the name and domain of this client.
    		
    		tmpgs->maptime = -1;
    		strcpy(tmpgs->playername, tmpclist->pname);
    		strcpy(tmpgs->playersuffix, tmpclist->suffix);
    		
    		single_client_phisto(tmpgs);
		}
		
		tmpclist = tmpclist->next;
	}
	
	free(tmpgs);

	return;
}




// create_ping_histo_onegame()
//
// Takes pointer to a particular seen_game entry, and adds the
// median ping times for each player that belongs to a specific set of
// client domains.
//
// Players are also skipped if they did not have enough ping samples
// to make a meaningful median ping estimate (generated during end-of-game
// "shutdown" processing). For gmod1.0 logfiles, this means at least
// three CPhisto2 entries. (For my earlier logfiles, at least 10 ClientPing
// lines are required.)
//

void create_ping_histo_onegame(int hour,
					  struct seen_games *glist,
					  struct global_state *gs)
{
	struct	seen_player *plist;
	int yesflag;

	plist = glist->c_all;	// Step through all the players seen in this game
	while (plist != NULL)
	{
		// If there's one or more IP addresses this player entry is NOT a bot

		if (plist->c_addr != NULL)
		{
			// If ONLY_UNRESOLVED is set then only count this player's
			// contribution if its addresses were not resolved to a
			// domain name.
			//
			// Otherwise, let is_in_region() decide if the resolvable
			// domain names are in the desired set of domains.

			yesflag = 0;

			if ((FLAG_SET(gs->graph_stats, ONLY_UNRESOLVED)) &&
				(is_in_unresolved(plist, gs)))
					yesflag = 1;
			if ((FLAG_UNSET(gs->graph_stats, ONLY_UNRESOLVED)) &&
				(is_in_region(plist, gs)))
					yesflag = 1;

			if (yesflag == 1)
			{
				// This client's median ping value has already been
				// calculated during Shutdown processing, and is
				// stored in gs->medianping. Use this value if it is less
				// than gs->max_ping and there were enough samples.
				
				if (plist->cph != NULL)
				{
					// This player logged CPhisto2 lines
    				if ((plist->cphcnt > 2) && (plist->medianping > 0)
    						 && (plist->medianping < gs->max_ping))
      	 				add_ping_to_histo(plist->medianping, 1.0, gs);
				}
				else if (plist->pinglist != NULL)
				{
					// This player logged ClientPing lines
    				if ((plist->pingcnt > 9) && (plist->medianping > 0)
    						 && (plist->medianping < gs->max_ping))
      	 				add_ping_to_histo(plist->medianping, 1.0, gs);
				}
				
			}
		}

		plist = plist->next;
	}

	return;

}

//
// create_selected_ping_histo()
//
// Walks through the seen_games list, and creates
// a histogram of median ping values for clients
// who fall under specified conditions (eg. a specific
// domain, multiple domains, or unresolved addresses)
//
// The histogram is dumped to gs->g_outf2.
// Cumulative distribution is dumped to gs->g_outf3.
//
// This routine is not called directly, rather it is
// called from create_ping_histo() which sets up the
// correct output files gs->g_outf2 and gs->g_outf3
// for each iteration.
//

void create_selected_ping_histo(struct global_state *gs,
								struct global_names *gn)
{
	struct  seen_games *glist;
	int week;
	char separator;
	int arraysize;
	int arrayptr;
	int arraylimit;
	float sum;
	
	if (gs->g_outf2 == NULL)
	{
		printf("create_selected_ping_histo(): gs->g_outf2 not set!?\n");
		fprintf(gs->g_outf, "create_selected_ping_histo(): gs->g_outf2 not set!?\n");
		return;
	}
	
	if (gs->g_outf3 == NULL)
	{
		printf("create_selected_ping_histo(): gs->g_outf3 not set!?\n");
		fprintf(gs->g_outf, "create_selected_ping_histo(): gs->g_outf3 not set!?\n");
		return;
	}
	
	// Calculate number of bytes we need to minimally reserve for
	// the histogram storage
	
   	arraysize = (1 + gs->max_ping/gs->ping_step) * sizeof(float);
	
	// Games are stored last->first hanging off glist. Walk the
	// list until we find the last entry. (We then walk backwards
	// up the list, which is chronologically forward in time)

	if ((glist = gs->gamelist) == NULL)
		return;
	while (glist->next != NULL)
		glist = glist->next;

	printf("Generating ping histogram\n");
	week = 0;

	if (FLAG_SET(gs->graph_stats, XGRAPH_OUT))
		separator = ' ';
	else
		separator = ',';

	// allocate space for the array of 'float'

	gs->ping_total = 0;
	if ((gs->ping_histo = (float *)malloc(arraysize))
		 == NULL)
	{
		printf("create_selected_ping_histo(): malloc failed, exiting\n");
		fprintf(gs->g_outf, "create_selected_ping_histo(): malloc failed, exiting\n");
		exit(-1);
	}
	else
	{
		// Zero all the entries in the array
		memset((char *)(gs->ping_histo), 0, arraysize);
	}

	// Begin stepping through the game list

	while (glist != NULL)
	{
		int gamelen = glist->t_end - glist->t_start;
		long realstart = glist->t_base + glist->t_start;	//Absolute time stamp
		long realmiddle = realstart + (gamelen/2);
		struct tm *tmnow;
		int hour;

		// The time of this game is calculated from the mid-point
		// of the game ('realmiddle' is halfway between absolute start
		// and end times)
		
		tmnow = localtime(&realmiddle);
		if (FLAG_SET(gs->graph_stats, HOUR_OF_DAY))
			hour = 2*tmnow->tm_hour + (tmnow->tm_min / 30);
		else
			hour = 24*(tmnow->tm_wday) + tmnow->tm_hour;

		create_ping_histo_onegame(hour, glist, gs);

		glist = glist->prev;
	}

	
	// Finished walking backwards through the seen_game list
	// Now dump appropriate results

	arrayptr = 0;
	arraylimit = gs->max_ping / gs->ping_step;
	sum = 0.0;
	while (arrayptr < arraylimit)
	{
		sum += (100.0*(gs->ping_histo)[arrayptr])/gs->ping_total;
		
		fprintf(gs->g_outf2, "%d %c %7.2f\n", arrayptr*gs->ping_step, separator,
			(100.0*(gs->ping_histo)[arrayptr])/gs->ping_total);

		if ((sum > 0.5) && (sum < 99.5))
			fprintf(gs->g_outf3, "%d %c %7.2f\n", arrayptr*gs->ping_step,
				separator, sum);

		arrayptr++;
	}

	// Clean up the array space, in case we're called again
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

		
	return;
}



//
// create_ping_histo()
//
// Calls create_selected_ping_histo() once or multiple
// times to generate one or more histograms of pings
// vs time. If GRAPH_MANY flag is set then histograms are
// generated for clients in each of the domains listed
// in gs->fragementlist. Otherwise, a single histogram is
// generated for clients who are members of any of the
// domains listed in gs->fragementlist.
//
// Also generates cumulative distribution of median pings
//

void create_ping_histo(struct global_state *gs,
								struct global_names *gn)
{
   	struct hostaddrlist *old_frglist;
   	struct hostaddrlist *nextfrg;
   	struct hostaddrlist tmpfragment;
   	char outfname[64];

	// Sanity checks
	
	if (gs->dnslist == NULL)
	{
		printf("create_ping_histo: there's no dns mapping list, exiting.\n");
		fprintf(gs->g_outf, "create_ping_histo: there's no dns mapping list, exiting.\n");
		return;
	}
	
	// If we're supposed to be doing histograms only for some specific
	// domains (either individually or as a group) then flag error if there's
	// nothing hanging off gs->fragmentlist
	
	if ((gs->fragmentlist == NULL) &&
		(FLAG_SET(gs->graph_stats, GRAPH_MANY) || FLAG_SET(gs->graph_stats, INCLUDE_REGIONS)))
	{
		printf("create_ping_histo: no target domain(s) specified, exiting.\n");
		fprintf(gs->g_outf, "create_ping_histo: no target domain(s) specified, exiting.\n");
		return;
	}

	// If GRAPH_MANY is set, begin to loop through all entries in
	// gs->fragmentlist, doing histograms for each one. Otherwise,
	// do one histogram (that will include clients from the domains
	// in gs->fragementlist, or all clients)
	
	if (FLAG_SET(gs->graph_stats, GRAPH_MANY))
	{
    	// Decouple the original gs->fragmentlist, and replace it with a
    	// single entry.

    	old_frglist = gs->fragmentlist;
    	nextfrg = old_frglist;
    	gs->fragmentlist = &tmpfragment;
    	memset((char *)gs->fragmentlist, 0, sizeof(struct hostaddrlist));

    	while (nextfrg != NULL)
    	{
    		// Establish new output files
    		sprintf(outfname,"CPH%s-pinghisto.txt\0", nextfrg->name);
    		gs->g_outf3 = fopen(outfname, "w");
    		if (gs->g_outf3 == NULL)
    		{
    			printf("Opening output file %s failed, aborting.\n",
    				outfname);
    			exit(1);
    		}
    		
    		sprintf(outfname,"PH%s-pinghisto.txt\0", nextfrg->name);
    		gs->g_outf2 = fopen(outfname, "w");
    		if (gs->g_outf2 == NULL)
    		{
    			printf("Opening output file %s failed, aborting.\n",
    				outfname);
    			exit(1);
    		}
    		
    		printf("Creating ping histogram for domain %s (file %s)\n",
    					nextfrg->name, outfname);
    		fprintf(gs->g_outf, "Creating ping histogram for domain %s (file %s)\n",
    					nextfrg->name, outfname);

    		// Modify gs->fragmentlist appropriately (adjust the name pointer
    		// to point to the real name in nextfrg->name, and leave the next
    		// pointer NULL so it looks like a fragmentlist with only one entry).

    		gs->fragmentlist->name = nextfrg->name;

    		// Call create_selected_ping_histo() to parse the seen_game list with
    		// a focus on the single entry in gs->fragmentlist

    		create_selected_ping_histo(gs, gn);
    		fclose(gs->g_outf3);
    		fclose(gs->g_outf2);
    		
    		nextfrg = nextfrg->next;
    	}

    	// Finished, clean up and return

    	gs->fragmentlist = old_frglist;
	}
	else
	{
		// This is the single-histogram case. Open one file, then
		// call create_selected_ping_histo(). But first, list
		// the specific domains (if applicable).
		
    	if (FLAG_UNSET(gs->graph_stats,ONLY_UNRESOLVED) &&
    			(gs->fragmentlist != NULL))
    	{
    		printf("Creating ping histogram for multiple domains\n");
    		fprintf(gs->g_outf, "Creating ping histogram for multiple domains\n");
    		if (FLAG_SET(gs->graph_stats, INCLUDE_REGIONS))
    		{
    			printf("Include list: ");
    			fprintf(gs->g_outf, "Include list: ");
    		}
    		if (FLAG_SET(gs->graph_stats, EXCLUDE_REGIONS))
    		{
    			printf("Exclude list: ");
    			fprintf(gs->g_outf, "Exclude list: ");
    		}
    		if (FLAG_SET(gs->graph_stats, EXCLUDE_REGIONS) ||
    			FLAG_SET(gs->graph_stats, INCLUDE_REGIONS))
    		{
    			nextfrg = gs->fragmentlist;
    			while (nextfrg != NULL)
    			{
    				printf("%s", nextfrg->name);
    				fprintf(gs->g_outf, "%s", nextfrg->name);
    				nextfrg = nextfrg->next;
    				if (nextfrg != NULL)
    				{
    					printf(", ");
    					fprintf(gs->g_outf, ", ");
    				}
    			}
    			printf("\n");
    			fprintf(gs->g_outf, "\n\n");
    		}
    	}
    	else
    	{
       		printf("Creating ping histogram for all domains\n");
    		fprintf(gs->g_outf, "Creating ping histogram for all domains\n");
	   	}

	   	// Open output file
   		sprintf(outfname,"CPH0-pinghisto.txt\0");
   		gs->g_outf3 = fopen(outfname, "w");
   		if (gs->g_outf3 == NULL)
   		{
   			printf("Opening output file %s failed, aborting.\n",
   				outfname);
   			exit(1);
   		}
    		
   		sprintf(outfname,"PH0-pinghisto.txt\0");
   		gs->g_outf2 = fopen(outfname, "w");
   		if (gs->g_outf2 == NULL)
   		{
   			printf("Opening output file %s failed, aborting.\n",
   				outfname);
   			exit(1);
   		}
   		else
   			fprintf(gs->g_outf, "Dumping histogram to %s\n", outfname);

   		create_selected_ping_histo(gs, gn);
    	fclose(gs->g_outf3);
    	fclose(gs->g_outf2);

	}
	
	
	return;
}



