// gmodstat_pre_gmod.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//

// This file contains various routines that I developed for analysing
// logfiles generated pre-gmod1.0. In particular, earlier logfiles contained
// ping samples each time a client picked up an item, or was killed. This
// was an early attempt at sampling client ping experience, and was superseded
// by the "CPhisto2:" compressed histogram lines in gmod1.0 and later.
//
// I keep these routines in gmodstat for my own use. They don't
// have much value to anyone using gmod1.0 or later.
//


#include "stdafx.h"
#include "gmodstat.h"



// clientping()
//
// Parses "ClientPing:" events which may appear in pre-gmod1.0
// logfiles, representing the current (server estimated)
// ping times for a particular Client ID. No guarantees that the
// actual reported ping times are particularly accurate.
//
// Format is "ClientPing: N M" where N is clientID and M is ping in ms.
//

void clientping(char *p, int ts, struct global_state *gs)
{
	struct client_id *pingedcid;
	struct seen_player *player;
	struct player_ping *tmpping;
	int pinged, ping;

	// Sanity check that a game as already begun, and space allocated.
	if (gs->gamelist == NULL)
	{
		fprintf(gs->g_outf, "clientping(): At %d:%02d gamelist is NULL.\n",
				ts/60, (ts % 60));
		return;
	}

	// If the gmod version hasn't been set yet, this logfile is
	// probably one of the early versions of gmod (call it "0.2")
	// that generated ClientPing but did not generate CPhisto2 lines.
	// (From version 1.0 onwards the gmod entity would log its version
	// number at the start of each logfile with "ModVersion:" line.)
	
	if (gs->gmod_ver == (float)0.0)
		gs->gmod_ver = (float)0.2;
		
	// Extract client ID and ping value from command line

	sscanf(p, "%d %d", &pinged, &ping);


	// Sanity check the returned ping value, and log an error message if
	// it outside the valid range (should never occur, but just in case
	// a logfile gets mangled somewhere...)

	if ((ping <= 0) || (ping > 999))
	{
		fprintf(gs->g_outf,"clientping(): At %d:%02d, ping of %d is out of range.\n",
			ts/60, ts%60, ping);
		return;
	}

	// If the ping value is 999, the server has actually lost track of
	// the client (probably disconnected ungracefully) so we ignore
	// these data points entirely.

	if (ping == 999)
		return;

	// Find the pinged client's client_id list entry

	pingedcid = findclient_id(gs->gamelist->c_now, pinged);
	if (pingedcid == NULL)
	{
		fprintf(gs->g_outf,"clientping(): At %d:%02d, client_id %d not in local list.\n",
			ts/60, ts%60, pinged);
		// Temporary hack - dump the client_id list
		if (FLAG_UNSET(gs->stats, HACK_STOP))
		{
			struct client_id *tmpcid = gs->gamelist->c_now;
			while (tmpcid != NULL)
			{
				fprintf(gs->g_outf, "ClientID %d, state %d, s_time %d, name %s\n",
					tmpcid->c_id, tmpcid->state, tmpcid->s_time, tmpcid->c_info->playername);
				tmpcid = tmpcid->next;			
			}
			gs->stats |= HACK_STOP;	
		}
		return;
	}

	// Sanity check the client is actually an active player

	player = pingedcid->c_info; // ClientID entry points back to the seen_player entry

	if (player == NULL)
	{
		fprintf(gs->g_outf,"clientping(): At %d:%02d, client_id %d has no playername.\n",
			ts/60, ts%60, pinged);
		return;
	}

	if (player->inout == NULL)
	{
		fprintf(gs->g_outf,"clientping(): At %d:%02d, client %s hasn't joined game yet!?.\n",
			ts/60, ts%60, player->playername);
		return;
	}

	// Note that since ClientPing events can occur *after* the end of a game,
	// only flag a "supposedly left game" error if the game is still in progress
	// (i.e. the "end of game" timestamp still hasn't been filled in). Yes, this
	// means we *might* miss an error condition of an invalid ClientPing just
	// after the end of the game, but really.... this error check is probably
	// pointless anyway.

	if ((player->inout->out_ts != -1) && (gs->gamelist->t_end == -1))
	{
		fprintf(gs->g_outf,"clientping(): At %d:%02d, client %s (%d,%d) has supposedly left game!.\n",
			ts/60, ts%60, player->playername, pinged, pingedcid->c_id);
		return;
	}

	// With a valid client ID pointer, update the list of seen pings
	// in the associated seen_player entry

	tmpping = (struct player_ping *)malloc(sizeof(struct player_ping));
	if (tmpping == NULL)
	{
		fprintf(gs->g_outf,"clientping(): At %d:%02d, malloc failed, exiting!.\n",
			ts/60, ts%60);
		exit(-1);
	}

	tmpping->ping = ping;
	tmpping->tstamp = (long)ts + gs->time_base;
	tmpping->next = player->pinglist;
	player->pinglist = tmpping;
	player->pingcnt++;

	return;
}




// single_client_pings()
//
// If the SINGLE_PINGS flag was set ('single_client_pings X Y' option)
// then walk through every game, and generate the scatter plot or histo
// of in-game ping values for playername X whose IP addr have prefix Y
// (i.e. specific client defined by playername and suffix).  Clients
// with unresolved playernames are represented by suffix of the
// form ".w-x.unresolved".
//
// The ping data used here is collected from ClientPing: lines in
// the logfile.
//
// Use "do_ping_histo" to specify creation of histogram rather than
// scatter plot.
// Use the "ping_histo_range N M" config option to specify bucket
// size N (ms) and maximum ping value of M (ms). (defaults are
// 1ms bucket, 998ms maximum ping)

void single_client_pings(struct global_state *gs)
{
	struct  seen_games *glist;
	struct	seen_player *plist;
	struct player_ping *pinglist;
	float weektime;
	struct tm *tmnow;
	long realtime;
	int totalpings;
	int tp2;
	char outfname[64];

	// Games are stored last->first hanging off glist. Walk the
	// list until we find the last entry, then step backwards up the
	// list, printing stats as we go

	if ((glist = gs->gamelist) == NULL)
		return;

	while (glist->next != NULL)
		glist = glist->next;

	// Establish new output file
	sprintf(outfname,"SCP-%s.txt\0", gs->playername);
	if ((gs->g_outf2 = fopen(outfname, "w")) == NULL)
	{
		printf("single_client_pings(): Opening output file %s failed, aborting.\n",
			outfname);
		exit(1);
	}
    		
	// glist points to entry at end of list, chronologically
	// the first one parsed from the logfile

	printf("Generating single_client_pings for %s/%s to file %s\n",
		gs->playername, gs->playersuffix, outfname);

	fprintf(gs->g_outf, "Generating single_client_pings for %s/%s to file %s\n",
		gs->playername, gs->playersuffix, outfname);

	// If we're doing a histogram of ping times, allocate space for the
	// array of 'float'

	gs->ping_histo = NULL;
	gs->ping_total = 0;
	if (FLAG_SET(gs->graph_stats, PING_HISTO))
	{
		int arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);
		gs->ping_histo = (float *)malloc(arraysize);
		if (gs->ping_histo == NULL)
		{
			printf("single_client_pings: malloc failed, exiting\n");
			fprintf(gs->g_outf, "single_client_pings: malloc failed, exiting\n");
			exit(-1);
		}
		else
		{
			// Zero all the entries in the array
			memset((char *)gs->ping_histo, 0, arraysize);
		}
	}

	// Walk the seen_game list backwards (forwards chronologically)
	totalpings = 0;
	tp2 = 0;
	while (glist != NULL)
	{
		if ((gs->maptime == -1) || (glist->t_start == gs->maptime))
		{
			// Step through all the players seen in this game
			plist = glist->c_all;
			while (plist != NULL)
			{
				int flag = 0;

				if (is_right_client(plist, gs))
				{
					if ((flag == 0) && FLAG_UNSET(gs->graph_stats, PING_HISTO) &&
						FLAG_UNSET(gs->graph_stats, DAY_OF_YEAR))
					{
						fprintf(gs->g_outf2, "\n\"Game: %d+%d\n",
							glist->t_start, glist->t_base);
						flag = 1;
					}

					totalpings += plist->pingcnt;
					tp2++;

					// Step through the ping entries for this client,
					// adding to histo or printing to outfile as appropriate
					pinglist = plist->pinglist;

					while (pinglist != NULL)
					{
						if (FLAG_SET(gs->graph_stats, PING_HISTO))
							add_ping_to_histo(pinglist->ping, 1.0, gs);
						else
						{
							realtime = pinglist->tstamp;
							tmnow = localtime(&realtime);
							weektime = (tmnow->tm_hour / (float)24) +
								(tmnow->tm_min / (float)1440) +
								(tmnow->tm_sec / (float)86400);

							// If DAY_OF_YEAR flag is set, the x-axis is
							// time of day of the year, otherwise time of
							// day of the week. The former provides a linear
							// sense of the player's overall trend, the latter
							// shows any weekly repeating patterns
							if (FLAG_SET(gs->graph_stats, DAY_OF_YEAR))
								weektime += tmnow->tm_yday;
							else
								weektime += tmnow->tm_wday;

							fprintf(gs->g_outf2, "%8.6f %d\n",
								weektime, pinglist->ping);					
						}

						pinglist = pinglist->next;
					}
				}
				plist = plist->next;
			}
		}
		glist = glist->prev;
	}

	if (totalpings == 0)
	{
    	fprintf(gs->g_outf, "Warning - Saw no ping samples for >%s< >%s<\n",
    		gs->playername, gs->playersuffix);	
    	printf("Warning - Saw no ping samples for >%s< >%s<\n",
    		gs->playername, gs->playersuffix);	
	}
	else
	{
    	fprintf(gs->g_outf, "Saw %d pings (%d games) for >%s< >%s<\n",
    		totalpings, tp2, gs->playername, gs->playersuffix);

    	// If we were collecting a histogram, dump its contents

    	if (FLAG_SET(gs->graph_stats, PING_HISTO))
    	{
    		int arrayptr = 0;
    		int arraylimit = gs->max_ping / gs->ping_step;
    		int pingcntr = 0;
    		int lowping, midping, highping;
    		float current, cumulative;
    		int x = 0;

    		// Find the lower, median and upper ping values that encompass
    		// N% of ping samples.
    		
    		find_ping_hilow(gs, &lowping, &midping, &highping);

    		fprintf(gs->g_outf, "   %7.2f elements, lower %d, median %d, upper %d\n",
    			gs->ping_total, lowping, midping, highping);
    			
    		fprintf(gs->g_outf2, "\"SCP-%s\n", gs->playername);

    		// Print out the histogram contents, up to the ping value below
    		// which 99% of all samples were seen (used to suppress zeros
    		// out to 998ms when most players show pings under 200ms)
			cumulative = 0.0;
    		while ((arrayptr < arraylimit) && ((cumulative <= 99.0) || (x < 100)))
    		{
    			current = ((float)100.0*(gs->ping_histo)[arrayptr])/gs->ping_total;
    			cumulative += current;
    			if (cumulative > 0.0)
    			{
    				x += gs->ping_step;
    				fprintf(gs->g_outf2, "%d %7.2f\n", pingcntr, current);
    			}
    			arrayptr++;
    			pingcntr += gs->ping_step;
    		}
    	}
	}

	// Clean up the array space, if used
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);
		
	if (gs->g_outf2 != NULL)
		fclose(gs->g_outf2);

	return;
}






// topN_client_pings()
//
// If the TOPN_PINGS flag was set walk through the seen_client pinglist
// and generate the scatter plot or histo of in-game ping times for the
// clients whose total playing time is greater than gs->clienttime.
// (This routine uses ClientPing: data rather than CPhisto2: data)
//
// In addition to the histograms per client, a final histogram is
// created reflecting ALL the ping values across the set of clients
// whose playing times were greater than gs->clienttime.
// NOTE: Not sure how valuable this "total" histogram is because it
// is possibly biassed towards players who generated more ClientPings,
// with no way to determine if though picking up Items, being killed,
// or doing the killing.
//
// Use the "ping_histo_range N M" config option to specify bucket
// size N (ms) and maximum ping value of M (ms). (defaults are
// 1ms bucket, 998ms maximum ping)

void topN_client_pings(struct global_state *gs)
{
	struct seen_client *tmpclist;
	struct player_ping *pinglist;
	float weektime;
	struct tm *tmnow;
	long realtime;
	int arraysize;
	float *totalarray;
	float totalcount;
	char outfname[64];

	// Sort the client list in descending order of number of
	// games played

	sort_seen_client(gs);

	// If we're doing a histogram of ping times, allocate space for the
	// array of 'float'

	gs->ping_histo = NULL;
	totalarray = NULL;
	if (FLAG_SET(gs->graph_stats, PING_HISTO))
	{
		arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);

		gs->ping_histo = (float *)malloc(arraysize);
		if (gs->ping_histo == NULL)
		{
			printf("topN_client_pings: malloc failed, exiting\n");
			fprintf(gs->g_outf, "topN_client_pings: malloc failed, exiting\n");
			exit(-1);
		}

		totalarray = (float *)malloc(arraysize);
		if (totalarray == NULL)
		{
			printf("topN_client_pings: malloc failed, exiting\n");
			fprintf(gs->g_outf, "topN_client_pings: malloc failed, exiting\n");
			exit(-1);
		}
		memset(totalarray, 0, arraysize);
		totalcount = 0.0;
	}

	// Walk the seen_clients list

	tmpclist = gs->clientlist;
	while ((tmpclist != NULL) && (tmpclist->totaltime >= gs->clienttime))
	{
		// If we're doing a new histo, zero the array
		if (FLAG_SET(gs->graph_stats, PING_HISTO))
		{
			memset((char *)gs->ping_histo, 0, arraysize);
			gs->ping_total = 0;
		}
		
     	// Establish new output file
     	sprintf(outfname,"topN-%d-%s.txt\0",
			tmpclist->pingcnt, tmpclist->pname);

     	if ((gs->g_outf2 = fopen(outfname, "w")) == NULL)
     	{
     		printf("topN_client_pings(): Opening output file %s failed, aborting.\n",
     			outfname);
     		exit(1);
     	}
  		
		// Now walk through the pinglist of this seen_client entry
		// and account for each entry.

		fprintf(gs->g_outf2, "\n\"%d-%s\"\n",
			tmpclist->pingcnt, tmpclist->pname);

		pinglist = tmpclist->pinglist;
		while (pinglist != NULL)
		{
			if (FLAG_SET(gs->graph_stats, PING_HISTO))
			{
				add_ping_to_histo(pinglist->ping, 1.0, gs);
			}
			else
			{
				realtime = pinglist->tstamp;
				tmnow = localtime(&realtime);
				weektime = tmnow->tm_wday +
					(tmnow->tm_hour / (float)24) +
					(tmnow->tm_min / (float)1440) +
					(tmnow->tm_sec / (float)86400);
				fprintf(gs->g_outf2, "%8.6f %d\n",
					weektime, pinglist->ping);					
			}
			pinglist = pinglist->next;
		}

		// If we were collecting a histogram, dump its contents

		if (FLAG_SET(gs->graph_stats, PING_HISTO))
		{
			int arrayptr = 0;
			int arraylimit = gs->max_ping / gs->ping_step;
			int pingcntr = 0;

			// Print out the histogram contents, and build the
			// "total" histogram.

			while (arrayptr < arraylimit)
			{
				fprintf(gs->g_outf2, "%d %7.2f\n", pingcntr,
					(100.0*(gs->ping_histo)[arrayptr])/gs->ping_total);

				new_add_ping_to_histo(pingcntr,
				   (100.0*(gs->ping_histo)[arrayptr])/gs->ping_total,
				    gs, totalarray, &totalcount);
				
				arrayptr++;
				pingcntr += gs->ping_step;
			}
		}
		
		fclose(gs->g_outf2);
		
		tmpclist = tmpclist->next;
	}

	// Clean up the array space, if used for histograms
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

	if (FLAG_SET(gs->graph_stats, PING_HISTO))
	{
		int arrayptr = 0;
		int arraylimit = gs->max_ping / gs->ping_step;
		int pingcntr = 0;

     	// Establish new output file
     	sprintf(outfname,"topN-TOTAL.txt\0");

     	if ((gs->g_outf2 = fopen(outfname, "w")) == NULL)
     	{
     		printf("topN_client_pings(): Opening output file %s failed, aborting.\n",
     			outfname);
     		exit(1);
     	}
  		
		fprintf(gs->g_outf2, "\n\"TOTAL HISTOGRAM\n");

		// Print out the histogram contents

		while (arrayptr < arraylimit)
		{
			fprintf(gs->g_outf2, "%d %7.2f\n", pingcntr,
				(100.0*(totalarray)[arrayptr])/totalcount);

			arrayptr++;
			pingcntr += gs->ping_step;
		}
		fclose(gs->g_outf2);
	}

	// Clean up the totalarray space, if used for histograms
	if (totalarray != NULL)
		free(totalarray);


	
	return;
}



// median_client_pings()
//
// Walk through the seen_client pinglist and generate scatter plot
// of median ping time vs 'jitter' (difference between the lower and
// upper ping times between which N% of each client's ping samples
// fell).
//
// We calculate the histogram of pings for each client in order to
// calculate the median ping and jitter. The histograms themselves
// are not output.
//

void median_client_pings(struct global_state *gs)
{
	struct seen_client *tmpclist;
	struct player_ping *pinglist;
	int arraysize;
	int lowping, midping, highping;

	// Use dumb bubble sort :)

	sort_seen_client(gs);

	// We're doing a histogram of ping times, so allocate space for the
	// array of 'float'

	gs->ping_histo = NULL;
	arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);
	gs->ping_histo = (float *)malloc(arraysize);
	if (gs->ping_histo == NULL)
	{
		printf("median_client_pings: malloc failed, exiting\n");
		fprintf(gs->g_outf, "median_client_pings: malloc failed, exiting\n");
		exit(-1);
	}

	// Walk the seen_clients list, and for each client create a histo
	// of its pinglist from which to calculate the median and jitter in the
	// ping samples.

	tmpclist = gs->clientlist;
	while (tmpclist != NULL)
	{
		// Zero the elements in the histo array

		gs->ping_total = 0;
		memset((char *)gs->ping_histo, 0, arraysize);

		// Now walk through the pinglist of this seen_client entry
		// and account for each entry.

		pinglist = tmpclist->pinglist;
		while (pinglist != NULL)
		{
			add_ping_to_histo(pinglist->ping, 1.0, gs);
			pinglist = pinglist->next;
		}

		// Find the lower, median and upper ping values that encompass
		// N% of ping samples, but only for clients having more than
		// 50 actual ping samples.
		if ((tmpclist->pingcnt > 50) && (tmpclist->totaltime > 3600))
		{
			find_ping_hilow(gs, &lowping, &midping, &highping);

			fprintf(gs->g_outf, "%d %d\n",
				midping, highping - lowping);
		}

		tmpclist = tmpclist->next;
	}

	// Clean up the array space, if used for histograms
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

	return;
}



// client_ping_vs_gamenum()
//
// Walk through the seen_client pinglist and generate scatter plot
// of median ping time vs the number of games that client was seen
// playing.
//
// We calculate the histogram of pings for each client in order to
// calculate the median ping. The histograms themselves
// are not output.
//
// If DO_JITTER is set, then plot jitter vs number of games rather
// than median ping.

void client_ping_vs_gamenum(struct global_state *gs)
{
	struct seen_client *tmpclist;
	struct player_ping *pinglist;
	int arraysize;
	int lowping, midping, highping;
#define MAXGAMES 120
	float *fraglist[MAXGAMES];	// Array of MAXGAMES pointers to arrays of floats
	float fragcntr[MAXGAMES];
	int n;


	fprintf(gs->g_outf, "Dumping client median pings vs number of played games\n");
	printf("Dumping client median pings vs number of played games\n");

	// Sort client list in descending order of games played

	sort_seen_client(gs);

	// We're doing a histogram of ping times, so allocate space for the
	// array of 'float'

	gs->ping_histo = NULL;
	arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);
	gs->ping_histo = (float *)malloc(arraysize);
	if (gs->ping_histo == NULL)
	{
		printf("client_ping_vs_gamenum: malloc failed, exiting\n");
		fprintf(gs->g_outf, "client_ping_vs_gamenum: malloc failed, exiting\n");
		exit(-1);
	}

	// Allocate space for the 2nd level ping histograms

	n = 0;
	while (n < MAXGAMES)
	{
		fraglist[n] = (float *)malloc(arraysize);
		if (fraglist[n] == NULL)
		{
			printf("client_ping_vs_gamenum: malloc failed, exiting\n");
			fprintf(gs->g_outf, "client_ping_vs_gamenum: malloc failed, exiting\n");
			exit(-1);
		}
		memset(fraglist[n], 0, arraysize);
		fragcntr[n] = 0;
		n++;
	}

	// Walk the seen_clients list, and for each client create a histo
	// of its pinglist then output the median and jitter in the
	// ping samples. Only consider clients who've accumulated more
	// than 10 ping samples (need a minimum for find_ping_hilow to work)

	tmpclist = gs->clientlist;
	while (tmpclist != NULL)
	{
		if ((tmpclist->pingcnt > 10) && (tmpclist->numseen > 1))
		{
			// Zero the elements in the histo array
			gs->ping_total = 0;
			memset((char *)gs->ping_histo, 0, arraysize);

			// Now walk through the pinglist of this seen_client entry
			// and account for each entry.

			pinglist = tmpclist->pinglist;
			while (pinglist != NULL)
			{
				add_ping_to_histo(pinglist->ping, 1.0, gs);
				pinglist = pinglist->next;
			}

			// Find the lower, median and upper ping values that encompass
			// N% of ping samples
		
			find_ping_hilow(gs, &lowping, &midping, &highping);

			if (FLAG_SET(gs->ping_stats, DO_JITTER))
			{
				fprintf(gs->g_outf, "%d %d\n",
					tmpclist->numseen, highping-lowping);
				new_add_ping_to_histo(highping-lowping, 1.0, gs,
					fraglist[tmpclist->numseen], &(fragcntr[tmpclist->numseen]));
			}
			else
			{
				fprintf(gs->g_outf, "%d %d\n",
					tmpclist->numseen, midping);
				new_add_ping_to_histo(midping, 1.0, gs,
					fraglist[tmpclist->numseen], &(fragcntr[tmpclist->numseen]));
			}

		}
		tmpclist = tmpclist->next;
	}

	// Clean up the array space, if used for histograms
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

	// Now process the 2nd order histograms (the histo of median pings)

	fprintf(gs->g_outf, "\n\"2nd order points\n");
	n = 0;
	while (n < MAXGAMES)
	{
		if (fragcntr[n] > 3.0)
		{
			new_find_ping_hilow(gs, fraglist[n], fragcntr[n],
				&lowping, &midping, &highping);
			fprintf(gs->g_outf, "%d %d\n", n, highping);
		}
		free(fraglist[n]);
		n++;
	}


	return;
}



// client_ping_vs_gametime()
//
// Walk through the seen_client pinglist and generate scatter plot
// of median ping time vs the total game playing time for each client
//
// We calculate the histogram of pings for each client in order to
// calculate the median ping. The histograms themselves
// are not output.
//
// If DO_JITTER is set, then plot jitter vs gametime rather
// than median ping.
//

void client_ping_vs_gametime(struct global_state *gs)
{
	struct seen_client *tmpclist;
	struct player_ping *pinglist;
	int arraysize;
	int lowping, midping, highping;

	fprintf(gs->g_outf, "Dumping client median pings vs total played time\n");
	printf("Dumping client median pings vs total played time\n");

	// Use dumb bubble sort :)

	sort_seen_client(gs);

	// We're doing a histogram of ping times, so allocate space for the
	// array of 'float'

	gs->ping_histo = NULL;
	arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);
	gs->ping_histo = (float *)malloc(arraysize);
	if (gs->ping_histo == NULL)
	{
		printf("client_ping_vs_gametime: malloc failed, exiting\n");
		fprintf(gs->g_outf, "client_ping_vs_gametime: malloc failed, exiting\n");
		exit(-1);
	}

	// Walk the seen_clients list, and for each client create a histo
	// of its pinglist then output the median and jitter in the
	// ping samples. Only consider clients who've accumulated more
	// than 10 ping samples (need a minimum for find_ping_hilow to work)

	tmpclist = gs->clientlist;
	while ((tmpclist != NULL) && (tmpclist->pingcnt > 10))
	{
		// Zero the elements in the histo array
		gs->ping_total = 0;
		memset((char *)gs->ping_histo, 0, arraysize);

		// Now walk through the pinglist of this seen_client entry
		// and account for each entry.

		pinglist = tmpclist->pinglist;
		while (pinglist != NULL)
		{
			add_ping_to_histo(pinglist->ping, 1.0, gs);
			pinglist = pinglist->next;
		}

		// Find the lower, median and upper ping values that encompass
		// N% of ping samples
		
		find_ping_hilow(gs, &lowping, &midping, &highping);

		if (FLAG_SET(gs->ping_stats, DO_JITTER))
			fprintf(gs->g_outf, "%d %d\n",
				tmpclist->totaltime, highping-lowping);
		else
			fprintf(gs->g_outf, "%d %d\n",
				tmpclist->totaltime, midping);

		tmpclist = tmpclist->next;
	}

	// Clean up the array space, if used for histograms
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

	return;
}



// player_ping_vs_frags()
//
// ACTUALLY: This is fragrate/minute vs ping or jitter.
//
// Walk through the seen_game pinglists and generate scatter plot
// of frag count/minute for each seen_player entry vs that player's
// median ping or jitter (as counted through the game, regardless of
// whether the player stayed for the whole game)
//
// Also calculates the average fragrate for each ping/jitter value
// between 0 to 998.
//
// ONLY consider games with 2 or more remote players (ie. 4 or
// more unique clients at any point in the game)
//
// ONLY consider clients with more than 10 ClientPing entries.
//

// We calculate the histogram of pings for each client in order to
// calculate the median ping. The histograms themselves
// are not output.
//
// If DO_JITTER is set, then plot jitter vs gametime rather
// than median ping.
//

void player_ping_vs_frags(struct global_state *gs,
							struct global_names *gn)
{
	struct  seen_games *glist;
	struct	seen_player *plist;
	struct player_ping *pinglist;
	int lowping, midping, highping;
	int arraysize;
	float sumtotal[999]; // Sum of fragrates at ping or jitter value N
	int cntr[999];	// Number of fragrates  at ping or jitter value N
	int totalcntr;	// Total number of fragrate samples measured
	int n;
	char tempname[128];
	FILE *ping_outf;
	FILE *ping_outf2;

	if (FLAG_UNSET(gs->graph_stats, PING_HISTO))
	{
		fprintf(gs->g_outf, "player_ping_vs_frags: do_ping_histo was not set.\n");
		return;
	}

	sprintf(tempname, "fragvsping-%s\0", gn->global_outfile);
	fprintf(gs->g_outf, "Dumping frag rate vs ping to %s\n",
		tempname);
	printf("Dumping frag rate vs ping to %s\n",
		tempname);
	if ((ping_outf = fopen(tempname, "w")) == NULL)
	{
		printf("Opening output file %s failed, aborting.\n",
			tempname);
		fprintf(gs->g_outf, "Opening output file %s failed, aborting.\n",
			tempname);
		exit(-1);
	}
	
	sprintf(tempname, "fragvsping-avg-%s\0", gn->global_outfile);
	if ((ping_outf2 = fopen(tempname, "w")) == NULL)
	{
		printf("Opening output file %s failed, aborting.\n",
			tempname);
		fprintf(gs->g_outf, "Opening output file %s failed, aborting.\n",
			tempname);
		exit(-1);
	}	
	
	
	// Games are stored last->first hanging off glist. Walk the
	// list until we find the last entry, then step backwards up the
	// list, printing stats as we go

	if ((glist = gs->gamelist) == NULL)
		return;

	while (glist->next != NULL)
		glist = glist->next;

	// glist points to entry at end of list, chronologically
	// the first one parsed from the logfile

	// Allocate space for the 1st level ping histograms

	arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);
	gs->ping_histo = (float *)malloc(arraysize);
	if (gs->ping_histo == NULL)
	{
		printf("player_ping_vs_frags: malloc failed, exiting\n");
		fprintf(gs->g_outf, "player_ping_vs_frags: malloc failed, exiting\n");
		exit(-1);
	}

	// Clear the arrays used to calculate the average fragrate
	// for each ping or jitter value from 0 to 998 ms.
	n = 0;
	while (n<999)
	{
		sumtotal[n] = (float)0.0;
		cntr[n] = 0;
		n++;
	}

	totalcntr = 0;

	// Walk the seen_game list backwards (forwards chronologically)

	while (glist != NULL)
	{
		int gamelen = (glist->t_end - glist->t_start);
		
		// Step through all the players seen in this game
		plist = glist->c_all;
		while ((plist != NULL) && (glist->max_c_id >= 4))
		{
			int tmpfrags;

			tmpfrags = frag_count(plist);

			if ((tmpfrags > 0) &&
				(plist->pingcnt > 10))
			{
				totalcntr++;

				// Zero all the entries in the array
				gs->ping_total = 0;
				memset((char *)gs->ping_histo, 0, arraysize);

				// Step through the ping entries for this player,
				// adding to histo
				pinglist = plist->pinglist;
				while (pinglist != NULL)
				{
					add_ping_to_histo(pinglist->ping, 1.0, gs);
					pinglist = pinglist->next;
				}

				// Find the lower, median and upper ping values that encompass
				// N% of ping samples
		
				find_ping_hilow(gs, &lowping, &midping, &highping);

				if (FLAG_SET(gs->ping_stats, DO_JITTER))
				{
					fprintf(ping_outf, "%d  %9.7f\n", highping-lowping,
						60*tmpfrags / (float)plist->totaltime);
					sumtotal[highping-lowping] += (60*tmpfrags / (float)plist->totaltime);
					cntr[highping-lowping]++;

				}
				else if (FLAG_SET(gs->ping_stats, DO_DEVIATION))
				{
					fprintf(ping_outf, "%10.7f  %9.7f\n",
						(highping-lowping)/(float)midping,
						60*tmpfrags / (float)plist->totaltime);
				}
				else
				{
					fprintf(ping_outf, "%d  %9.7f\n", midping,
						60*tmpfrags / (float)plist->totaltime);
					sumtotal[midping/4] += (60*tmpfrags / (float)plist->totaltime);
					cntr[midping/4]++;

				}

			}
	
			plist = plist->next;
		}
		glist = glist->prev;
	}

	// Clean up the array space, if used
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

	// Walk though the sumtotal array and calculate the average
	// fragrates. Print to output if the cntr[] entry indicates	
	// enough sample points

	fprintf(ping_outf2, "\n\"Average fragrates\n");
	n = 0;
	while (n < 999)
	{
		if (cntr[n] >= 4)
		{
			fprintf(ping_outf2, "%d %9.7f\n", 4*n, sumtotal[n]/cntr[n]);
		}
		n++;
	}

	// Walk though the cntr[] array and and for each
	// cntr[] entry taht is non-zero, print its value
	// as a percentage of the total number of fragrate
	// samples collected

	fprintf(ping_outf2, "\n\"%% of fragrate samples\n");
	n = 0;
	while (n < 999)
	{
		if (cntr[n] >= 4)
		{
			fprintf(ping_outf2, "%d %6.2f\n", 4*n, 100*cntr[n]/(float)totalcntr);
		}
		n++;
	}
	
	fclose(ping_outf);
	fclose(ping_outf2);

	return;
}




// player_ping_vs_gameratio()
//
// Walk through the seen_game pinglists and generate scatter plot
// of playedtime ratio for each seen_player entry vs that player's
// median ping or jitter (as counted through the game, regardless of
// whether the player stayed for the whole game)
//
// Also calculates the average playedtime ratio for each median
// ping/jitter value between 0 to 998.
//
// ONLY consider games with 2 or more remote players (ie. 4 or
// more unique clients at any point in the game)
//
// ONLY consider clients with more than 10 ClientPing entries.
//

// We calculate the histogram of pings for each client in order to
// calculate the median ping. The histograms themselves
// are not output.
//
// If DO_JITTER is set, then plot jitter vs gametime rather
// than median ping.
//

void player_ping_vs_gameratio(struct global_state *gs,
							struct global_names *gn)
{
	struct  seen_games *glist;
	struct	seen_player *plist;
	struct player_ping *pinglist;
	int lowping, midping, highping;
	int arraysize;
	float sumtotal[999]; // Sum of playedtimeratio at ping or jitter value N
	int cntr[999];	// Number of playedtimeratios  at ping or jitter value N
	int totalcntr;	// Total number of playedtimeratio samples measured
	int n;
	char tempname[128];
	FILE *ping_outf;
	FILE *ping_outf2;

	if (FLAG_UNSET(gs->graph_stats, PING_HISTO))
	{
		fprintf(gs->g_outf, "player_ping_vs_gameratio: do_ping_histo was not set.\n");
		return;
	}

	sprintf(tempname, "gratiovsping-%s\0", gn->global_outfile);
	fprintf(gs->g_outf, "Dumping playedtimeratio vs ping to %s\n",
		tempname);
	printf("Dumping playedtimeratio vs ping to %s\n",
		tempname);
	if ((ping_outf = fopen(tempname, "w")) == NULL)
	{
		printf("Opening output file %s failed, aborting.\n",
			tempname);
		fprintf(gs->g_outf, "Opening output file %s failed, aborting.\n",
			tempname);
		exit(-1);
	}
	
	sprintf(tempname, "gratiovsping-avg-%s\0", gn->global_outfile);
	if ((ping_outf2 = fopen(tempname, "w")) == NULL)
	{
		printf("Opening output file %s failed, aborting.\n",
			tempname);
		fprintf(gs->g_outf, "Opening output file %s failed, aborting.\n",
			tempname);
		exit(-1);
	}	
	
	
	// Games are stored last->first hanging off glist. Walk the
	// list until we find the last entry, then step backwards up the
	// list, printing stats as we go

	if ((glist = gs->gamelist) == NULL)
		return;

	while (glist->next != NULL)
		glist = glist->next;

	// glist points to entry at end of list, chronologically
	// the first one parsed from the logfile

	// Allocate space for the 1st level ping histograms

	arraysize = (1+ (gs->max_ping / gs->ping_step)) * sizeof(float);
	gs->ping_histo = (float *)malloc(arraysize);
	if (gs->ping_histo == NULL)
	{
		printf("player_ping_vs_frags: malloc failed, exiting\n");
		fprintf(gs->g_outf, "player_ping_vs_frags: malloc failed, exiting\n");
		exit(-1);
	}

	// Clear the arrays used to calculate the average playedtimeratio
	// for each ping or jitter value from 0 to 998 ms.
	n = 0;
	while (n<999)
	{
		sumtotal[n] = (float)0.0;
		cntr[n] = 0;
		n++;
	}

	totalcntr = 0;

	// Walk the seen_game list backwards (forwards chronologically)

	while (glist != NULL)
	{
		int gamelen = (glist->t_end - glist->t_start);
		
		// Step through all the players seen in this game
		plist = glist->c_all;
		while ((plist != NULL) && (glist->max_c_id >= 4))
		{
			int tmpfrags;

			tmpfrags = frag_count(plist);

			if ((tmpfrags > 0) &&
				(plist->pingcnt > 10))
			{
				totalcntr++;

				// Zero all the entries in the array
				gs->ping_total = 0;
				memset((char *)gs->ping_histo, 0, arraysize);

				// Step through the ping entries for this player,
				// adding to histo
				pinglist = plist->pinglist;
				while (pinglist != NULL)
				{
					add_ping_to_histo(pinglist->ping, 1.0, gs);
					pinglist = pinglist->next;
				}

				// Find the lower, median and upper ping values that encompass
				// N% of ping samples
		
				find_ping_hilow(gs, &lowping, &midping, &highping);

				if (FLAG_SET(gs->ping_stats, DO_JITTER))
				{
					fprintf(ping_outf, "%d  %9.7f\n", highping-lowping,
						plist->totaltime/(float)(gamelen));
					sumtotal[highping-lowping] += plist->totaltime/(float)(gamelen);
					cntr[highping-lowping]++;

				}
				else if (FLAG_SET(gs->ping_stats, DO_DEVIATION))
				{
					fprintf(ping_outf, "%10.7f  %9.7f\n",
						(highping-lowping)/(float)midping,
						plist->totaltime/(float)(gamelen));
				}
				else
				{
					fprintf(ping_outf, "%d  %9.7f\n", midping,
						plist->totaltime/(float)(gamelen));
					sumtotal[midping/4] += plist->totaltime/(float)(gamelen);
					cntr[midping/4]++;

				}

			}
	
			plist = plist->next;
		}
		glist = glist->prev;
	}

	// Clean up the array space, if used
	if (gs->ping_histo != NULL)
		free(gs->ping_histo);

	// Walk though the sumtotal array and calculate the average
	// fragrates. Print to output if the cntr[] entry indicates	
	// enough sample points

	fprintf(ping_outf2, "\n\"Average playedtimeratio\n");
	n = 0;
	while (n < 999)
	{
		if (cntr[n] >= 4)
		{
			fprintf(ping_outf2, "%d %9.7f\n", 4*n, sumtotal[n]/cntr[n]);
		}
		n++;
	}

	// Walk though the cntr[] array and and for each
	// cntr[] entry taht is non-zero, print its value
	// as a percentage of the total number of fragrate
	// samples collected

	fprintf(ping_outf2, "\n\"%% of playedtimeratio samples\n");
	n = 0;
	while (n < 999)
	{
		if (cntr[n] >= 4)
		{
			fprintf(ping_outf2, "%d %6.2f\n", 4*n, 100*cntr[n]/(float)totalcntr);
		}
		n++;
	}
	
	fclose(ping_outf);
	fclose(ping_outf2);

	return;
}





// add_ping_to_histo()
//
// Calls new_add_ping_to_histo() with gs->pinghisto as pointer
// to the base array holding a histogram of ping times.
//
void add_ping_to_histo(int newping,
					  float incr,
					  struct global_state *gs)
{
	new_add_ping_to_histo(newping, incr, gs,
		gs->ping_histo, &(gs->ping_total));
	return;
}



//
// Calls new_find_ping_hilow() in order to find lower,
// upper, and median ping values of gs->ping_histo

void find_ping_hilow(struct global_state *gs,
					 int *lping,
					 int *mping,
					 int *hping)
{

	new_find_ping_hilow(gs, gs->ping_histo, gs->ping_total,
		lping, mping, hping);
	return;
}


