// gmodstat_startup.cpp
//
// Copyright (c) 2001, Grenville Armitage
// gj_armitage@yahoo.com
//
//
// +++++++++++++++++++++++ License and Conditions +++++++++++++++++++++++++
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   A copy of this license was supplied with gmodstat in the
//   file ../COPYING. Alternatively, a copy may be obtained from
//   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA  02111-1307, USA.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Routines for parsing the config file for startup
// parameters.
//

#include "stdafx.h"
#include "gmodstat.h"

//
// parse_config_file()
//
// Parses the gmodstat config file to set all parameters
// controlling the subsequent logfile analysis.
//
void parse_config_file(struct global_names *gn,
					   struct global_state *gs)
{
	char confstr[128];	// Store for each line of config file
	FILE *conf = NULL;	// File handle for config file
	int confeof;

	// Set a few default flags

	// Open and parse the config file

	conf = fopen(gn->configfile, "rb");
	if (conf == NULL)
	{
		printf("Opening config file %s failed.\n",
			gn->configfile);
		return;
	}

	confeof = 0;
	while (confeof != 1)
	{
		memset(confstr, 0, 128);
		if (fgets(confstr, 126, conf) == NULL)
			 confeof = 1;
		else
		{			
			
			if (confstr[0] != '#')
				parse_config_line(gn, gs, confstr);

		}
	}
}

//
// parse_config_line()
//
// Parses one line of the gmodstat config file to set parameters
// controlling the subsequent logfile analysis.
//
void parse_config_line(struct global_names *gn,
					   struct global_state *gs,
					   char *line)
{
	char	c1[128];	// First config line param
	char	c2[128];	// Second config line param
	char	c3[128];	// Third config line param
	char	*cspace;	// Pointer to each word in config line

	if (sscanf(line, "%s", c1) <= 0)
		return;

	memset(c2, 0, 128);
	memset(c3, 0, 128);

	cspace = strchr(line, ' ');
	if (cspace != NULL)
	{
		sscanf(cspace, "%s", c2);
		cspace = strchr(cspace+strlen(c2), ' ');
		if (cspace != NULL)
			sscanf(cspace, "%s", c3);
	}

	if (strcmp(c1, "sourcefile\0") == 0)
	{	// Specify the filename for a single source logfile
		strcpy(gn->logfile, c2);
	}

	else if (strcmp(c1, "filelist\0") == 0)
	{	// Specify a file containing a list of source logfile names
		// ('sourcefile' option is ignored)
		strcpy(gn->logfile2, c2);
	}

	else if (strcmp(c1, "filelist_base\0") == 0)
	{	// Full path name of log files listed in the filelist file
		// is found by prefixing this string to each filename
		// (default is null)
		strcpy(gn->logfile2_base, c2);
	}

	else if (strcmp(c1, "gamestats\0") == 0)
	{	// List all games, start and stop times, etc
		 gs->stats |= GAME_STATS;
		 gs->stats &= ~GAME_NAME;
		 gs->stats &= ~GAME_TIME;
	}

	else if (strcmp(c1, "gamestats_bymap\0") == 0)
	{	// List all games using a specific map <map>
		 strcpy(gs->mapname, c2);
		 gs->stats |= GAME_STATS;
		 gs->stats |= GAME_NAME;
		 gs->stats &= ~GAME_TIME;
	}

	else if (strcmp(c1, "playerstats\0") == 0)
	{	// List all player stats for each game
		 gs->stats |= PLAYER_STATS;
	}

	else if (strcmp(c1, "playerstats_byplayer\0") == 0)
	{	// List a specific player <playername>
		 strcpy(gs->playername, c2);
		 gs->stats |= PLAYER_STATS;
		 gs->stats |= PLAYER_NAME;
	}

	else if (strcmp(c1, "gamestats_bytime\0") == 0)
	{	// List game starting at time <time> (in integer seconds)
		 gs->maptime = strtol(c2,(char **)NULL,0);
		 gs->stats |= GAME_STATS;
		 gs->stats &= ~GAME_NAME;
		 gs->stats |= GAME_TIME;
	}

	else if (strcmp(c1, "saylines\0") == 0)
	{	// Collect "say" lines for client IDs >= parameter NN
		// (requires the "gamestats" option in order to be printed out)
		 gs->saycid = strtol(c2,(char **)NULL,0);
		 if (gs->saycid < 0)
		 {
			 printf("Config error: 'saylines %d' - setting to 'saylines 0'\n", gs->saycid);
			 gs->saycid = 0;
		 }
	}

	else if (strcmp(c1, "humanreadable\0") == 0)
	{	// General directive to print output in more human readable form
		gs->stats |= HUMAN_READABLE;
	}

	else if (strcmp(c1, "resolve_again\0") == 0)
	{	// retry gethostbyname() for addresses that either do not exist
		// in the hosts_file or exist but failed earlier attempts
		// at name resolution.
		gs->dom_stats |= RESOLVE_AGAIN;
	}

	else if (strcmp(c1, "resolve_missing\0") == 0)
	{	// retry gethostbyname() for addresses NOT found in the file
		// specified by "hosts_file" (but don't attempt to re-resolve
		// addresses that are in the hosts_file but previously failed
		// resolution)
		gs->dom_stats |= RESOLVE_MISSING;
	}

	else if (strcmp(c1, "fake_resolve1\0") == 0)
	{	// Unresolved addresses are placed in fake meta-domain 
		// of "w-x-y-z.unresolved". When using domain name suffix
		// matching this will cause all unresolved ipaddr entries for
		// a given playername to count as a single client, regardless of
		// the actual ipaddr.
		gs->dom_stats |= FAKE_RESOLVE;
	}

	else if (strcmp(c1, "fake_resolve2\0") == 0)
	{	// Unresolved addresses are placed in fake meta-domain
		// of "w-x-y-z.w-x-y-z.unresolved". This differs from fake_resolve1
		// in that playernames with only one unresolved ipaddr but multiple
		// unique ports will count as one client, but multiple unresolved
		// ipaddr will still count as multiple clients when using name suffix
		// matching.
		gs->dom_stats |= FAKE_RESOLVE2;
	}

	else if (strcmp(c1, "faked_suffix_range\0") == 0)
	{	// Sets 1, 2, 3, or 4 to specify how many bytes of an unresolved
		// ipaddr to use when constructing a fake domain suffix for
		// seen_client list processing
		int xx = strtol(c2,(char **)NULL,0);
		if ((xx < 1) || (xx > 4))
		{
			 printf("Config error: faked_suffix_range <N> must be between 1 and 4.\n");
			 printf("Setting to faked_suffix_range to 4.\n");
			 xx = 4;
		}
		gs->faked_suffix = xx;
	}

	else if (strcmp(c1, "hosts_file\0") == 0)
	{	// Specify an ipaddr:domainname mapping file to use instead
		// of gethostbyaddr()
		strcpy(gn->resolvfile, c2);
	}

	else if (strcmp(c1, "hosts_file_uniq\0") == 0)
	{	
		// When reading in the ipaddr:name mappings from hosts_file,
		// check each entry and flag any redundant mappings
		gs->dom_stats |= CHK_HLIST_UNIQ;
	}

	else if (strcmp(c1, "resolve_asnum\0") == 0)
	{	// Resolve all the ip addresses into AS num information
		// (uses local AS mapping file, and also queries a remote
		// 'whois' database if "do_whois" config option is selected)
		gs->dom_stats |= RESOLVE_ASNUM;
	}

	else if (strcmp(c1, "asnum_file\0") == 0)
	{	// Specify an ipaddr->ASnumber mapping file to use instead
		// of querying the RADB online database
		strcpy(gn->asnumfile, c2);
	}

	else if (strcmp(c1, "do_whois\0") == 0)
	{	// Use the remote whois database to resolve ipaddr->ASnum
		gs->stats |= OPEN_WHOIS;
	}

	else if (strcmp(c1, "homedomain_file\0") == 0)
	{	// Specify file of domain suffixes to use when estimating
		// how many clients belong to a particular group (e.g. home users)
		strcpy(gn->namefragmentfile, c2);

	}

	else if (strcmp(c1, "retry_resolve_file\0") == 0)
	{
		// If you have a file containing "dig -x" results on a group
		// of IP addresses, these can also be added to the internal
		// host list using retry_resolve_file
		strcpy(gn->namereresolvfile, c2);
	}

	else if (strcmp(c1, "valid_domains\0") == 0)
	{	// List of domains that are valid for estimating total population
		// (particularly when calculating % of users who fall under
		// the domain specified by 'homedomainfile')
		strcpy(gn->domainlimitfile, c2);
	}

	else if (strcmp(c1, "include_unnamedplayer\0") == 0)
	{	// Include the UnnamedPlayer in aggregate stats calculations
		 gs->stats |= UNNAMED_PLAYER;
	}


	else if (strcmp(c1, "commandtokens\0") == 0)
	{	// List command token stats (-C) or player name stats (-CC)
		 gs->stats |= C_TOKEN_STATS;
	}

	else if (strcmp(c1, "clientnames\0") == 0)
	{
		// Dump sorted list of all seen clients
		gs->stats |= CLIENT_NAMES;
	}

	else if (strcmp(c1, "clients_NAT_range\0") == 0)
	{	// Use domain suffix matching to list client<->playername<->IPaddr stats
		// for a range of total client played times from 0 to <limit> (minutes)
		int xx = strtol(c2,(char **)NULL,0);
		if ((xx < 0) || (xx > 2000))
		{
			 printf("Config error: clients_NAT_range <N> must be between 0 and 2000.\n");
			 printf("Setting <N> to 0.\n");
			 xx = 0;
		}
		gs->playtimemin = xx * 60;
		gs->dom_stats |= SUFFIX_STATS;
	}

	else if (strcmp(c1, "reverse_NAT_range\0") == 0)
	{	// If this flag is UNSET, clients_NAT_range processing includes only
		// clients who played MORE than a range of playing times. If this flag
		// is SET we include only clients who played LESS.
		 gs->stats |= MAX_GAMETIME;
	}

	else if (strcmp(c1, "dump_dns_maps\0") == 0)
	{	// After resolving as many ip addresses as possible, dump the
		// mapping list gs->dnslist contents
		 gs->dom_stats |= DUMP_DOMAINMAP;
	}

	else if (strcmp(c1, "dump_domain_trees\0") == 0)
	{	// While calculating how many playername/clients exist who played
		// more than a certain number of minutes, dump the domain suffix
		// hierarchy of those clients to file "domain_trees.txt". Right here
		// we reset the file to null by opening then closing immediately.
		// (it will be reopened later for appending)
		FILE *tmpf;
		if ((tmpf = fopen("domain_trees.txt\0", "w")) == NULL)
		{
			printf("Fatal error: Couldn't initialize domain_trees.txt for writing. Exiting.\n");
			exit(-1);
		}
		else
			fclose(tmpf);
			
		 gs->dom_stats |= DUMP_DTREE;
	}

	else if (strcmp(c1, "dump_while_resolving\0") == 0)
	{	// Print the domain name resolution process to the specified
		// output file while it is occuring (subtle difference to the
		// 'dumpdomains' option in that "Failed:" entries will have
		// a valid Winsock2 error code rather than '0')
		 gs->dom_stats |= DUMP_RESOLVES;
	}

	else if (strcmp(c1, "dump_asnums_raw\0") == 0)
	{	// After reading in the AS number mappings from file, dump the
		// raw gs->asnumbers list
		 gs->dom_stats |= DUMP_ASNUMMAP;
	}

	else if (strcmp(c1, "dump_asnums_tree\0") == 0)
	{	// After reading in the game log files, dump the AS numbers in sorted
		// order of appearance.
		 gs->dom_stats |= DUMP_ASTREE;
	}

	else if (strcmp(c1, "range_increment\0") == 0)
	{	// Specify the increment for the player playtime parameter if needed
		int xx = strtol(c2,(char **)NULL,0);
		if ((xx < 0) || (xx > 40))
		{
			 printf("Config error: range_increment <N> must be between 0 and 40.\n");
			 printf("Setting range_increment to 4 minutes.\n");
			 xx = 4;
		}
		gs->rangeinc = xx * 60;
	}

	else if (strcmp(c1, "graph_include_regions\0") == 0)
	{	// For certain per-game/per-client stats, only count players
		// in "home domains"
		 gs->graph_stats |= INCLUDE_REGIONS;
	}

	else if (strcmp(c1, "graph_exclude_regions\0") == 0)
	{	// For certain per-game/per-client stats, only count players
		// NOT in "home domains"
		 gs->graph_stats |= EXCLUDE_REGIONS;
	}

	else if (strcmp(c1, "graph_unresolved_regions\0") == 0)
	{	// For certain per-game/per-client stats, only count players
		// with <unresolved> domain names
		 gs->graph_stats |= ONLY_UNRESOLVED;
	}

	else if (strcmp(c1, "graph_many\0") == 0)
	{	// For a number of per-game analyses where the
		// domains listed as 'home domains' limit the
		// clients being analysed, setting 'graph_many' will
		// cause analysis to be performed for EACH 'home domain'
		// separately, rather than as a union of all 'home domains'.
		 gs->graph_stats |= GRAPH_MANY;
	}

	else if (strcmp(c1, "graph_ping_histo\0") == 0)
	{
		// Create histograms of median ping times for every
		// client over every game. Client selection permuted
		// by graph_include_regions, graph_many, and graph_unresolved_regions
		 gs->graph_stats |= GRAPH_PINGS;
	}

	else if (strcmp(c1, "ping_histo_range\0") == 0)
	{	// Specify the max allowed end-of-game ping
		int xx1 = strtol(c2,(char **)NULL,0);
		int xx2 = strtol(c3,(char **)NULL,0);
		if ((xx1 < 1) || (xx1 > 64))
		{
			 printf("Config error: ping_histo_range <N> %d must be from 1 to 64\n", xx2);
			 printf("Setting it to 6 milliseconds\n");
			 xx1 = 6;
		}
		if ((xx2 < 10) || (xx2 > 998))
		{
			 printf("Config error: ping_histo_range %d <N> must be from 10 to 998\n", xx1);
			 printf("Setting it to 998 milliseconds\n");
			 xx2 = 998;
		}
		gs->ping_step = xx1;
		gs->max_ping = xx2;
		printf("ping_histo_range: bucket size %d ms, range up to %d ms\n", xx1, xx2);
		fprintf(gs->g_outf, "ping_histo_range: bucket size %d ms, range up to %d ms\n", xx1, xx2);
	}

	else if (strcmp(c1, "do_ping_histo\0") == 0)
	{
		// Force a number of ping scatter plot routines
		// to generate histograms instead
		gs->graph_stats |= PING_HISTO;
	}

	else if (strcmp(c1, "do_phisto_total\0") == 0)
	{
		// Force the routines dealing with CPhisto2: data to
		// create aggregate histograms rather than output
		// the individual histograms represented by each
		// CPhisto2: line
		gs->graph_stats |= PHISTO_TOTAL;
	}

	else if (strcmp(c1, "do_phisto_pergame\0") == 0)
	{
		// Force the routines dealing with CPhisto2: data to
		// create aggregate histograms rather than output
		// the individual histograms represented by each
		// CPhisto2: line
		gs->graph_stats |= PHISTO_GAME;
	}

	else if (strcmp(c1, "do_phisto_median\0") == 0)
	{
		// Force single_client_phisto() to dump XY points
		// of median ping (from CPhisto2 lines) vs timestamp
		// of the CPhisto2.
		gs->graph_stats |= PHISTO_MEDIAN;
	}

	else if (strcmp(c1, "do_phisto_jitter\0") == 0)
	{
		// Force single_client_phisto() to dump XY points
		// of ping jitter (from CPhisto2 lines) vs timestamp
		// of the CPhisto2.
		gs->graph_stats |= PHISTO_JITTER;
	}
	
	else if (strcmp(c1, "graph_ptime_histo\0") == 0)
	{	// Construct histogram(s) of total played time per hour versus
		// time (of day, or of week). May be used in conjunction
		// with 'graph_many' and/or 'graph_include_regions'.
		 gs->graph_stats |= PTIME_HISTO;
	}

	else if (strcmp(c1, "day_of_year\0") == 0)
	{
		// By default, played time histos are plotted vs hour
		// of the week. Set 'day_of_year' to change the x-axis
		// to be hour of day of year (mod ~365 days)
		 gs->graph_stats |= DAY_OF_YEAR;
	}

	else if (strcmp(c1, "hour_of_day\0") == 0)
	{
		// By default, played time histos are plotted vs hour
		// of the week. Set 'hour_of_day' to change the x-axis
		// to be hour of day (mod 24 hours)
		 gs->graph_stats |= HOUR_OF_DAY;
	}

	else if (strcmp(c1, "single_client_pings\0") == 0)
	{
		// Specify that the intra-game ping times of a specific
		// client (playername N, player suffix M) are to be analysed
		// (based on ClientPing: entries in the logfiles)
		strcpy(gs->playername, c2);
		strcpy(gs->playersuffix, c3);
		gs->ping_stats |= SINGLE_C_PING;
		printf("single_client_pings: %s/%s\n",
			gs->playername, gs->playersuffix);
		fprintf(gs->g_outf, "single_client_pings: %s/%s\n",
			gs->playername, gs->playersuffix);
	}

	else if (strcmp(c1, "single_client_phisto\0") == 0)
	{
		// Specify that the intra-game ping histos of a specific
		// client (playername N, player suffix M) are to be analysed
		// (based on CPhisto2: entries in the logfiles)
		strcpy(gs->playername, c2);
		strcpy(gs->playersuffix, c3);
		gs->ping_stats |= SINGLE_C_PHISTO;
		printf("single_client_phisto: %s/%s\n",
			gs->playername, gs->playersuffix);
		fprintf(gs->g_outf, "single_client_phisto: %s/%s\n",
			gs->playername, gs->playersuffix);
	}

	else if (strcmp(c1, "single_client_starttime\0") == 0)
	{
		// For single_client_pings or single_client_phisto,
		// further restrict analysis to the
		// game starting this many seconds from beginning of
		// single log file.
		 gs->maptime = strtol(c2,(char **)NULL,0);
	}

	else if (strcmp(c1, "topN_client_pings\0") == 0)
	{
		// Walk the seen_client list, and for all clients who totaled
		// more than N minutes playing time dump pings (or ping histogram)
		int xx = strtol(c2,(char **)NULL,0);
		if (xx < 0)
		{
			 printf("Config error: topN_client_pings <N> must be > 0\n");
			 printf("Setting it to 10 minutes\n");
			 xx = 6;
		}
		gs->ping_stats |= TOPN_CLIENT;
		gs->clienttime = xx*60;
		printf("topN_client_pings = %d minutes\n", xx);
		fprintf(gs->g_outf, "topN_client_pings = %d minutes\n", xx);
	}

	else if (strcmp(c1, "topN_client_phistos\0") == 0)
	{
		// Walk the seen_client list, and for all clients who totaled
		// more than N minutes playing time call single_client_phistos()
		int xx = strtol(c2,(char **)NULL,0);
		if (xx < 0)
		{
			 printf("Config error: topN_client_phistos <N> must be > 0\n");
			 printf("Setting it to 10 minutes\n");
			 xx = 6;
		}
		gs->ping_stats |= TOPN_PHISTOS;
		gs->clienttime = xx*60;
		printf("topN_client_phistos = %d minutes\n", xx);
		fprintf(gs->g_outf, "topN_client_phistos = %d minutes\n", xx);
	}

	else if (strcmp(c1, "median_client_pings\0") == 0)
	{
		// Walk the seen_client list and calculate/plot the median ping
		// vs 'jitter' for each client's pinglist.
		 gs->ping_stats |= MEDIAN_PINGS;
	}

	else if (strcmp(c1, "jitter_width\0") == 0)
	{
		// Jitter is the difference between the upper and lower bounds
		// of the ping histogram that covers N% of the ping samples.
		// This parameter sets "N/100"
		float xx = (float)atof(c2);
		if ((xx < 0.1) || (xx > 0.95))
		{
			 printf("Config error: jitter_width <N> must be between 0.1 and 0.95\n");
			 printf("Setting it to 0.9\n");
			 xx = (float)0.90;
		}
		gs->jitterwidth = xx;
		printf("jitter_width = %6.4f (%6.2f%%)\n", xx, 100*xx);
		fprintf(gs->g_outf, "jitter_width = %6.4f (%6.2f%%)\n", xx, 100*xx);
	}

	else if (strcmp(c1, "minimum_itemratio\0") == 0)
	{
		// If a player collects less than a certain number of items per
		// minute during a game that player is considered irrelevant ('tourist')
		// This config option sets the minimum value of 'itemratio'
		float xx = (float)atof(c2);
		if (xx < 0.0)
		{
			 printf("Config error: minimum_itemratio <N> must be above 0.0\n");
			 printf("Setting it to 1.0\n");
			 xx = (float)1.0;
		}
		gs->itemratio = xx;
		printf("itemratio = %6.4f\n", xx);
		fprintf(gs->g_outf, "itemratio = %6.4f\n", xx);
	}

	else if (strcmp(c1, "ping_vs_gamenum\0") == 0)
	{
		// Scatter plot median ping times vs total number of played games
		// for the entire seen_client list
		 gs->ping_stats |= PINGVSGAMENUM;
	}

	else if (strcmp(c1, "ping_vs_gametime\0") == 0)
	{
		// Scatter plot median ping times vs total played gametime
		// for the entire seen_client list
		 gs->ping_stats |= PINGVSGAMETIME;
	}

	else if (strcmp(c1, "ping_vs_frags\0") == 0)
	{
		// Scatter plot median ping times vs frag rate
		// for each player, each game.
		 gs->ping_stats |= PINGVSFRAGS;
	}

	else if (strcmp(c1, "ping_vs_gameratio\0") == 0)
	{
		// Scatter plot median ping times vs playedtimeratio
		// for each player, each game.
		 gs->ping_stats |= PINGVSGRATIO;
	}

	else if (strcmp(c1, "do_jitter_not_ping\0") == 0)
	{
		// Modifies ping_vs_gamenum and  ping_vs_gametime to plot
		// jitter rather than median ping.
		 gs->ping_stats |= DO_JITTER;
	}

	else if (strcmp(c1, "do_jitter_ping_ratio\0") == 0)
	{
		// Modifies ping_vs_gamenum and  ping_vs_gametime to plot
		// ratio of jitter to median ping.
		 gs->ping_stats |= DO_DEVIATION;
	}

	else if (strcmp(c1, "xgraph_output\0") == 0)
	{
		// For certain histos output xgraph-style breaks between weeks
		 gs->graph_stats |= XGRAPH_OUT;
	}

	else if (strcmp(c1, "extract_base_time\0") == 0)
	{
		// Extract base time for each segment of logfile from
		// logfile names themselves ("gja?-ddmmyy-hhmm-xxxx.log")
		 gs->stats |= BASE_TIME;
	}

	else if (strcmp(c1, "cumulative_gametime\0") == 0)
	{
		// Plot cumulative played time vs day/time of year
		 gs->stats |= CUMULATE_TIME;
	}

	else if (strcmp(c1, "do_half_days\0") == 0)
	{
		// Only output XY data points every 12 hours of calendar time
		// (initially for CUMULATE_TIME mode)
		 gs->graph_stats |= HALF_DAYS;
	}

	else if (strcmp(c1, "base_time_offset\0") == 0)
	{
		// Specify the number of hours to add to the basetimes
		// extracted from logfile (logfile name) when calculating
		// the 'real' time at which each event happens
		int xx = strtol(c2,(char **)NULL,0);
		if ((xx < -23) || (xx > 23))
		{
			 printf("Config error: base_time_offset <N> must be from -23 to 23 hours\n");
			 printf("Setting it to 0 hours\n");
			 xx = 0;
		}
		gs->time_offset = xx; // time_offset is in seconds, rather than hours.
		printf("base_time_offset = %d hours\n", xx);
		fprintf(gs->g_outf, "base_time_offset = %d hours\n", xx);
	}

	else if (strcmp(c1, "no_client_calculations\0") == 0)
	{
		// Suppress creation of the seen_client list
		// while parsing each logfile. Saves cycles when the ultimate goal
		// is to post-process the per-game stats only (the seen_game list)
		 gs->stats |= NO_CLIENT_CALC;
	}

	else if (strcmp(c1, "process_unix_ping\0") == 0)
	{
		strcpy(gn->unixpingfile, c2);
		gs->ping_stats |= PROCESS_NIXPING;
	}

	else
	{
		printf("Config file error: %s is not a valid token.\n", c1);
		fprintf(gs->g_outf, "Config file error: %s is not a valid token.\n", c1);
	}

	return;
}

// init_dns_lists()
//
// If required, read into memory the external lists of
// IPaddr:domainname mappings, and optionally the list of
// domain name fragments used to estimate categories of
// users.
//
// Also, if we're compiling under Windows then initialize
// Winsock2 so we can use gethostbyaddr() calls for name
// resolution.

void init_dns_lists(struct global_names *gn,
					struct global_state *gs)
{
	FILE *file;

	// Attempt to pre-open the ipaddr:domainname resolver file
	// and initialize winsock2 in case DNS resolution is required

	if (gn->resolvfile[0] != 0)
	{
		if ((file = fopen(gn->resolvfile, "rb")) == NULL)
		{
			printf("Opening ipaddr:name resolver file %s failed, aborting.\n",
				gn->resolvfile);
			exit(1);
		}

		// Read in the file and create 'hlist'
		printf("Loading DNS mappings from %s...", gn->resolvfile);
		fprintf(gs->g_outf, "Loading DNS mappings from %s...", gn->resolvfile);
		fflush(NULL);
		gs->dnslist = create_hlist(gs, file);
	}

	if (gn->namereresolvfile[0] != 0)
	{
		if ((file = fopen(gn->namereresolvfile, "rb")) == NULL)
		{
			printf("Opening ipaddr:name re-resolver file %s failed, aborting.\n",
				gn->namereresolvfile);
			exit(1);
		}

		// Read in the file and create 'hlist'
		printf("Updating DNS mappings from %s...", gn->namereresolvfile);
		fprintf(gs->g_outf, "Updating DNS mappings from %s...", gn->namereresolvfile);
		fflush(NULL);
		
		add_to_hlist(gs, file, &(gs->dnslist));
	}

	if (gn->namefragmentfile[0] != 0)
	{
		if ((file = fopen(gn->namefragmentfile, "rb")) == NULL)
		{
			printf("Opening home domain fragment file %s failed, aborting.\n",
				gn->namefragmentfile);
			exit(1);
		}

		// Read in the file and create 'flist'
		printf("Loading 'homedomains' from %s...", gn->namefragmentfile);
		fprintf(gs->g_outf, "Loading 'homedomains' from %s...", gn->namefragmentfile);
		fflush(NULL);
		
		gs->fragmentlist = install_fragmentlist(gs, file);
		fprintf(gs->g_outf, "Only domains listed in %s will be counted as HOME domains.\n",
			gn->namefragmentfile);
	}

	if (gn->domainlimitfile[0] != 0)
	{
		if ((file = fopen(gn->domainlimitfile, "rb")) == NULL)
		{
			printf("Opening domain limit list file %s failed, aborting.\n",
				gn->domainlimitfile);
			exit(1);
		}

		// Read in the file and create 'flist'
		printf("Loading limitlist from %s...", gn->domainlimitfile);
		fprintf(gs->g_outf, "Loading limitlist from %s...", gn->domainlimitfile);
		fflush(NULL);
		
		gs->limitlist = install_fragmentlist(gs, file);
		fprintf(gs->g_outf, "Only domain suffixes listed in %s will be counted in aggregate.\n",
			gn->domainlimitfile);
	}

#ifdef WIN32
	// Initialize winsock2
	unsigned short wvers;
	WSADATA wsadata;
	int err;
	wvers = MAKEWORD( 1, 1 ); 
	err = WSAStartup( wvers, &wsadata );
	if ( err != 0 )
	{
	 printf("WSAStartup() failed, quitting.\n");
	 exit(-1);
	}
#endif

	printf("\n");
	fprintf(gs->g_outf,"\n");
	
}



// init_asnum_lists()
//
// If required, read into memory the external list of
// ipaddr->ASnumber mappings.
// NB. The asnumber information is added to fields in
// in gs->dnslist (new entries are created if the asnum
// mapping file contains entries for ip addresses that aren't
// already in the gs->dnslist)

void init_asnum_list(struct global_names *gn,
					struct global_state *gs)
{
	char descr[256];
	char ipnum[32];
	char prefix[32];
	int asnum;
	int prefix_len;
	int multirec;
	int multiasnum;
	unsigned int tmpaddr;
	unsigned int tmpprefix;

	FILE *file;

	if (gn->asnumfile[0] == 0)
		return;

	if ((file = fopen(gn->asnumfile, "rb")) == NULL)
	{
		printf("Opening ipaddr->ASnumber mapping file %s failed, aborting.\n",
			gn->asnumfile);
		exit(1);
	}

	// Begin reading the ipaddr->ASnumber mapping file, line by
	// line. If the ipaddress matches one already in gs->dnslist
	// then update the AS number, prefix, and prefix_len information
	// otherwise create a new hostaddrlist entry and fill in the
	// AS information + a null domainname

	printf("Loading ASnumbers...");
	fflush(NULL);
	
	// Each line is of the following form:
	// w.x.y.z   ASnum   prefix  prefix_len   descr_string
	// (the final descr_string is an arbitrary human-readable description
	// string to associate with the ASnumber)
	// (the 'prefix' field is an unsigned 32 bit integer representing
	// the dotted-quad prefix as returned by previous 'whois' query to the RADB)

	while (fscanf(file, "%s %d %s %d %d %d %[^\r\n]",
		ipnum, &asnum, prefix, &prefix_len, &multirec, &multiasnum, descr) > 0)
	{

		descr[255] = 0; // Make sure the ascii description is null terminated

		tmpaddr = inet_addr(ipnum);	// Network byte order
		tmpprefix = inet_addr(prefix);

		push_new_asnum(gs, tmpaddr, asnum, tmpprefix,
			prefix_len, multirec, multiasnum, descr);

	}

	printf(" %d unique ASnum entries\n", gs->asnum_cnt);
	fprintf(gs->g_outf,"init_asnum_list: Saw %d unique ASnum entries\n", gs->asnum_cnt);
	
}


//
// install_fragmentlist()
//
// Build a memory-resident list of domain name fragments
// from the ascii text in file pointed to by fragmentfile
//
struct hostaddrlist *install_fragmentlist(struct global_state *gs,
											FILE *fragmentfile)
{
	struct hostaddrlist *hlist = NULL;
	struct hostaddrlist *tmphlist;
	char dname[128], dname2[256];
	int entries = 0;
	int redundant = 0;

	if (fragmentfile == NULL)
		return (NULL);

	// If rfile is non-null it points
	// to a file containing domain name fragments in ASCII
	// Read in that file to a local linear linked list

	while (fscanf(fragmentfile, "%[^\r\n]\n", dname2) > 0)
	{

		sscanf(dname2, "%s", dname);	// Strip leading/trailing spaces

		// First scan the existing entries to see if this new
		// entry is redundant
		if (dname[0] != '#') // Ignore comment lines
		{
			tmphlist = hlist;
			while (tmphlist != NULL)
			{
				if (strcmp(tmphlist->name, dname) == 0)
					break;
				tmphlist = tmphlist->next;
			}

			if (tmphlist == NULL)
			{
				// Getting to here means this is a new domain name fragment,
				// so create a new hostaddrlist entry

				if ((tmphlist =
					 (struct hostaddrlist *)malloc(sizeof(struct hostaddrlist)))
					 == NULL)
				{
				  fprintf(gs->g_outf, "install_fragmentlist(): Memory allocation failure, quitting.\n");
				  exit(-1);
				}
				
				memset((char *)tmphlist, 0, sizeof(struct hostaddrlist));

				// Allocate space for the name fragment, copy it over
				
				if ((tmphlist->name =
						(char *)malloc(strlen(dname)))
						 == NULL)
				{
				 fprintf(gs->g_outf, "install_fragmentlist(): Memory allocation failure, quitting.\n");
				 exit(-1);
				}

				strcpy(tmphlist->name, dname);

				// Put this new entry at head of hlist

				tmphlist->ipaddr = 0;
				tmphlist->next = hlist;
				hlist = tmphlist;

				entries++;
			}
			else
			{
				redundant++;
			}
		}

	}

	printf("%d unique entries (%d redundant entries).\n",
		entries, redundant);
	fprintf(gs->g_outf, "%d unique entries (%d redundant entries).\n",
		entries, redundant);

	return(hlist);
}


// sanity_check()
//
// Look for wierd combinations of flags after we've parsed the
// config file. Flag with error messages and return 0, or return
// with 1 for particularly bad combinations to indicate the main
// program should just quit.
//
// This section is totally incomplete.
//

int sanity_check(struct global_state *gs)
{

	// Check conditions associated with GRAPH_MANY mode

	if (FLAG_SET(gs->graph_stats, GRAPH_MANY))
	{
		if (FLAG_UNSET(gs->graph_stats, INCLUDE_REGIONS) &&
			FLAG_UNSET(gs->graph_stats, EXCLUDE_REGIONS))
		{
			printf("sanity_check failed: graph_many must also set graph_include_regions, exiting.\n");
			fprintf(gs->g_outf, "sanity_check failed: graph_many must also set graph_include_regions, exiting.\n");
			return(1);
		}
	
		if (gs->dnslist == NULL)
		{
			printf("sanity_check failed: graph_many requires dns mapping list, exiting.\n");
			fprintf(gs->g_outf, "sanity_check failed: graph_many requires dns mapping list, exiting.\n");
			return(1);
		}

		if (gs->fragmentlist == NULL)
		{
			printf("sanity_check failed: graph_many needs a home domain file, exiting.\n");
			fprintf(gs->g_outf, "sanity_check failed: graph_many needs a home domain file, exiting.\n");
			return(1);
		}
	}

	// Default for now is to accept everything if we get to here
	return(0);
}
